Package: WebGestaltR
Type: Package
Title: Gene Set Analysis Toolkit WebGestaltR
Version: 0.4.6
Date: 2023-05-31
Authors@R: c(
        person("Jing", "Wang", email = "jingwang.uestc@gmail.com", role = "aut"),
        person("Yuxing", "Liao", email = "yuxingliao@gmail.com", role = c("aut", "cre")),
        person("Eric", "Jaehnig", email = "Eric.Jaehnig@bcm.edu", role = c("ctb")),
        person("Zhiao", "Shi", email = "Zhiao.Shi@bcm.edu", role = c("ctb")),
        person("Quanhu", "Sheng", email = "shengqh@gmail.com", role = c("ctb"))
        )
Description: The web version WebGestalt <https://www.webgestalt.org> supports 12 organisms, 354 gene identifiers and 321,251 function categories. Users can upload the data and functional categories with their own gene identifiers. In addition to the Over-Representation Analysis, WebGestalt also supports Gene Set Enrichment Analysis and Network Topology Analysis. The user-friendly output report allows interactive and efficient exploration of enrichment results. The WebGestaltR package not only supports all above functions but also can be integrated into other pipeline or simultaneously analyze multiple gene lists.
License: LGPL
URL: https://github.com/bzhanglab/WebGestaltR
LazyLoad: yes
Depends: R (>= 3.3)
Imports: methods, dplyr, doRNG, readr, parallel (>= 3.3.2), doParallel
        (>= 1.0.10), foreach (>= 1.4.0), jsonlite, httr, rlang,
        svglite, igraph, whisker, apcluster, Rcpp
NeedsCompilation: yes
LinkingTo: Rcpp
RoxygenNote: 7.2.3
Packaged: 2023-06-01 15:42:39 UTC; Yuxing
Author: Jing Wang [aut],
  Yuxing Liao [aut, cre],
  Eric Jaehnig [ctb],
  Zhiao Shi [ctb],
  Quanhu Sheng [ctb]
Maintainer: Yuxing Liao <yuxingliao@gmail.com>
Repository: CRAN
Date/Publication: 2023-06-01 16:00:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-25 12:38:39 UTC; unix
Archs: WebGestaltR.so.dSYM

# Generated by roxygen2: do not edit by hand

export(GOSlimSummary)
export(IDMapping)
export(WebGestaltR)
export(WebGestaltRBatch)
export(WebGestaltR_batch)
export(affinityPropagation)
export(formatCheck)
export(goSlimSummary)
export(idMapping)
export(idToSymbol)
export(listArchiveURL)
export(listArchiveUrl)
export(listGeneSet)
export(listIDType)
export(listIdType)
export(listOrganism)
export(listReferenceSet)
export(loadGeneSet)
export(prepareGseaInput)
export(prepareInputMatrixGsea)
export(readGmt)
export(swGsea)
export(weightedSetCover)
import(grDevices)
import(graphics)
import(methods)
import(utils)
importFrom(Rcpp,sourceCpp)
importFrom(apcluster,apcluster)
importFrom(doParallel,registerDoParallel)
importFrom(doRNG,"%dorng%")
importFrom(dplyr,"%>%")
importFrom(dplyr,arrange)
importFrom(dplyr,bind_rows)
importFrom(dplyr,desc)
importFrom(dplyr,distinct)
importFrom(dplyr,filter)
importFrom(dplyr,inner_join)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,right_join)
importFrom(dplyr,select)
importFrom(foreach,"%dopar%")
importFrom(foreach,foreach)
importFrom(httr,GET)
importFrom(httr,POST)
importFrom(httr,content)
importFrom(httr,modify_url)
importFrom(igraph,V)
importFrom(igraph,get.adjacency)
importFrom(igraph,graph.edgelist)
importFrom(jsonlite,fromJSON)
importFrom(jsonlite,toJSON)
importFrom(parallel,makeCluster)
importFrom(parallel,mclapply)
importFrom(parallel,stopCluster)
importFrom(readr,cols)
importFrom(readr,read_tsv)
importFrom(readr,stop_for_problems)
importFrom(readr,write_tsv)
importFrom(rlang,.data)
importFrom(stats,IQR)
importFrom(stats,median)
importFrom(stats,p.adjust)
importFrom(stats,phyper)
importFrom(stats,quantile)
importFrom(stats,rnorm)
importFrom(stats,sd)
importFrom(svglite,svglite)
importFrom(tools,file_ext)
importFrom(whisker,rowSplit)
importFrom(whisker,whisker.render)
useDynLib(WebGestaltR)

WebGestalt R package is the R version of our well-known web application tool WebGestalt (www.webgestalt.org) that has on average 27,000 users from 140 countries and territories per year and has also been cited 371 in 2016. The advantage of this R package is that it can be easily integrated to other pipelines or simultaneously analyze multiple gene lists.

WebGestaltR function can perform popular enrichment analyses: ORA (Over-Representation Analysis), GSEA (Gene Set Enrichment Analysis) and NTA (Network Topology Analysis). Based on the user-uploaded gene list or gene list with scores (for GSEA method), WebGestaltR function will first map the gene list to entrez gene IDs and then summarize the gene list based on the GO (Gene Ontology) Slim data. After performing the enrichment analysis, WebGestaltR function also returns an user-friendly HTML report containing GO Slim summary and enrichment analysis result. If the functional categories have the DAG (directed acyclic graph) structure, the structure of the enriched categories can also be visualized in the report.

#' createReport
#'
#' Generate HTML report for ORA and GSEA
#'
#' @importFrom jsonlite toJSON
#' @importFrom whisker whisker.render
#'
#' @keywords internal
#'
createReport <- function(hostName, outputDirectory, organism="hsapiens", projectName, enrichMethod, geneSet, geneSetDes, geneSetDag, geneSetNet, interestingGeneMap, referenceGeneList, enrichedSig, geneTables, clusters, background, enrichDatabase=NULL, enrichDatabaseFile=NULL, enrichDatabaseType=NULL, enrichDatabaseDescriptionFile=NULL, interestGeneFile=NULL, interestGene=NULL, interestGeneType=NULL, collapseMethod="mean", referenceGeneFile=NULL, referenceGene=NULL, referenceGeneType=NULL, referenceSet=NULL, minNum=10, maxNum=500, fdrMethod="BH", sigMethod="fdr", fdrThr=0.05, topThr=10, reportNum=20, perNum=1000, p=1, dagColor="binary") {
	outputHtmlFile <- file.path(outputDirectory, paste0("Project_", projectName), paste0("Report_", projectName, ".html"))

	# if hostname starts with "file://", it is used as WebGestaltReporter
	if (startsWith(hostName, "file://")) {
		# change back hostName for web assets and browsers will cache it.
		hostName <- "http://www.webgestalt.org"
	}

	numAnnoRefUserId <- NULL
	dagJson <- list()
	allEnrichedSig <- enrichedSig
	repAdded <- FALSE
	if (organism!="others") {
		if (!is.null(enrichedSig) && reportNum < nrow(enrichedSig)) {
			if (enrichMethod == "ORA") {
				enrichedSig <- enrichedSig[1:reportNum, ]
			} else if (enrichMethod == "GSEA") {
				enrichedSig <- getTopGseaResults(enrichedSig, reportNum / 2)[[1]]
			}
			# Add representatives if they are not in top ReportNum. So could be more if ReportNum.is small and high redundancy in top
			numRes <- nrow(enrichedSig)
			enrichedSig <- keepRep(enrichedSig, allEnrichedSig, clusters$ap$representatives)
			enrichedSig <- keepRep(enrichedSig, allEnrichedSig, clusters$wsc$representatives)
			repAdded <- nrow(enrichedSig) > numRes
		}
		standardId <- interestingGeneMap$standardId
		if (enrichMethod == 'ORA') {
			interestGeneList <- unique(interestingGeneMap$mapped[[standardId]])
			numAnnoRefUserId <- length(intersect(interestGeneList,
				intersect(referenceGeneList, geneSet$gene)))
		}

		##### Summary Tab ########
		bodyContent <- summaryDescription(projectName, organism, interestGeneFile, interestGene, interestGeneType, enrichMethod, enrichDatabase, enrichDatabaseFile, enrichDatabaseType, enrichDatabaseDescriptionFile, interestingGeneMap, referenceGeneList, referenceGeneFile, referenceGene, referenceGeneType, referenceSet, minNum, maxNum, sigMethod, fdrThr, topThr, fdrMethod, allEnrichedSig, reportNum, perNum, p, geneSet, repAdded, numAnnoRefUserId, hostName)

		########### GOSlim summary #########################
		if(standardId=="entrezgene"){
			bodyContent <- paste(bodyContent, goSlimReport(projectName), sep='\n')
		}

		############ Enrichment result ##################
		if (!is.null(enrichedSig)) {
			bodyContent <- paste(bodyContent, enrichResultSection(enrichMethod, enrichedSig, geneSet, geneSetDes, geneSetDag, geneSetNet, clusters), seq='\n')
			if (!is.null(geneSetDag)) {
				if (!is.vector(geneSetDag)) {
					# for backward compatibility, it is unlisted for single dataset
					geneSetDag <- list(geneSetDag)
					names(geneSetDag) <- ifelse(is.character(enrichDatabase), enrichDatabase, gsub(".gmt", "", basename(enrichDatabaseFile), fixed=TRUE))
				}
				for (name in names(geneSetDag)) {
					dag <- geneSetDag[[name]]
					if (is.null(dag)) {
						#dagJson[[name]] <- list(NULL)
						next
					}
					dagRes <- expandDag(enrichedSig$geneSet, dag)
					dagEdges <- dagRes$edges
					dagNodes <- getDagNodes(enrichedSig, dagRes$allNodes, geneSetDes, enrichMethod, dagColor)
					dagJson[[name]] <- c(dagEdges, dagNodes)
				}
			}
		}
	} else {
		########### Organism is others. No mapping information #############
		############# Summary for the analysis ###################
		if (enrichMethod == 'ORA') {
			numAnnoRefUserId <- length(intersect(interestingGeneMap,
				intersect(referenceGeneList, geneSet$gene)))
		}
		if (!is.null(enrichedSig) && reportNum < nrow(enrichedSig)) {
			if (enrichMethod == "ORA") {
				enrichedSig <- enrichedSig[1:reportNum, ]
			} else if (enrichMethod == "GSEA") {
				enrichedSig <- getTopGseaResults(enrichedSig, reportNum / 2)[[1]]
			}
			# Add representatives if they are not in top ReportNum. So could be more if ReportNum.is small and high redundancy in top
			numRes <- nrow(enrichedSig)
			enrichedSig <- keepRep(enrichedSig, allEnrichedSig, clusters$ap$representatives)
			enrichedSig <- keepRep(enrichedSig, allEnrichedSig, clusters$wsc$representatives)
			repAdded <- nrow(enrichedSig) > numRes
		}

		bodyContent <- summaryDescription(projectName, organism, interestGeneFile, interestGene, interestGeneType, enrichMethod, enrichDatabase, enrichDatabaseFile, enrichDatabaseType, enrichDatabaseDescriptionFile, interestingGeneMap, referenceGeneList, referenceGeneFile, referenceGene, referenceGeneType, referenceSet, minNum, maxNum, sigMethod, fdrThr, topThr, fdrMethod, allEnrichedSig, reportNum, perNum, p, geneSet, repAdded, numAnnoRefUserId, hostName)

		############## Enrich Result ################
		if (!is.null(enrichedSig)) {
			bodyContent <- paste(bodyContent, enrichResultSection(enrichMethod, enrichedSig, geneSet, geneSetDes, geneSetDag, geneSetNet, clusters), seq='\n')
		}
		standardId <- NULL
	}
	if (is.null(enrichedSig)) {
		enrichedSig <- data.frame()
	}
	if (is.null(background)) {
		background <- data.frame()
	}
	version <- packageVersion("WebGestaltR")
	# use major and minor version numbers for JS lib. If API changes, version should be bumped
	# patch number should not matter
	version <- paste(version[1, 1], version[1, 2], sep=".")
	hasGeneSetDag = !is.null(geneSetDag)
	hasCytoscape <- hasGeneSetDag || !is.null(geneSetNet) # DAG or network needs cytoscape
	allDbNames <- unlist(c(enrichDatabase, unname(sapply(enrichDatabaseFile, function(x) {
		gsub(".gmt", "", basename(x), fixed=TRUE)
	}))))  # sapply on NULL will return a list

	header <- readLines(system.file("templates/header.mustache", package="WebGestaltR"))
	footer <- readLines(system.file("templates/footer.mustache", package="WebGestaltR"))
	template <- readLines(system.file("templates/template.mustache", package="WebGestaltR"))
	data <- list(hostName=hostName, bodyContent=bodyContent,
				organism=organism, enrichDatabaseJson=toJSON(allDbNames, auto_unbox=TRUE),
				sigJson=toJSON(enrichedSig, digits=16), insigJson=toJSON(background, digits=16),
				dagJson=toJSON(dagJson, auto_unbox=TRUE), hasGeneSetDag=hasGeneSetDag, version=version,
				clusterJson=toJSON(clusters), hasCytoscape=hasCytoscape,
				geneTableJson=toJSON(geneTables), standardId=standardId, numAnnoRefUserId=numAnnoRefUserId,
				methodIsGsea=enrichMethod=="GSEA", hasGeneSetDes=!is.null(geneSetDes)
				)
	cat(whisker.render(template, data, partials=list(header=header, footer=footer)), file=outputHtmlFile)
}

#' Wrapper convenient utility for mapping ids to "genesymbol"
#'
#' @inheritParams idMapping
#'
#' @importFrom tools file_ext
#'
#' @export
#' @rdname idMapping
idToSymbol <- function(organism="hsapiens", dataType="list", inputGeneFile=NULL,
                       inputGene=NULL, sourceIdType="ensembl_gene_id", collapseMethod="mean",
                       mappingOutput=FALSE, outputFileName=NULL, cache=NULL,
                       hostName="http://www.webgestalt.org/") {
	# various error checking
	errorTest <- .hostNameError(hostName)
	if(!is.null(errorTest)){
		return(errorTest)
	}
	errorTest <- .organismError(organism, hostName)
	if(!is.null(errorTest)){
		return(errorTest)
	}
	if(!(dataType %in% c("list", "rnk", "matrix"))){
		error <- paste("ERROR: Data type ",dataType," is not supported by idToSymbol.
						Please select from 'list', 'rnk' and 'matrix'.",sep="")
		cat(error)
		return(error)
	}
	errorTest <- .collapseMethodError(collapseMethod)
	if(!is.null(errorTest)){
		return(errorTest)
	}
	errorTest <- idTypeError(idType=sourceIdType, organism=organism, hostName=hostName, cache=cache)
	if(!is.null(errorTest)){
		return(errorTest)
	}
	# actual mapping
	if(dataType == "list" || dataType == "rnk"){
		if(mappingOutput && is.null(outputFileName)){
			outputFileName <- paste0("wgr_", dataType)
		}
		geneMap <- idMappingGene(organism=organism, dataType=dataType, inputGeneFile=inputGeneFile, inputGene=inputGene, sourceIdType=sourceIdType, targetIdType="genesymbol",  collapseMethod=collapseMethod, mappingOutput=mappingOutput, outputFileName=outputFileName, hostName=hostName)
		return(geneMap)
	} else if(dataType == "matrix"){
		if(mappingOutput && is.null(outputFileName)){
			outputFileName <- paste0("wgr_", dataType, "_converted.txt")
		}
		# cct or cbt
		if(!is.null(inputGeneFile)){
			inputMat <- .testMatrixFormat(inputGeneFile, collapseMethod)
		} else if(!is.null(inputGene)){
			inputMat <- .testMatrixFormat(inputGene, collapseMethod)
		}
		inputId <- rownames(inputMat)
		geneMap <- idMappingGene(organism=organism, dataType="list", inputGeneFile=NULL, inputGene=inputId, sourceIdType=sourceIdType, targetIdType="genesymbol",  mappingOutput=FALSE, hostName=hostName)
		idMap <- geneMap$mapped[,c(1,2)]
		id <- as.vector(idMap[,2])
		inputMat <- inputMat[idMap[,1],]
		inputMat <- mergeDuplicate(id,inputMat,collapseMethod)
		re <- list(data=inputMat,idMap=idMap)
		# write to output file
		if(mappingOutput){
			mtrx <- re$data
			mtrx <- cbind(rownames(mtrx),mtrx)
			colnames(mtrx)[1] <- "GeneSymbol"
			write.table(mtrx,file=outputFileName,row.names=F,col.names=T,sep="\t",quote=F)
		}
		return(re)
	}
}

.testMatrixFormat <- function(inputMat, collapseMethod="maxSD"){
	if(is.character(inputMat)){
		if(file_ext(inputMat)!="cct" && file_ext(inputMat)!="cbt"){
			stop("The extension of the input file should be 'cct' or 'cbt'. \n")
		}else{
			inputMat <- read.table(inputMat,header=TRUE,sep="\t",stringsAsFactors=FALSE,check.names=FALSE)
			geneId <- inputMat[,1]
			if(length(geneId)!=length(unique(geneId))){
				cat("The input data contain duplicate Id. The function will use ",collapseMethod," to collapse duplicate Id in each sample!\n",sep="")
				inputMat <- mergeDuplicate(geneId,inputMat[,2:ncol(inputMat)] ,collapseMethod)
			}else{
				inputMat <- inputMat[,c(2:ncol(inputMat))]
				rownames(inputMat) <- geneId
			}
		}
	}else{
		if(!is.matrix(inputMat) && !is.data.frame(inputMat)){
			stop("The type of input data should be a matrix or data.frame object. Other types of data are not supported by this package.!\n")
		}else{
			x <- apply(inputMat,2,function(e) return(class(e)=="numeric" || class(e)=="integer"))
			y <- all(x==TRUE)
			if(y==FALSE){
				stop("The input matrix or data.frame object should only contain numeric or integer values.\n")
			}
		}
	}
	if(ncol(inputMat)<6){
		stop("The data should contain at least six samples!\n")
	}
	if(length(which(inputMat %in% c("Inf","-Inf")))>0){
		stop("The input data contain Inf which may be gernated by some wrong operation, such as log(0) or 1/0. Please re-process the data and remove the Inf\n")
	}
	return(inputMat)
}

#' Create HTML Report for NTA
#'
#' @importFrom readr read_tsv cols
#' @importFrom jsonlite fromJSON toJSON
#' @importFrom whisker whisker.render rowSplit
#'
#' @keywords internal
#'
createNtaReport <- function(networkName, method, sigMethod, fdrThr, topThr, highlightType, outputDirectory, projectDir, projectName, hostName) {
	namePrefix <- paste(projectName, networkName, method, sep=".")
	seedsFn <- file.path(projectDir, paste0(namePrefix, "_seedsInSubnetwork.txt"))
	networkFn <- file.path(projectDir, paste0(namePrefix, "_randomWalkNetwork.txt"))
	candidateFn <- file.path(projectDir, paste0(namePrefix, "_candidate.txt"))
	enrichFn <- file.path(projectDir, paste0(namePrefix, "_enrichedResult.txt"))
	summaryFn <- file.path(projectDir, paste0(namePrefix, "_resultSummary.txt"))
	jsonFn <- file.path(projectDir, paste0(namePrefix, ".json"))

	if (startsWith(hostName, "file://")) {
		# change back hostName for web assets and browsers will cache it.
		hostName <- "http://www.webgestalt.org"
	}

	if (method == "Network_Retrieval_Prioritization") {
		highSeedsFn <- file.path(projectDir, paste0(namePrefix, "_highlightedSeeds.txt"))
	}
	## JSON for cytoscape of GO DAG
	dagJson <- readLines(jsonFn)
	dagData <- fromJSON(dagJson)
	## Get all GO term nodes, combine ID and Name
	goDataList <- lapply(dagData, function(x) paste(x[["data"]][["id"]], x[["data"]][["name"]]))
	goDataList <- as.character(goDataList[sapply(goDataList, length) > 0])

	seeds <- scan(seedsFn, "character")
	candidates <- read_tsv(candidateFn, col_names=c("candidate", "score"), col_types="cd-")
	candidates$score <- sprintf("%2.2E", candidates$score)

	enrichment <- read_tsv(enrichFn, col_names=c("goId", "goName", "c", "o", "geneInfo", "expect", "ratio", "rawP", "adjP"), skip=1, col_types=cols())

	summary <- readLines(summaryFn)

	network <- read_tsv(networkFn, col_names=c("source", "target"), col_types="cc")
	allNodes <- unique(c(network, recursive=TRUE))

	## Prepare JSON data for cytoscape of gene network, add edges data first
	cyJson <- lapply(split(network, seq(nrow(network))), function(x) list(data=as.list(x)))
	## Highlight (larger node, colored red in table for expansion)
	## for expansion, option to decide seeds or neighbors
	## for retrieval, highlight seeds
	if (method == "Network_Expansion") {
		seeds <- seeds
		highlight <- sapply(allNodes, function(x) x %in% seeds)
		if (highlightType != "Seeds") {
			highlight <- !highlight
		}
	} else {
		seeds <- scan(highSeedsFn, "character")
		highlight <- sapply(allNodes, function(x) x %in% seeds)
	}
	## add node data
	cyJson <- c(cyJson, mapply(function(x, y) list(data=list(id=x, highlight=y)), allNodes, highlight, SIMPLIFY=FALSE))
	cyJson <- toJSON(unname(cyJson), auto_unbox=TRUE)

	## Prepare GO enrichment table data
	#rowSplit still keeps each list element as data.frame
	enrichmentList <- lapply(unname(rowSplit(enrichment)), as.list)
	for (i in 1:nrow(enrichment)) {
		#decode geneInfo
		geneInfoList <- unlist(strsplit(enrichment[[i, "geneInfo"]], ";", fixed=TRUE))
		geneData <- vector("list", length(geneInfoList))
		for (j in 1:length(geneInfoList)) {
			splitInfo <- unlist(strsplit(geneInfoList[[j]], "|", fixed=TRUE))
			geneName <- splitInfo[1]
			label <- splitInfo[2]
			if ((method == "Network_Expansion" && highlightType == "Seeds" && label != 0) ||
					(method == "Network_Expansion" && highlightType != "Seeds" && label == 0)) {
				highlight <- TRUE
			} else {
				highlight <- FALSE
			}
			geneData[[j]] <- list(geneName=geneName, highlight=highlight)
		}
		enrichmentList[[i]]$geneInfo <- geneData
	}
	enrichmentJson <- toJSON(enrichmentList, auto_unbox=TRUE)

	version <- packageVersion("WebGestaltR")
	version <- paste(version[1, 1], version[1, 2], sep=".")
	netName <- paste(networkName, "net", sep="_")
	dagName <- paste(networkName, "dag", sep="_")

	data <- list(networkName=networkName, summary=summary, highlightType=highlightType,
				hostName=hostName,
				dagName=dagName, netName=netName,
				zipPath=paste0(projectName, ".zip"),
				toolboxDag=list(name=dagName, nodes=goDataList)
				)
	template <- readLines(system.file("templates/networkContent.mustache", package="WebGestaltR"))
	content <- whisker.render(template, data)

	data <- list(hostName=hostName,
				networkName=networkName,
				dagJson=dagJson,
				version=version,
				networkJson=cyJson,
				enrichmentJson=enrichmentJson,
				candidateJson=toJSON(candidates),
				seedJson=toJSON(seeds),
				method=method, sigMethod=sigMethod,
				threshold=ifelse(sigMethod=='fdr', fdrThr, topThr),
				networkContent=content
				)
	header <- readLines(system.file("templates/header.mustache", package="WebGestaltR"))
	footer <- readLines(system.file("templates/footer.mustache", package="WebGestaltR"))
	partials <- list(header=header, footer=footer)
	template <- readLines(system.file("templates/ntaTemplate.mustache", package="WebGestaltR"))
	outFn <- file.path(projectDir, paste0("Report_", projectName, ".html"))
	cat(whisker.render(template, data, partials=partials), file=outFn)

	file.remove(jsonFn)
}

#' Modify the link to highlight the genes in the pathways
#'
#' Currently, we only have wikipathway and kegg pathways that need to modify the link
#'
#' @keywords internal
linkModification <- function(enrichMethod, enrichPathwayLink, geneList, interestingGeneMap) {

	if (grepl("www.kegg.jp", enrichPathwayLink, fixed=TRUE)) {
		link <- keggLinkModification(enrichPathwayLink, geneList)
		return(link)
	}
	if (grepl("www.wikipathways.org", enrichPathwayLink, fixed=TRUE)) {
		link <- wikiLinkModification(enrichMethod, enrichPathwayLink, geneList, interestingGeneMap)
		return(link)
	}
	return(enrichPathwayLink)
}

keggLinkModification <- function(enrichPathwayLink,geneList){
	geneList <- gsub(";","+",geneList)
	enrichPathwayLink <- paste(enrichPathwayLink,"+",geneList,sep="")
	return(enrichPathwayLink)
}

wikiLinkModification <- function(enrichMethod, enrichPathwayLink, geneList, interestingGeneMap) {
	geneMap <- interestingGeneMap$mapped
	geneList <- unlist(strsplit(geneList,";"))
	geneMap <- filter(geneMap, .data$entrezgene %in% geneList)
	enrichPathwayLink <- paste0(enrichPathwayLink,
		paste0(sapply(geneMap$geneSymbol, function(x) paste0("&label[]=", x)), collapse="")
		#not many pathway have entrezgene xref. Using both also seem to interfere with coloring
		#paste0(sapply(geneMap$entrezgene, function(x) paste0("&xref[]=", x, ",Entrez Gene")), collapse="")
	)
	if (enrichMethod == "ORA") {
		enrichPathwayLink <- paste0(enrichPathwayLink, "&colors=", colorPos)
	} else if (enrichMethod == "GSEA") {
		scores <- filter(interestingGeneMap$mapped, .data$entrezgene %in% geneList)[["score"]]
		maxScore <- max(scores)
		minScore <- min(scores)
		tmp <- getPaletteForGsea(maxScore, minScore)
		palette <- tmp[[1]]
		breaks <- tmp[[2]]
		colors <- sapply(scores, function(s) palette[max(which(breaks <= s))])
		colorStr <- paste(gsub("#", "%23", colors, fixed=TRUE), collapse=",")
		enrichPathwayLink <- paste0(enrichPathwayLink, "&colors=", colorStr)
	}
	return(enrichPathwayLink)
}

loadInterestGene <- function(organism="hsapiens", dataType="list", inputGeneFile=NULL, inputGene=NULL, geneType="entrezgene", collapseMethod="mean", cache=NULL, hostName="https://www.webgestalt.org/", geneSet) {
	if (is.null(inputGeneFile) && is.null(inputGene)) {
		stop(interestGeneError(type="empty"))
	} else {
		if (organism!="others") {
			if (is.null(geneType)) {
				stop(interestGeneError(type="emptyType"))
			} else {
				mapRe <- .uploadGeneExistingOrganism(organism=organism, dataType=dataType, inputGeneFile=inputGeneFile, inputGene=inputGene, geneType=geneType, collapseMethod=collapseMethod, geneSet=geneSet, cache=cache, hostName=hostName)
			}
		} else {
			mapRe <- .uploadGeneOthers(dataType=dataType, inputGeneFile=inputGeneFile, inputGene=inputGene, geneSet=geneSet)
		}
	}

	#if organism is not others, the function will return a mapping result with mapped and unmapped list
	#if organism is others, the function will return a matrix with gene list
	return(mapRe)
}

#' @importFrom httr content
#' @importFrom readr read_tsv
loadReferenceGene <- function(organism="hsapiens", referenceGeneFile=NULL, referenceGene=NULL, referenceGeneType="entrezgene", referenceSet=NULL, collapseMethod="mean", hostName="https://www.webgestalt.org/", geneSet, interestGeneList, cache=NULL) {
	referenceGeneList <- NULL
	referenceGeneMap <- NULL

	if (is.null(referenceGeneFile) && is.null(referenceGene) && is.null(referenceSet)) {
		stop(referenceGeneError(type="empty"))
	} else {
		if (organism!="others") {
			if (!is.null(referenceGeneFile) || !is.null(referenceGene)) {
				if (is.null(referenceGeneType)) {
					stop(referenceGeneError(type="emptyType"))
				} else {
					mapRe <- .uploadGeneExistingOrganism(organism=organism, dataType="list", inputGeneFile=referenceGeneFile, inputGene=referenceGene, geneType=referenceGeneType, collapseMethod=collapseMethod, geneSet=geneSet, cache=cache, hostName=hostName)
					geneStandardId <- identifyStandardId(hostName=hostName, idType=referenceGeneType, organism=organism, type="interest", cache=cache)
					referenceGeneList <- mapRe$mapped[[geneStandardId]]
				}
			} else { ### referenceGeneFile and referenceGene are both NULL. But referenceSet is not NULL
				refS <- listReferenceSet(organism=organism, hostName=hostName, cache=cache)
				if (length(which(refS==referenceSet))==0) {
					stop(referenceGeneError(type="existingRef"))
				}
				refStandardId <- identifyStandardId(hostName=hostName, idType=referenceSet, organism=organism, type="reference", cache=cache)
				if (startsWith(hostName, "file://")) {
					# Getting data from local directory in the old way
					refPath <- removeFileProtocol(file.path(hostName, "reference", paste0(paste(organism, referenceSet, refStandardId, sep="_"), ".table")))
					referenceGeneList <- read_tsv(refPath, col_names=FALSE, col_types="c-")[[1]]
				} else {
					response <- cacheUrl(file.path(hostName, "api", "reference"), cache=cache, query=list(organism=organism, referenceSet=referenceSet, standardId=refStandardId))
					if (response$status_code != 200) {
						stop(webRequestError(response))
					}
					# API now just returns one single column
					referenceGeneList <- read_tsv(content(response), col_names=FALSE, col_types="c")[[1]]
				}
			}
		} else { ## For other organisms
			if (!is.null(referenceGeneFile) || !is.null(referenceGene)) {
				referenceGeneList <- .uploadGeneOthers(dataType="list", inputGeneFile=referenceGeneFile, inputGene=referenceGene, geneSet=geneSet)
				referenceGeneList <- unique(referenceGeneList)
			} else {
				stop(referenceGeneError(type="empty"))
			}
		}
	}

	## compare interest gene list and reference gene list
	if (length(intersect(interestGeneList, intersect(referenceGeneList, geneSet$gene)))==0) {
		stop(referenceGeneError(type="interestEmpty"))
	}
	return(referenceGeneList)
}


#' @importFrom dplyr filter
.uploadGeneExistingOrganism <- function(organism, dataType, inputGeneFile, inputGene, geneType, collapseMethod, geneSet, cache, hostName) {
	geneMap <- idMapping(organism=organism, dataType=dataType, inputGeneFile=inputGeneFile, inputGene=inputGene, sourceIdType=geneType, targetIdType=NULL, collapseMethod=collapseMethod, mappingOutput=FALSE, cache=cache, hostName=hostName)

	#gene_standardId <- identifyStandardId(hostName=hostName,idtype=geneType,organism=organism,type="interest")  ##identifyStandardId in idMappingComponent.R
	#if(gene_standardId!=databaseStandardId){  ###the standardId of the input genes should be the same with the standardarId of the functional database
#	return(interestGeneError(type="unmatch"))
#}

	geneMapMappedList <- geneMap$mapped
	standardId <- geneMap$standardId

	geneList <- as.character(unique(geneMapMappedList[[standardId]]))
	ov <- intersect(geneList, geneSet$gene)

	if (length(ov)==0) {
		stop(interestGeneError(type="unannotated"))
	}

	###Because if all genes are annotated to only one category, GSEA will return the error, we need to avoid this error by reporting the error in the R#
	geneSets <- unique((filter(geneSet, .data$gene %in% geneList))[["geneSet"]])
	if (length(geneSets) == 1) {
		stop(interestGeneError(type="onlyOne"))
	}
	return(geneMap)
}


#' @importFrom dplyr filter
.uploadGeneOthers <- function(dataType,inputGeneFile,inputGene,geneSet){
	inputGene <- formatCheck(dataType=dataType,inputGeneFile=inputGeneFile,inputGene=inputGene)

	if (dataType == "list") {
		geneList = inputGene
	} else if (dataType == "rnk") {
		geneList = inputGene$gene
	}

	ov <- intersect(geneList, geneSet$gene)
	if (length(ov)==0) {
		stop(interestGeneError(type="unannotated"))
	}

	###Because if all genes are annotated to only one category, GSEA will return the error, we need to avoid this error by reporting the error in the R#
	geneSets <- unique((filter(geneSet, .data$gene %in% geneList))[["geneSet"]])
	if (length(geneSets) == 1) {
		stop(interestGeneError(type="onlyOne"))
	}
	return(inputGene)
}

idMappingInput <- function(dataType="list",inputGeneFile,inputGene){
	if(dataType=="gmt"){
		if(!is.null(inputGeneFile)){
			inputGene <- readGmt(inputGeneFile)
			return(inputGene)
		}else{
			stop(gmtFormatError("empty"))
		}
	}else{
		inputGene <- formatCheck(dataType=dataType,inputGeneFile=inputGeneFile,inputGene=inputGene)
		return(inputGene)
	}
}

#' @importFrom httr content
#' @importFrom jsonlite fromJSON
identifyStandardId <- function(hostName, idType, organism, type, cache) {
	if (startsWith(hostName, "file://")) {
		if (type=="interest") {
			summaryPath <- removeFileProtocol(file.path(hostName, "idtypesummary.json"))
		}
		if (type=="reference") {
			summaryPath <- removeFileProtocol(file.path(hostName, "referencesetsummary.json"))
		}
		jsonData <- fromJSON(summaryPath)
		idTypes <- jsonData[[organism]]
		names <- idTypes$name
		standardIds <- idTypes$type
	} else {
		if (type=="interest") {
			response <- cacheUrl(file.path(hostName, "api", "summary", "idtype"), cache)
		}
		if (type=="reference") {
			response <- cacheUrl(file.path(hostName, "api", "summary", "referenceset"), cache)
		}
		if (response$status_code != 200) {
			stop(webRequestError(response))
		}
		jsonData <- content(response)
		idTypes <- jsonData[[organism]]
		names <- unlist(lapply(idTypes, function(e) return(e$name)))
		standardIds <- unlist(lapply(idTypes,function(e) return(e$type)))
	}
	idTypes <- data.frame(name=names, standardId=standardIds, stringsAsFactors=FALSE)
	return(filter(idTypes, .data$name == idType)[[1, "standardId"]])
}

#' @importFrom dplyr select distinct %>%
#' @importFrom readr write_tsv
idMappingOutput <- function(outputFileName, mappingList, unmappedList, dataType, sourceIdType, targetIdType) {
	if (length(unmappedList)>0) {
		write(unmappedList, paste0(outputFileName, "_unmappedList.txt"))
	}
	if (dataType == "list" | dataType == "rnk") {
		dataType = "txt"
	}
	fileName <- paste0(outputFileName, "_mappedList_from_", sourceIdType, "_to_", targetIdType, ".", dataType)
	if(dataType=="gmt"){
		genes <- tapply(mappingList[[targetIdType]], mappingList$geneSet, paste, collapse="\t")
		gmtDf <- mappingList %>% select(.data$geneSet, .data$link) %>% distinct()
		gmtDf$genes = genes[gmtDf$geneSet]
		write_tsv(gmtDf, fileName, col_names=FALSE)
	}else{
		write_tsv(mappingList, fileName)
	}
}

replace_null <- function(x) {
	lapply(x, function(x) {
		if (is.list(x)) {
			replace_null(x)
		} else {
		if (is.null(x)) NA else(x)
		}
	})
}

#' @importFrom whisker whisker.render
goSlimReport <- function(projectName) {
	goSlimPicPath <- paste0("goslim_summary_", projectName, ".png")
	template <- readLines(system.file("templates/goSlimReport.mustache", package="WebGestaltR"))
	return(whisker.render(template, list(goSlimPicPath=goSlimPicPath)))
}

#' @importFrom httr POST content
#' @importFrom dplyr inner_join select filter left_join %>%
idMappingGene <- function(organism="hsapiens", dataType="list", inputGeneFile=NULL, inputGene=NULL, sourceIdType, targetIdType, collapseMethod="mean", mappingOutput=FALSE,  outputFileName="", hostName="https://www.webgestalt.org/") {

	###########Check input data type###############
	inputGene <- idMappingInput(dataType=dataType,inputGeneFile=inputGeneFile,inputGene=inputGene)

	##########ID Mapping Specify to gene level###############
	if(dataType=="list"){
		inputGeneL <- unique(inputGene)
	}

	if(dataType=="rnk"){
		######Collapse the gene ids with multiple scores##########
		x <- tapply(inputGene$score, inputGene$gene, collapseMethod)
		inputGene <- data.frame(gene=names(x),score=as.numeric(x),stringsAsFactors=FALSE)
		inputGeneL <- inputGene$gene
		colnames(inputGene) <- c(sourceIdType,"score")
	}

	if(dataType=="gmt"){
		colnames(inputGene) <- c("geneSet", "description", sourceIdType)
		inputGeneL <- unique(inputGene[[sourceIdType]])
	}
	if (startsWith(hostName, "file://")) {
		# old way of mapping with mapping files. Now only used for WebGestaltReporter when hostName is file protocol
		sourceMap <- read_tsv(
			removeFileProtocol(file.path(hostName, "xref", paste(organism, sourceIdType, "entrezgene.table", sep="_"))),
			col_names=c("entrezgene", "userId"), col_types="cc", quote=""
		) %>% filter(.data$userId %in% inputGeneL)
		symbolMap <- read_tsv(
			removeFileProtocol(file.path(hostName, "xref", paste(organism, "genesymbol", "entrezgene.table", sep="_"))),
			col_names=c("entrezgene", "geneSymbol"), col_types="cc", quote=""
		)
		nameMap <- read_tsv(
			removeFileProtocol(file.path(hostName, "xref", paste(organism, "genename", "entrezgene.table", sep="_"))),
			col_names=c("entrezgene", "geneName"), col_types="cc", quote=""
		)
		sourceMap <- sourceMap %>% left_join(symbolMap, by=c("entrezgene")) %>% left_join(nameMap, by=c("entrezgene"))

		if (targetIdType %in% c("entrezgene", sourceIdType)) {
			mappedInputGene <- sourceMap
		} else {
			targetMap <- read_tsv(removeFileProtocol(file.path(hostName, "xref", paste(organism, targetIdType, "entrezgene.table", sep="_"))),
				col_names=c("entrezgene", targetIdType), col_types="cc", quote="")
			mappedInputGene <- inner_join(sourceMap, targetMap, by=c("entrezgene"))
		}
		if (nrow(mappedInputGene) == 0) { return(idMappingError("empty")) }
		mappedInputGene <- mappedInputGene %>%
			select(.data$userId, .data$geneSymbol, .data$geneName, targetIdType)
		unmappedIds <- setdiff(inputGeneL, mappedInputGene$userId)

	} else {
		# new way uses web server API
		mapR <- POST(file.path(hostName, "api", "idmapping"), encode="json",
			body=list(organism=organism, sourceType=sourceIdType,
			targetType=targetIdType, ids=inputGeneL)
		)

		if (mapR$status_code != 200) {
			stop(webRequestError(mapR))
		}
		mapR <- content(mapR)
		if (mapR$status == 1) {
			stop(webApiError(mapR))
		}

		mappedIds <- mapR$mapped
		unmappedIds <- unlist(mapR$unmapped)
		if (is.null(targetIdType)) {
			targetIdType <- mapR$standardId
		}

		if (length(mappedIds) == 0) { stop(idMappingError("empty")) }

		names <- c("sourceId", "geneSymbol", "geneName", "targetId")
		mappedInputGene <- data.frame(matrix(unlist(lapply(replace_null(mappedIds), FUN=function(x) { x[names] })), nrow=length(mappedIds), byrow=TRUE), stringsAsFactors=FALSE)
		colnames(mappedInputGene) <- c("userId", "geneSymbol", "geneName", targetIdType)
	}

	if (dataType=="list") {
		inputGene <- mappedInputGene
	} else if (dataType=="rnk") {
		inputGene <- inner_join(mappedInputGene, inputGene, by=c("userId"=sourceIdType))
	} else if (dataType=="gmt") {
		inputGene <- inner_join(mappedInputGene, inputGene, by=c("userId"=sourceIdType)) %>%
			select(.data$geneSet, .data$description, .data$userId, .data$geneSymbol, .data$geneName, targetIdType)
	}

	if (targetIdType != "entrezgene" && sourceIdType!=targetIdType) {
		entrezgeneMapRes <- idMappingGene(organism, dataType="list", inputGene=inputGeneL, sourceIdType=sourceIdType, targetIdType="entrezgene", hostName=hostName)
		inputGene <- left_join(inputGene, entrezgeneMapRes$mapped, by="userId")

		if (dataType=="list") {
			inputGene <- select(inputGene, .data$userId, geneSymbol=.data$geneSymbol.x, geneName=.data$geneName.x, .data$entrezgene, targetIdType)
		} else if (dataType=="rnk") {
			inputGene <- select(inputGene, .data$userId, geneSymbol=.data$geneSymbol.x, geneName=.data$geneName.x, .data$entrezgene, targetIdType, .data$score)
		} else if (dataType=="gmt") {
			inputGene <- select(inputGene, .data$geneSet, .data$description, .data$userId, geneSymbol=.data$geneSymbol.x, geneName=.data$geneName.x, .data$entrezgene, targetIdType)
		}
	}
	inputGene$gLink <- paste0("https://www.ncbi.nlm.nih.gov/gene/?term=", inputGene$entrezgene)

	#############Output#######################
	if (mappingOutput) {
		idMappingOutput(outputFileName, inputGene, unmappedIds, dataType, sourceIdType, targetIdType)
	}
	r <- list(mapped=inputGene,unmapped=unmappedIds)
	return(r)
}

#' List WebGestalt Servers
#'
#' List available WebGestalt servers.
#'
#'
#' @return A data frame of available servers.
#'
#' @importFrom readr read_tsv
#' @export
#' @aliases listArchiveURL
#'
listArchiveUrl <- function(){
	archiveUrl <- read_tsv("http://www.webgestalt.org/archiveURL.txt", col_names=FALSE)
	return(archiveUrl)
}

#' @export
listArchiveURL <- function(...) {
	warning("Function listArchiveURL is deprecated and changed to listArchiveUrl!\n")
	return(listArchiveUrl(...))
}

#' List Organisms
#'
#' List supported organisms on WebGestalt server.
#'
#' @inheritParams WebGestaltR
#'
#' @return A list of supported organisms.
#'
#' @importFrom httr content
#' @importFrom jsonlite fromJSON
#' @export
#'
listOrganism <- function(hostName="http://www.webgestalt.org/", cache=NULL) {
	if (startsWith(hostName, "file://")) {
		jsonData <- fromJSON(removeFileProtocol(file.path(hostName, "idtypesummary.json")))
	} else {
		response <- cacheUrl(file.path(hostName, "api", "summary", "idtype"), cache)
		if (response$status_code != 200) {
			return(webRequestError(response))
		}
		jsonData <- content(response)
	}
	organisms <- names(jsonData)
	return(organisms)
}

#' enrichResultSection
#'
#' Conditionally render template of main result section. Actual work is carried out in front end
#'
#' @importFrom whisker whisker.render
#' @importFrom dplyr select distinct filter
#' @importFrom jsonlite toJSON
#' @keywords internal
#'
enrichResultSection <- function(enrichMethod, enrichedSig, geneSet, geneSetDes, geneSetDag, geneSetNet, clusters) {
	if ('database' %in% colnames(geneSet)) {
		#multiple databases
		netDatabases <- names(geneSetNet[!sapply(geneSetNet, is.null)])
		setSource <- geneSet %>% select(.data$geneSet, .data$database) %>%
			distinct() %>%
			filter(.data$geneSet %in% enrichedSig$geneSet)
		setsWithNetJson <- toJSON((filter(setSource, .data$database %in% netDatabases))$geneSet)
		hasGeneSetDag <- length(geneSetDag[!sapply(geneSetDag, is.null)]) > 0
		hasMultipleDatabases <- TRUE
	} else {
		setsWithNetJson <- toJSON(!is.null(geneSetNet), auto_unbox=TRUE)
		hasGeneSetDag <- !is.null(geneSetDag)
		hasMultipleDatabases <- FALSE
	}

	data <- list(methodIsOra=enrichMethod=='ORA',
				hasGeneSetDag=hasGeneSetDag,
				hasMultipleDatabases=hasMultipleDatabases,
				setsWithNetJson=setsWithNetJson,
				hasAp=!is.null(clusters$ap),
				hasWsc=!is.null(clusters$wsc)
				)
	template <- readLines(system.file("templates/enrichResultSection.mustache", package="WebGestaltR"))
	return(whisker.render(template, data))
}

mapUserId <- function(enrichedSig,geneColumn,interestingGeneMap){
	####map entrez gene back to the original user id and add one more column to the enrichedSig
	standardId <- interestingGeneMap$standardId
	mapgene <- interestingGeneMap$mapped[, c("userId", standardId)]
	gene <- enrichedSig[[geneColumn]]
	gene <- strsplit(gene,";")
	gene <- unlist(lapply(gene,geneM,mapgene))
	enrichedSig <- data.frame(enrichedSig, userId=gene, stringsAsFactors=FALSE)
	return(enrichedSig)
}

geneM <- function(geneList,mappingTable){
	if(length(geneList)==1 && is.na(geneList)){
		###The categories outputted from GSEA may not have leading edge genes. TODO: obsolete
		return(NA)
	}else{
		u <- mappingTable[mappingTable[[2]] %in% geneList, ][[1]]
		# although user ID could contain ;, like in some gene symbols.
		# but this is only concatenated in output
		u <- paste(u,collapse=";")
		return(u)
	}
}

#' @importFrom dplyr select
getGeneTables <- function(organism, enrichedSig, geneColumn, interestingGeneMap) {
	if (organism != "others") {
		standardId <- interestingGeneMap$standardId
		mapping <- select(interestingGeneMap$mapped, .data$userId, .data$geneSymbol, .data$geneName, .data$gLink, standardId)
		if ("score" %in% colnames(interestingGeneMap$mapped)) {
			mapping$score <- interestingGeneMap$mapped$score
		}
	}
	table <- list()
	for (i in 1:nrow(enrichedSig)) {
		genes <- enrichedSig[[i, geneColumn]]
		geneSetId <- enrichedSig[[i, "geneSet"]]
		if (length(genes) == 1 && is.na(genes)) {
			table[[geneSetId]] <- list()
		} else {
			genes <- unlist(strsplit(genes, ";"))
			if (organism != "others") {
				table[[geneSetId]] <- mapping[mapping[[standardId]] %in% genes, ]
			} else {
				table[[geneSetId]] <- data.frame("userId"=genes)
			}
		}
	}
	return(table)
}

#' @importFrom dplyr filter bind_rows
getTopGseaResults <- function(results, topThr) {
	if (is.wholenumber(topThr)) {
		posThr <- topThr
		negThr <- topThr
	} else {
		posThr <- floor(topThr) + 1
		negThr <- floor(topThr)
	}
	posRes <- filter(results, .data$normalizedEnrichmentScore > 0)
	if (nrow(posRes) > posThr) {
		posSig <- posRes[1:posThr, ]
		posInsig <- posRes[(posThr+1):nrow(posRes), ]
	} else {
		posSig <- posRes
		posInsig <- NULL
	}
	negRes <- filter(results, .data$normalizedEnrichmentScore < 0)
	if (nrow(negRes) > negThr) {
		negSig <- negRes[1: negThr, ]
		negInsig <- negRes[(negThr+1):nrow(negRes), ]
	} else {
		negSig <- negRes
		negInsig <- NULL
	}
	sig <- bind_rows(posSig, negSig)
	insig <- bind_rows(posInsig, negInsig)
	if (nrow(sig) == 0) {
		sig <- NULL
	}
	if (nrow(insig) == 0) {
		insig <- NULL
	}
	return(list(sig, insig))
}

#' keepRep
#'
#' Add representatives of redundancy-reduced clusters to
#' topResult if they are missing.
#'
#' @keywords internal
#'
keepRep <- function(topResult, allResult, reps) {
	missing <- NULL
	for (rep in reps) {
		if (!rep %in% topResult$geneSet) {
			missing <- c(missing, rep)
		}
	}
	if (!is.null(missing)) {
		topResult <- rbind(topResult, allResult[allResult$geneSet %in% missing, ])
	}
	return(topResult)
}

removeFileProtocol <- function(path) {
	return(normalizePath(sub("^file://", "", path), mustWork=FALSE))
}

sanitizeFileName <- function(name) {
	return(gsub("[[:punct:]]", "_", name))
}

#' Affinity Propagation
#'
#' Use affinity propagation to cluster similar gene sets to reduce redundancy in report.
#'
#' @param idsInSet A list of set names and their member IDs.
#' @param score A vector of addible scores with the same length used to assign input preference;
#'  higher score has larger weight, i.e. -logP.
#'
#' @return A list of \code{clusters} and \code{representatives} for each cluster.
#' \describe{
#'  \item{clusters}{A list of character vectors of set IDs in each cluster.}
#'  \item{representatives}{A character vector of representatives for each cluster.}
#' }
#'
#' @export
#' @importFrom apcluster apcluster
#' @importFrom stats rnorm
#' @author Zhiao Shi, Yuxing Liao
affinityPropagation <- function(idsInSet, score) {
	cat("Begin affinity propagation...\n")
	# compute the similiarity and input preference vector
	ret <- jaccardSim(idsInSet, as.numeric(score))

	sim.mat <- ret$sim.mat
	ip.vec <- ret$ip.vec

	apRes <- apcluster(sim.mat,p=ip.vec)
	#sort clusters to make exemplar the first member
	clusters <- vector(mode="list", length(apRes@clusters))
	if(length(apRes@clusters) == 0){
		return(NULL)
	}
	for (i in 1:length(apRes@clusters)) {
		exemplar <- apRes@exemplars[[i]]
		clusters[[i]] <- apRes@clusters[[i]][order(apRes@clusters[[i]] == exemplar, decreasing=TRUE)]
	}
	cat("End affinity propagation...\n")
	return(list(clusters=sapply(clusters, names), representatives=names(apRes@exemplars)))
}

#' Jaccard Similarity
#'
#' Calculate Jaccard Similarity.
#'
#' @inheritParams affinityPropagation
#'
#' @return A list of similarity matrix \code{sim.mat} and input preference vector \code{ip.vec}.
#'
#' @importFrom stats median
#' @author Zhiao Shi, Yuxing Liao
jaccardSim <- function(idsInSet, score){
	# first find out the union of sets, sorted
	all.genes <- sort(unique(unlist(idsInSet)))
	overlap.mat <- sapply(idsInSet, function(x) {as.integer(all.genes %in% x)})

	num <- length(idsInSet)
	sim.mat <- matrix(1, num, num)
	colnames(sim.mat) <- colnames(overlap.mat)

	if (num == 1) {
		return(list(sim.mat=sim.mat, ip.vec=c(1)))
	}

	for (i in 1:(num-1)) {
		for (j in (i+1):num) {
			jaccardIndex <- sum(bitwAnd(overlap.mat[, i], overlap.mat[, j])) / sum(bitwOr(overlap.mat[, i], overlap.mat[, j]))
			sim.mat[i, j] <- jaccardIndex
			sim.mat[j, i] <- jaccardIndex
		}
	}
	# if there is no overlap, set the similarity to -Inf
	sim.mat[sim.mat == 0] <- -Inf
	# check sim.mat to see if it is identical for each pair
	if (max(sim.mat) == min(sim.mat)) {
		# this will generate error, so randomy add some noise to off diagonal elements
		mat.siz <- dim(sim.mat)[1]
		rand.m <- matrix(rnorm(mat.siz*mat.siz,0,0.01),mat.siz)
		# make it symmetric
		rand.m[lower.tri(rand.m)] = t(rand.m)[lower.tri(rand.m)]
		sim.mat <- sim.mat + rand.m
		# make diagonal all 1
		diag(sim.mat) <- 1
	}
	# set the input preference (IP) for each gene set
	# give higher IP to gene set with larger -logP (remove sign)
	# IP <- maxScore for gene set with largest -logP value
	# IP <- minScore for gene set with smallest -logP value
	# other gene sets will have linearly interpolated IP value
	max.sig <- max(score)
	min.sig <- min(score)

	minScore <- 0
	tmp.sim.mat <- sim.mat
	tmp.sim.mat[!is.finite(tmp.sim.mat)] <- NA
	# get the median excluding -Inf
	maxScore <- median(tmp.sim.mat, na.rm=TRUE)
	if (abs(max.sig - min.sig) < .Machine$double.eps^0.5) {
		ip.vec <- NA
	} else{
		ip.vec <- minScore + (maxScore-minScore) * (score-min.sig) / (max.sig-min.sig)
	}
	return(list(sim.mat=sim.mat, ip.vec=ip.vec))
}

#' Read GMT File
#'
#' @inheritParams WebGestaltR
#' @param gmtFile The file path or URL of the GMT file.
#'
#' @return A data frame with columns of "geneSet", "description", "gene".
#'
#' @importFrom httr content
#' @importFrom tools file_ext
#' @export
#'
readGmt <- function(gmtFile, cache=NULL) {
#####Change a gmt file to a three column matrix (gene set name, gene set description and genes)#######
	if (startsWith(gmtFile, "http://") || startsWith(gmtFile, "https://")) {
		response <- cacheUrl(gmtFile, cache)
		if (response$status_code == 200) {
			data <- unlist(strsplit(content(response, "text"), "\n", fixed=TRUE))
		} else {
			stop(webRequestError(response))
		}
	} else {
		if (file_ext(gmtFile) != "gmt") {
			stop(gmtFormatError("empty"))
		}
		# remove BOM in some windows files
		data <- gsub("\xEF\xBB\xBF", "", readLines(gmtFile, skipNul=TRUE), useBytes=TRUE)
	}
	data <- strsplit(data, "\t", useBytes=TRUE)
	data <- lapply(data,.toList)
	data <- do.call("rbind",data)

	if (is.null(data)) {
		stop(gmtFormatError("incorrect"))
	} else {
		data <- as.data.frame(data, stringsAsFactors=FALSE)
		colnames(data) <- c("geneSet", "description", "gene")
		return(data)
	}
}
readGMT <- readGmt


.toList <- function(data) {
	if (length(data)>2) {
		data <- data[!is.na(data) & !is.null(data)]
		# replace % in gene set names to _, because png treats % in filename specially
		data1 <- cbind(rep(gsub('%', '_', data[1], fixed=TRUE), length(data)-2), rep(data[2], length(data)-2), data[c(-1,-2)])
		return(data1)
	} else {
		return(NULL)
	}
}

#' Prepare Input Matrix for GSEA
#'
#' @param rank A 2 column Data Frame of gene and score
#' @param gmt 3 column Data Frame of geneSet, description, and gene
#'
#' @return A matrix used for input to \code{swGsea}.
#'
#' @importFrom dplyr filter select distinct inner_join %>%
#' @export
#'
prepareInputMatrixGsea <- function(rank, gmt) {
	genes <- rank$gene
	gmt <- gmt %>% filter(.data$gene %in% genes)
	geneSets <- (gmt %>% select(.data$geneSet, .data$description) %>% distinct())$geneSet
	# 0 or 1 matrix indicating gene and gene set relationship
	rel <- fillInputDataFrame(gmt, genes, geneSets)
	# R implementation
	# rel <- matrix(0, nrow=length(genes), ncol=length(geneSets), dimnames=list(genes, geneSets))
	#
	# for (i in 1:nrow(gmt)) {
	# 	rel[gmt[i, "gene"], gmt[i, "geneSet"]] <- 1
	# }
	# rel <- as.data.frame(rel)
	# rel$gene <- genes
	return(inner_join(rank, rel, by="gene"))
}

readGMT <- function(...) {
	warning("Function readGMT is deprecated and changed to readGmt!\n")
	return(readGmt(...))
}

parameterErrorMessage <- function(hostName="https://www.webgestalt.org/", cache=NULL, ...){
	errorTest <- NULL
	argList <- list(...)

	argNames <- names(argList)
	# special cases, assuming minNum, maxNum appear together
	if("minNum" %in% argNames && "maxNum" %in% argNames){
		errorTest <- .minMaxNumError(minNum=argList$minNum,maxNum=argList$maxNum)
		if(!is.null(errorTest)){
			return(errorTest)
		}
		argList$minNum <- NULL
		argList$maxNum <- NULL
	}

	if("organism" %in% argNames){
		errorTest <- .organismError(organism=argList$organism, hostName=hostName, cache=cache)
		if(!is.null(errorTest)){
			return(errorTest)
		}
		argList$organism <- NULL
	}

	# individual checking function has name with format:
	# "." + parameterName + "Error"
	for(i in seq_along(argList)){
		errorTest <- do.call(paste0(".", names(argList[i]),"Error"), argList[i])
		if(!is.null(errorTest)){
			return(errorTest)
		}
	}
	return(errorTest)
}

.enrichMethodError <- function(enrichMethod){ #####Input method error
	existingMethods <- c("ORA", "GSEA", "NTA")

	if(!(enrichMethod %in% existingMethods)){
		error <- paste0("ERROR: The enrichment method '",enrichMethod, "' is not supported.")
		cat(error)
		return(error)
	}else{
		return(NULL)
	}
}

.organismError <- function(organism, hostName, cache) { ####Input organism error
	organisms <- listOrganism(hostName=hostName, cache=cache)
	organisms <- c(organisms, "others")
	if(!(organism %in% organisms)){
		error <- paste0("ERROR: The organism '",organism,"' is not supported.")
		cat(error)
		return(error)
	}else{
		return(NULL)
	}
}

.collapseMethodError <- function(collapseMethod){ ###Input collapseMethod error
	collapseMethodList <- c("mean","median","min","max")
	if(!(collapseMethod %in% collapseMethodList)){
		error <- "ERROR: WebGesalt only supports the following collapse methods: mean, median, min and max."
		cat(error)
		return(error)
	}else{
		return(NULL)
	}
}

.minMaxNumError <- function(minNum,maxNum){  #####Input the minimum and maximum number of category error
	if(!is.wholenumber(minNum) || minNum<1){
		error <- "ERROR: The minimum number of genes annotated to the category should be an positive integer."
		cat(error)
		return(error)
	}else{
		if(!is.wholenumber(maxNum) || maxNum<1){
			error <- "ERROR: The maximum number of genes annotated to the category should be an positive integer."
			cat(error)
			return(error)
		}else{
			if(minNum>=maxNum){
				error <- "ERROR: The minimum number of genes annotated to the category should be less than the maximum number."
				cat(error)
				return(error)
			}else{
				return(NULL)
			}
		}
	}
}

.fdrMethodError <- function(fdrMethod){   ##Input fdrMethod error
	fdrMethodList <- c("holm","hochberg","hommel","bonferroni","BH","BY")

	if(!(fdrMethod %in% fdrMethodList)){
		error <- "ERROR: WebGesalt only supports the following FDR methods: holm, hochberg, hommel, bonferroni, BH and BY."
		cat(error)
		return(error)
	}else{
		return(NULL)
	}
}

.sigMethodError <- function(sigMethod){   ##Input sig method error
	sigMethodList <- c("fdr","top")
	if(!(sigMethod %in% sigMethodList)){
		error <- "ERROR: WebGesalt only supports two methods to identify the enriched categories: 'fdr' method identifies the categories with FDR less than 'fdrThr' and 'top' method ranks all categories based on the FDR and selects the 'topThr' categories."
		cat(error)
		return(error)
	}else{
		return(NULL)
	}
}

.fdrThrError <- function(fdrThr){   ##Input fdr threshold error
	if(!is.numeric(fdrThr) || fdrThr<0 || fdrThr>1){
		error <- "ERROR: The FDR threshold should be a numberic from 0 to 1."
		cat(error)
		return(error)
	}else{
		return(NULL)
	}
}

.topThrError <- function(topThr){   ##Input top threshold error
	if(!is.wholenumber(topThr) || topThr<1){
		error <- "ERROR: The number of top categories should be a positive integer."
		cat(error)
		return(error)
	}else{
		return(NULL)
	}
}

.reportNumError <- function(reportNum){   ##Input the number of visualized enriched category error
	if (!is.wholenumber(reportNum) || reportNum<0) {
		error <- "ERROR: The number of enriched categories shown in the final report should be a positive integer."
		cat(error)
		return(error)
	}else{
		return(NULL)
	}
}

.perNumError <- function(perNum){   ##Input the number of permutation error
	if(!is.wholenumber(perNum) || perNum<0 || perNum>10000){
		error <- "ERROR: The number of permutation for GSEA method should be a positive integer and less than 10000."
		cat(error)
		return(error)
	}else{
		return(NULL)
	}
}

.isOutputError <- function(isOutput){  ##Input isOutput error
	if(!is.logical(isOutput)){
		error <- "ERROR: isOutput should be an R logical object (TRUE or FALSE)."
		cat(error)
		return(error)
	}else{
		return(NULL)
	}
}

.outputDirectoryError <- function(outputDirectory){   ##Input directory error
	if(!dir.exists(outputDirectory)){
		error <- paste("ERROR: The output directory ",outputDirectory," does not exist. please change another directory or create the directory.",sep="")
		cat(error)
		return(error)
	}else{
		return(NULL)
	}
}

.dagColorError <- function(dagColor){   ##Input dag color type error
	dagColorList <- c("binary","continuous")
	if(!(dagColor %in% dagColorList)){
		error <- "ERROR: dagColor should be binary or continuous."
		cat(error)
		return(error)
	}else{
		return(NULL)
	}
}

.hostNameError <- function(hostName){ ####Input hostName error
		if(grepl("^file://", hostName, perl=TRUE)){
			return(NULL)
		}
		archives <- listArchiveUrl()
		if(!(hostName %in% archives[,2])){
			error <- paste("ERROR: The host name ",hostName," is incorrect. Please use listArchiveURL function to find the correct host name.",sep="")
			cat(error)
			return(error)
		}else{
			return(NULL)
		}
}

.dataTypeError <- function(dataType){  ######Input data type error for ID Mapping function
	dataTypeA <- c("list","rnk","gmt")
	if(!(dataType %in% dataTypeA)){
		error <- paste("ERROR: Data type ",dataType," can not be supported. Please select from 'list', 'rnk' and 'gmt'.",sep="")
		cat(error)
		return(error)
	}else{
		return(NULL)
	}
}

.mappingOutputError <- function(mappingOutput){  ##Input mappingOutput error for ID Mapping function
	if(!is.logical(mappingOutput)){
		error <- "ERROR: mappingOutput should be an R logical object (TRUE or FALSE)."
		cat(error)
		return(error)
	}else{
		return(NULL)
	}
}

##############Other functions for the parameter error test############
testNull <- function(parameter){
	if(!is.null(parameter)){
		if(length(parameter)==1 && length(which(parameter=="NULL"))==1){
			parameter <- NULL
		}
	}
	return(parameter)
}

is.wholenumber <- function(x, tol = .Machine$double.eps^0.5){
	re <- tryCatch(abs(x - round(x)) < tol,warning=function(e){return(FALSE)},error=function(e){return(FALSE)})
	return(re)
}


##################Other errors###############

descriptionFileError <- function(type){
	if(type=="format"){
		error <- "ERROR: The description file for the functional categetories should have a 'des' extension."
	}

	if(type=="columnNum"){
		error <- "ERROR: The description file should contain two columns: the first column is the ID of the gene sets that should be the same with the uploaded enrichment gene sets, the second column is the description of the gene sets."
	}

	if(type=="overlap"){
		error <- "ERROR: The ID types of the uploaded functional database file and description file are different. Please check the uploaded files."
	}

	cat(error)
	return(error)
}

gmtFormatError <- function(type){
	if(type=="empty"){
		error <- "ERROR: Please upload a file with extension 'gmt'."
	}
	if(type=="incorrect"){
		error <- "ERROR: Invalid GMT file format. Please check the format of the GMT file from https://www.webgestalt.org/WebGestalt_2019_Manual.pdf."
	}
	cat(error)
	return(error)
}


enrichDatabaseError <- function(type,enrichDatabase="",organism=""){
	if(type=="unsupported"){
		error <- paste("ERROR: ",enrichDatabase," can not be supported for organism ",organism,".",sep="")
	}

	if(type=="empty"){
		error <- "ERROR: Please select the functional database or select 'others' to upload the functional database."
	}

	if(type=="others"){
		error <- "ERROR: Please upload a 'gmt' file as the functional database for 'others' organism."
	}

	cat(error)
	return(error)
}


idTypeError <- function(idType, organism, hostName, cache) {
		idTypes <- listIdType(organism=organism, hostName=hostName, cache=cache)
		if(!(idType %in% idTypes)){
				error <- paste("ERROR: The ID type ",idType," can not be supported for organism ",organism,".",sep="")
				cat(error)
				return(error)
		}else{
				return(NULL)
		}
}

targetIdTypeError <- function(idType, organism, hostName, cache) {
		idTypes <- listIdType(organism=organism, hostName=hostName, cache=cache)
		if(!(idType %in% idTypes)){
				error <- paste("ERROR: The target ID type ",idType," can not be supported for organism ",organism,".",sep="")
				cat(error)
				return(error)
		}else{
				return(NULL)
		}
}

stardardDiffError <- function(standardSource,standardTarget){
	if(standardSource!=standardTarget){
		error <- paste(standardSource," based ID type can not map to ",standardTarget," based ID type.",sep="")
		cat(error)
		return(error)
	}else{
		return(NULL)
	}

}

idMappingError <- function(type,idType="",topF=""){
	if(type=="unmapped"){
		error <- paste("ERROR: The ID type of the uploaded list is not consistent with the input ID type ",idType,". Examples of the input ID type: ",topF,".",sep="")
	}
	if(type=="empty"){
		error <- "ERROR: No IDs are mapped. Please check your input."
	}

	cat(error)
	return(error)
}

interestGeneError <- function(type){
	if(type=="empty"){
		error <- "ERROR: Please upload an interesting ID list."
	}
	if(type=="emptyType"){
		error <- "ERROR: Please select the ID type of the interesting ID list."
	}
	if(type=="unmatch"){
		error <- "The ID type of the input ID list can not match to the ID type of the functional database."
	}

	if(type=="unannotated"){
		error <- "ERROR: All IDs in the uploaded list are not annotated to any category of the functional database."
	}

	if(type=="onlyOne"){
		error <- "ERROR: All IDs in the uploaded list can only annotate to one category, which can not perform the enrichment analysis."
	}

	cat(error)
	return(error)
}

referenceGeneError <- function(type){
	if(type=="empty"){
		error <- "ERROR: Please upload a reference list or select an existing reference set."
	}
	if(type=="emptyType"){
		error <- "ERROR: Please select the ID type of the reference list."
	}
	if(type=="unmatch"){
		error <- "The ID type of the input reference list can not match to the ID type of the functional database."
	}
	if(type=="unmatch_set"){
		error <- "The ID type of the selected reference set can not match to the ID type of the functional database."
	}
	if(type=="existingRef"){
		error <- "ERROR: Please select an existing reference set."
	}
	if(type=="interestEmpty"){
		error <- "ERROR: All IDs in the interesting list are not included in the overlapping IDs between IDs in the reference list and IDs annotated to the functional database."
	}

	cat(error)
	return(error)
}

webRequestError <- function(response) {
	error <- paste(response["status_code"], content(response)["message"], "at", response["url"], "\n")
	cat(error)
	return(error)
}

webApiError <- function(response) {
	error <- response[["message"]]
	cat(paste0(error, "\n"))
	return(error)

}

.hasError <- function(obj){
	if(is.character(obj) && length(obj)==1 && length(grep("ERROR:",obj))>0) {
		return(TRUE)
	}
	return(FALSE)
}

#' Load gene set data
#'
#' @inheritParams WebGestaltR
#'
#' @return A list of \code{geneSet}, \code{geneSetDes}, \code{geneSetDag}, \code{geneSetNet}, \code{standardId}.
#' \describe{
#'  \item{geneSet}{Gene set: A data frame with columns of "geneSet", "description", "genes"}
#'  \item{geneSetDes}{Description: A data frame with columns of two columns of gene set ID and description}
#'  \item{geneSetDag}{DAG: A edge list data frame of two columns of parent and child. Or a list of data frames if multilple databases are given.}
#'  \item{geneSetNet}{Network: A edge list data frame of two columns connecting nodes. Or a list of data frames if multilple databases are given.}
#'  \item{standardId}{The standard ID of the gene set}
#' }
#'
#' @importFrom dplyr select distinct filter %>%
#' @importFrom httr modify_url
#' @export
loadGeneSet <- function(organism="hsapiens", enrichDatabase=NULL, enrichDatabaseFile=NULL, enrichDatabaseType=NULL, enrichDatabaseDescriptionFile=NULL, cache=NULL, hostName="http://www.webgestalt.org/") {
	# TODO: multiple custom database ID types?
	geneSet <- NULL    ##gene sets
	geneSetDes <- NULL ##gene set description file
	geneSetDag <- list() ##gene set DAG file
	geneSetNet <- list() ##gene set network file
	standardId <- NULL

	if (organism != "others" && !is.null(enrichDatabaseFile) && is.null(enrichDatabaseType)) {
		stop("The ID type should be given in enrichDatabaseType for custom GMT files, e.g. genesymbol.")
	}
	# necessary because loop with length used below. enrichDatabase will skip NULL
	if (!is.vector(enrichDatabaseFile)) {
		enrichDatabaseFile = list(enrichDatabaseFile)
	}
	if (!is.vector(enrichDatabaseDescriptionFile)) {
		enrichDatabaseDescriptionFile = list(enrichDatabaseDescriptionFile)
	}
	if (length(enrichDatabaseFile) != length(enrichDatabaseDescriptionFile)) {
		stop("The number of custom database and its description files should be equal. Use NULL for placeholder.")
	}
	if (organism != "others") {  # supported organism
		geneSetInfo <- listGeneSet(organism=organism, hostName=hostName, cache=cache)
		#  load build-in databases
		for (enrichDb in enrichDatabase) {
			if (is.null(enrichDb) || enrichDb == "others") { next }  # just for backward compatibility
			if (!enrichDb %in% geneSetInfo$name) {
				warning("Database ", enrichDb, " is not supported")
				next
			}
			# get the ID type of the enriched database, such as entrezgene or phosphsiteSeq
			thisStandardId <- filter(geneSetInfo, .data$name==enrichDb)[[1, "idType"]]
			if (!is.null(standardId) && standardId != thisStandardId) {
				stop("Databases have inconsistent ID types. Mixed gene annotation databases with phosphosite databases?")
			}
			standardId <- thisStandardId

			#########Read GMT file from the existing database###########
			if (startsWith(hostName, "file://")) {
				gmtPath <- removeFileProtocol(file.path(hostName, "geneset", paste0(paste(organism, enrichDb, standardId, sep="_"), ".gmt")))
				thisGeneSet <- readGmt(gmtPath)
				thisGeneSet$database <- enrichDb # add a column for database source
				geneSet <- rbind(geneSet, thisGeneSet)
			} else {
				gmtUrl <- modify_url(file.path(hostName, "api", "geneset"), query=list(organism=organism, database=enrichDb, standardId=standardId, fileType="gmt"))
				thisGeneSet <-  readGmt(gmtUrl, cache=cache)
				thisGeneSet$database <- enrichDb
				oriCnt <- nrow(thisGeneSet)
				thisGeneSet <- thisGeneSet %>% filter(!(.data$geneSet %in% unique(!!geneSet$geneSet)))
				if (nrow(thisGeneSet) < oriCnt) {
				  warning(paste("Duplicate gene set names in", enrichDb, "have been ignored."))
				}
				geneSet <- rbind(geneSet, thisGeneSet)
			}

			#########Read the description file#############
			#geneSetDes <- rbind(geneSetDes, .loadGeneSetData(hostName, organism, enrichDb, standardId, "des", cache))
			thisGeneSetDes <- .loadGeneSetData(hostName, organism, enrichDb, standardId, "des", cache)
			if (!is.null(thisGeneSetDes) && !is.null(geneSetDes)) {
			  thisGeneSetDes <- thisGeneSetDes %>% filter(!(.data$geneSet %in% unique(!!geneSetDes$geneSet)))
			}
			geneSetDes <- rbind(geneSetDes, thisGeneSetDes)

			###########Try to load the DAG file#################
			# assignment considering possible return of NULL
			# list[] <- NULL will delete the element
			geneSetDag[enrichDb] <- list(.loadGeneSetData(hostName, organism, enrichDb, standardId, "dag", cache))

			###########Try to load the network file if the gene sets are generated from the network##########
			geneSetNet[enrichDb] <- list(.loadGeneSetData(hostName, organism, enrichDb, standardId, "net", cache))
		}

		# load local database files
		for (i in 1:length(enrichDatabaseFile)) {
			enrichDbFile <- enrichDatabaseFile[[i]]
			if (is.null(enrichDbFile)) { next }

			thisGeneSet <- idMapping(organism=organism, dataType="gmt", inputGeneFile=enrichDbFile, sourceIdType=enrichDatabaseType, targetIdType=NULL, mappingOutput=FALSE, cache=cache, hostName=hostName)
			thisStandardId <- thisGeneSet$standardId  # should be just enrichDatabaseType here
			if (!is.null(standardId) && standardId != thisStandardId) {
				stop("Databases have inconsistent ID types. Mixed gene annotation databases with phosphosite databases?")
			}
			standardId <- thisStandardId

			thisGeneSet <- thisGeneSet$mapped %>% select(.data$geneSet, .data$description, gene=standardId) %>% distinct()

			# load local description files
			enrichDbDesFile <- enrichDatabaseDescriptionFile[[i]]
			if (!is.null(enrichDbDesFile)) {
				thisGeneSetDes <- .loadEnrichDatabaseDescriptionFile(thisGeneSet, enrichDbDesFile)
				if (!is.null(thisGeneSetDes) && !is.null(geneSetDes)) {
				  thisGeneSetDes <- thisGeneSetDes %>% filter(!(.data$geneSet %in% unique(!!geneSetDes$geneSet)))
				}
				geneSetDes <- rbind(geneSetDes, thisGeneSetDes)
			}

			fileName <- gsub(".gmt", "", basename(enrichDbFile), fixed=TRUE)
			thisGeneSet$database <- fileName
			oriCnt <- nrow(thisGeneSet)
			thisGeneSet <- thisGeneSet %>% filter(!(.data$geneSet %in% unique(!!geneSet$geneSet)))
			if (nrow(thisGeneSet) < oriCnt) {
			  warning(paste("Duplicate gene set names in", fileName, "have been ignored."))
			}
			geneSet <- rbind(geneSet, thisGeneSet)

			geneSetDag[fileName] <- list(NULL) # correct way to assign NULL to list element
			geneSetNet[fileName] <- list(NULL)
		}
	} else { # custom organisms
		for (i in 1:length(enrichDatabaseFile)) {
			enrichDbFile <- enrichDatabaseFile[[i]]
			if (is.null(enrichDbFile)) { next }
			thisGeneSet <- readGmt(enrichDbFile)

			enrichDbDesFile <- enrichDatabaseDescriptionFile[[i]]
			if (!is.null(enrichDbDesFile)) {
				thisGeneSetDes <- .loadEnrichDatabaseDescriptionFile(thisGeneSet, enrichDbDesFile)
				if (!is.null(thisGeneSetDes) && !is.null(geneSetDes)) {
				  thisGeneSetDes <- thisGeneSetDes %>% filter(!(.data$geneSet %in% unique(!!geneSetDes$geneSet)))
				}
				geneSetDes <- rbind(geneSetDes, thisGeneSetDes)
			}
			fileName <- gsub(".gmt", "", basename(enrichDbFile), fixed=TRUE)
			thisGeneSet$database <- fileName
			oriCnt <- nrow(thisGeneSet)
			thisGeneSet <- thisGeneSet %>% filter(!(.data$geneSet %in% unique(!!geneSet$geneSet)))
			if (nrow(thisGeneSet) < oriCnt) {
				warning(paste("Duplicate gene set names in", enrichDb, "have been ignored."))
			}
			geneSet <- rbind(geneSet, thisGeneSet)

			geneSetDag[fileName] <- list(NULL)
			geneSetNet[fileName] <- list(NULL)
		}
	}
	if (is.null(geneSet)) { stop(enrichDatabaseError(type="empty")) }

	# unlist if just one database, for backward compatibility
	if (length(geneSetDag) == 1) { geneSetDag <- geneSetDag[[1]] }
	if (length(geneSetNet) == 1) { geneSetNet <- geneSetNet[[1]] }

	if (length(unique(geneSet$database)) == 1) {
		# remove database column for single source
		geneSet$database<- NULL
	}
	re <- list(geneSet=geneSet, geneSetDes=geneSetDes, geneSetDag=geneSetDag, geneSetNet=geneSetNet,standardId=standardId)
	return(re)
}

#' @importFrom httr content
#' @importFrom readr read_tsv
.loadGeneSetData <- function(hostName, organism, database, standardId, fileType, cache=NULL) {
	# read gene set files from API or returns NULL
	if (startsWith(hostName, "file://")) {
		geneSetPath <- removeFileProtocol(file.path(hostName, "geneset", paste(paste(organism, database, standardId, sep="_"), fileType, sep=".")))
		if (file.exists(geneSetPath)) {
			geneSetData <- read_tsv(geneSetPath, col_names=FALSE, col_types="cc", quote="")
		} else {
			geneSetData <- NULL
		}
	} else {
		geneSetUrl <- file.path(hostName,"api","geneset")
		response <- cacheUrl(geneSetUrl, cache=cache, query=list(organism=organism, database=database, standardId=standardId, fileType=fileType))
		if (response$status_code == 200) {
			geneSetData <- read_tsv(content(response), col_names=FALSE, col_types="cc", quote="")
		} else {
			geneSetData <- NULL
		}
	}
	if (!is.null(geneSetData) && fileType == "des") {
		colnames(geneSetData) <- c("geneSet", "description")
		geneSetData <- geneSetData %>% distinct(.data$geneSet, .keep_all=TRUE)
	}
	return(geneSetData)
}

#' @importFrom readr read_tsv
#' @importFrom tools file_ext
.loadEnrichDatabaseDescriptionFile <- function(geneSet, enrichDatabaseDescriptionFile) {
	if (file_ext(enrichDatabaseDescriptionFile) != "des") {
		warning(descriptionFileError("format"))
		return(NULL)
	} else {
		geneSetDes <- read_tsv(enrichDatabaseDescriptionFile, col_names=c("geneSet", "description"), col_types="cc", quote="") %>%
			distinct(geneSet, .keep_all=TRUE)
		if (ncol(geneSetDes)!=2) {
			warning(descriptionFileError("columnNum"))
			return(NULL)
		} else {
			if (length(intersect(unique(geneSet$geneSet), geneSetDes$geneSet)) < 0.6 * length(unique(geneSet[,1]))) {
				warning(descriptionFileError("overlap"))
				return(NULL)
			} else {
				return(geneSetDes)
			}
		}
	}
}

#' List Reference Sets
#'
#' List available reference sets for the given organism on WebGestalt server.
#'
#' @inheritParams WebGestaltR
#'
#' @return A list of reference sets.
#'
#' @importFrom httr content
#' @export
#'
listReferenceSet <- function(organism="hsapiens", hostName="http://www.webgestalt.org/", cache=NULL) {
	if (startsWith(hostName, "file://")) {
		jsonData <- fromJSON(removeFileProtocol(file.path(hostName, "referencesetsummary.json")))
		idType <- jsonData[[organism]]
		idType <- idType$name
	} else {
		response <- cacheUrl(file.path(hostName, "api", "summary", "referenceset"), cache)
		if (response$status_code != 200) {
			return(webRequestError(response))
		}
		jsonData <- content(response)
		idType <- jsonData[[organism]]
		idType <- sapply(idType,function(e){return(e$name)})
	}
	return(idType)
}

#' Expand enriched GO IDs to include ancestors up to the root
#'
#' Returns expanded nodes and DAG tree edges
#'
#' @importFrom dplyr filter
#' @keywords internal
#'
expandDag <- function(goTermList, dagEdgeList) {
	colnames(dagEdgeList) <- c("source", "target")
	goTermList <- goTermList[goTermList %in% unique(unlist(dagEdgeList, use.names=FALSE))]
	queue <- as.list(goTermList)
	## expand to include all linked nodes
	edges <- list()
	while (length(queue) > 0 ) {
		goTerm <- queue[[1]]
		if (length(queue) == 1) {
			queue <- list()
		} else {
			queue <- queue[2:length(queue)]
		}
		inEdges <- filter(dagEdgeList, .data$target == goTerm)
		if (nrow(inEdges) > 0) {
			edges <- c(edges, unname(lapply(split(inEdges, seq(nrow(inEdges))), function(x) {
				return(list(data=as.list(x)))
			})))
		}
		for (parentNode in inEdges$source) {
			if (!parentNode %in% goTermList) {
				queue <- c(queue, parentNode)
				goTermList <- c(goTermList, parentNode)
			}
		}
	}

	return(list(allNodes=goTermList, edges=edges))
}

#' @importFrom dplyr filter
getDagNodes <- function(enrichedRes, allGoList, goIdName, enrichMethod, dagColorSchema) {
	if (!is.null(goIdName)) {
		colnames(goIdName) <- c("id", "name")
	}
	palette <- getColorPalette(enrichedRes, enrichMethod, dagColorSchema)
	return(unname(lapply(allGoList, function(x) {
		if (!is.null(goIdName) && x %in% goIdName$id) {
			goName <- filter(goIdName, .data$id == x)[[1, "name"]]
		} else {
			goName <- ""
		}
		color <- palette(x)
		return(list(
			data=list(
				id=x,
				name=goName,
				color=color
			)
		))
	})))
}


colorPos <- "steelblue"
colorNeg <- "darkorange"
colorNeutral <- "white"

#' @importFrom dplyr filter
getColorPalette <- function(enrichedRes, enrichMethod, schema) {
	if (schema == "binary") {
		if (enrichMethod == "ORA") {
			return(function(goTerm) {
				return(ifelse(goTerm %in% enrichedRes$geneSet, colorPos, colorNeutral))
			})
		} else if (enrichMethod == "GSEA") {
			return(function(goTerm) {
				nes <- filter(enrichedRes, .data$geneSet == goTerm)[["normalizedEnrichmentScore"]]
				if (length(nes) == 0) {
					return(colorNeutral)
				} else {
					return(ifelse(nes > 0, colorPos, colorNeg))
				}
			})
		}
	} else if (schema == "continuous") {
		if (enrichMethod=="ORA") {
			minFdr <- min(enrichedRes$FDR)
			minFdrLog <- ifelse(minFdr==0, -log10(.Machine$double.eps), -log10(minFdr))
			colorPalette <- colorRampPalette(c(colorNeutral, colorPos))(128)
			myBreak <- seq(0, minFdrLog + 0.01, length.out=129)

			return(function(goTerm) {
				fdr <- filter(enrichedRes, .data$geneSet == goTerm)[["FDR"]]
				if (length(fdr) == 0) {
					return(colorNeutral)
				} else {
					fdrLog <- ifelse(fdr == 0, -log10(.Machine$double.eps), -log10(fdr))
					return(colorPalette[max(which(myBreak <= fdrLog))])
				}
			})
		} else if (enrichMethod=="GSEA") {
			fdr <- enrichedRes$FDR
			fdr[fdr == 0] <- .Machine$double.eps
			fdr <- sign(enrichedRes$normalizedEnrichmentScore) * (-log10(fdr))
			minFdrLog <- min(fdr)
			maxFdrLog <- max(fdr)
			tmp <- getPaletteForGsea(maxFdrLog, minFdrLog)
			colorPalette <- tmp[[1]]
			myBreak <- tmp[[2]]
			return(function(goTerm) {
				row <- filter(enrichedRes, .data$geneSet == goTerm)
				if (nrow(row) == 0) {
					return(colorNeutral)
				} else {
					fdr <- row[["FDR"]]
					nes <- row[["normalizedEnrichmentScore"]]
					fdrLog <- ifelse(fdr == 0, sign(nes) * (-log10(.Machine$double.eps)), sign(nes) * (-log10(fdr)))
					return(colorPalette[max(which(myBreak <= fdrLog))])
				}
			})
		}
	}
}

getPaletteForGsea <- function(maxScore, minScore) {
	if (minScore > 0) {
		colorPalette <- colorRampPalette(c(colorNeutral, colorPos))(128)
		myBreak <- seq(0, maxScore + 0.01, length.out=129)
	}else{
		if (maxScore < 0) {
			colorPalette <- colorRampPalette(c(colorNeg, colorNeutral))(128)
			myBreak <- seq(minScore- 0.01, 0, length.out=129)
		}else{
			if (abs(minScore) > maxScore) {
				colorPalette <- colorRampPalette(c(colorNeg, colorNeutral, colorPos))(256)
				myBreak <- c(seq(minScore-0.01, -0.01, length.out=128), 0, seq(0.01, -minScore+0.01, length.out=128))
			}else{
				colorPalette <- colorRampPalette(c(colorNeg, colorNeutral, colorPos))(256)
				myBreak <- c(seq(-maxScore-0.01, -0.01, length.out=128), 0, seq(0.01, maxScore+0.01, length.out=128))
			}
		}
	}
	return(list(colorPalette, myBreak))
}

#' ID Mapping
#'
#' ID mapping utility with WebGestalt server.
#'
#' @inheritParams WebGestaltR
#' @param dataType Type of data, either \code{list}, \code{rnk} or \code{gmt}.
#'   Could be \code{list}, \code{rnk} or \code{matrix} for \code{idToSymbol}.
#' @param inputGeneFile The data file to be mapped.
#' @param inputGene Or the input could be given as an R object.
#'   GMT file should be read with \code{readGmt}.
#' @param sourceIdType The ID type of the data.
#' @param targetIdType The ID type of the mapped data.
#' @param mappingOutput Boolean if the mapping output is written to file.
#' @param outputFileName The output file name.
#'
#' @return A list of \code{mapped} and \code{unmapped} IDs.
#' @export
#' @aliases IDMapping
#'
idMapping <- function(organism="hsapiens", dataType="list", inputGeneFile=NULL, inputGene=NULL, sourceIdType, targetIdType=NULL, collapseMethod="mean", mappingOutput=FALSE, outputFileName="", cache=NULL, hostName="http://www.webgestalt.org/") {
	#############Check general parameters########
	errorTest <- parameterErrorMessage(organism=organism, dataType=dataType, collapseMethod=collapseMethod, hostName=hostName, mappingOutput=mappingOutput, cache=cache)

	if(!is.null(errorTest)){
		stop(errorTest)
	}

	############Check source id type#########
	errorTest <- idTypeError(idType=sourceIdType, organism=organism, hostName=hostName, cache=cache)
	if(!is.null(errorTest)){
		stop(errorTest)
	}

	##########Identify the standardId for the input ID type###########
	standardSource <- identifyStandardId(hostName=hostName, idType=sourceIdType, organism=organism, type="interest", cache=cache)

	############Check target id type#########
	if(!is.null(targetIdType)){
		errorTest <- targetIdTypeError(idType=targetIdType, organism=organism, hostName=hostName, cache=cache)
		if(!is.null(errorTest)){
			stop(errorTest)
		}else{
			standardTarget <- identifyStandardId(hostName=hostName, idType=targetIdType, organism=organism, type="interest", cache=cache)
			errorTest <- stardardDiffError(standardSource=standardSource,standardTarget=standardTarget)
			if(!is.null(errorTest)){
				stop(errorTest)
			}
		}
	}else{
		targetIdType <- standardSource
	}

	##########gene level ID Mapping##########
	if(standardSource=="entrezgene"){
		idMap <- idMappingGene(organism=organism, dataType=dataType, inputGeneFile=inputGeneFile, inputGene=inputGene, sourceIdType=sourceIdType, targetIdType=targetIdType, collapseMethod=collapseMethod, mappingOutput=mappingOutput, outputFileName=outputFileName, hostName=hostName)
	} else {
		idMap <- idMappingPhosphosite(organism=organism, dataType=dataType, inputGeneFile=inputGeneFile, inputGene=inputGene, sourceIdType=sourceIdType, targetIdType=targetIdType, collapseMethod=collapseMethod, mappingOutput=mappingOutput, outputFileName=outputFileName, hostName=hostName)
	}


	idMap$standardId <- standardSource
	return(idMap)
}

#' @export
IDMapping <- function(...) {
	warning("Function IDMapping is deprecated and changed to idMapping!\n")
	return(idMapping(...))
}

#' Site Weighted Gene Set Enrichment Analysis
#'
#' Performs site weighted gene set enrichment analysis or standard GSEA when
#' likelihood/weight columns in \code{input_df} are 1 or 0, \code{p=1},
#' \code{q=1} and \code{thresh_type="val"}.
#'
#' The formula for weighting is as follows
#' \deqn{\frac{s_{j}^{q}|r_{j}|^{p}}{\sum s^{q}|r|^{p}}}{(s_j^q * abs(r_j)^p) / (\sum s^q * abs(r)^p)}
#' Where r is log ratio score, s is likelihood score, j is the index of the gene.
#'
#' @param input_df A data frame in which first column is name of item of interest
#' (gene, protein, phosphosite, etc.), the second is the correlation of that item
#' of interest with the phenotype (typically log ratio of expression for phenotype
#' vs. normal), and the remaining columns are the scores for the likelihood that
#' the item belongs in each set (one column per set).
#' @param thresh_type The type of \code{thresh}. Use 'percentile' to include all
#' scores over that percentile given in \code{thresh} (i.e., 0.9 would be all items
#' in 90th percentile, or top 10 percent); 'list' to include a list of set lists
#' where the set lists are in the same order as the corresponding set columns in
#' the \code{input_df}; 'val' to apply a single threshold value to all sets; or
#' 'values' to use a vector of unique cutoffs for each set (needs to be in the
#' same order as the sets are specified in the columns of \code{input_df}")
#' @param thresh Depends on \code{thresh_type}. A list of lists of the items in
#' each set (with same names as colnames of the scores); a numeric vector of
#' threshold scores for each set (in the same order as the colnames of the scores
#' in the input_df), or a single percentile value between 0 and 1 (i.e., if
#' \code{thresh}=0.9, the 90th percentile of the score or the highest scoring 10%
#' of of the items are included in the set for each scoring regimen) (\code{thresh}
#' ="all" is not supported at this time, as it doesn't result in a Kolgorov-Smirnoff
#' statistic; this may be worked in as an alternate scoring method later on).
#' @param thresh_action Either "include", "exclude (default)", or "adjust";
#' this specifies how to treat each set if it doesn't contain a minimum number of
#' items or contains all of the items; this option cannot be used with predefined
#' lists of items in sets (if the number of items in a given set doesn't meet
#' requirements, that set will be skipped).
#' @param min_set_size,max_set_size The minimum/maximum number of items each
#' set needs for the analysis to proceed.
#' @param max_score,min_score A optional numeric vector of minimum/maximum boundaries
#' to clip scores for each set.
#' @param psuedocount Psuedocount (pc) is used for rescaling set scores:
#' \code{(score - min_score + pc)/(max_score - min_score +pc)}; this is needed to
#' prevent division by 0 if \code{max_score==min_score} (in this case, all scores
#' for items in set will be 1, which is equivalent to standard GSEA); it also allows
#' users to adjust weights for scores that are close to the minimum for the scores in
#' the set (unless min_score==max_score): as psuedocount value approaches 0, scaled
#' minimum scores also approach 0; as psuedocount approaches infinity, scaled minimum
#' scores approach the scaled maximum scores (which equal 1); this value must be
#' larger than 0.
#' @param perms The number of permutations.
#' @param p The exponential scaling factor of the phenotype score (second column in
#' \code{input_df}).
#' @param q The exponential scaling factor of the likelihood score (weights).
#' @param nThreads The number of threads to use in calculating permutaions.
#' @param rng_seed Random seed.
#' @param fork A boolean. Whether pass "fork" to \code{type} parameter of
#' \code{makeCluster} on Unix-like machines.
#'
#' @return A list of \code{Enrichment_Results}, \code{Items_in_Set} and \code{Running_Sums}.
#' \describe{
#'  \item{Enrichment_Results}{A data frame with row names of gene set and columns of
#'  "ES", "NES", "p_val", "fdr".}
#'  \item{Items_in_Set}{A list of one-column data frames. Describes genes and their
#'  ranks in each set.}
#'  \item{Running_Sums}{Running sum scores along genes sorted by ranked scores,
#'  with gene sets as columns.}
#' }
#'
#' @importFrom stats quantile
#' @importFrom dplyr arrange desc
#' @importFrom parallel makeCluster stopCluster
#' @importFrom doParallel registerDoParallel
#' @importFrom foreach foreach
#' @importFrom doRNG %dorng%
#' @export
#' @author Eric Jaehnig
#'
swGsea <- function(input_df, thresh_type="percentile", thresh=0.9, thresh_action="exclude", min_set_size=10, max_set_size=500, max_score="max", min_score="min", psuedocount=0.001, perms=1000, p=1, q=1, nThreads=1, rng_seed=1, fork=FALSE){
	# check input parameters
	if(thresh_type!="percentile"&thresh_type!="list"&thresh_type!="val"&thresh_type!="values"){
		stop("invalid thresh_type specified; needs to be set to 'percentile' to include all scores over that percentile (i.e., 0.9 would be all items in 90th percentile, or top 10 percent) 'list' to include a list of set lists where the set lists are in the same order as the corresponding set columns in the input_df, 'val' to apply a single threshold value to all sets, or 'values' to use a vector of unique cutoffs for each set (needs to be in the same order as the sets are specified in the columns of input_df")
	}
	if(thresh_action!="exclude"&thresh_action!="include"&thresh_action!="adjust"){
		stop("invalid thresh_action specified; needs to be set to 'exclude' to skip set if it contains no items after applying score threshold (or contains all items), or 'include' to include values for the set at the end of the results (ES and NES automatically set to 0 and pval to 1) or 'adjust' to adjust threshold to add at least min_set_size items below thresh to set (or remove all items equal to the minimum set score value from the set)")
	}
	if(min_set_size<3){
		stop("please set 'min_set_size' to 3 or greater (default is 5)")
	}
	if(max_score!="max"){
		if(length(max_score)!=(ncol(input_df)-2)|(!is.numeric(max_score))){
			stop("max_score needs to be set to max or contain a numeric vector of maximum scores for each set")
		}
	}
	if(min_score!="min"){
		if(length(min_score)==(ncol(input_df)-2)|(!is.numeric(min_score))){
			stop("min_score needs to be set to min or contain a numeric vector of minimum scores for each set")
		}
	}
	if(!psuedocount>0){
		stop("psuedocount must be greater than 0")
	}
	pc <- psuedocount

	# re-order by log ratios (rank order from highest to lowest)
	expt <- colnames(input_df)[2]
	enr_test <- colnames(input_df)[3:ncol(input_df)]
	colnames(input_df)[c(1,2)] <- c("item", "expression_val")
	input_df <- arrange(input_df, desc(.data$expression_val))

	# get and check size of set items; build in-set matrix of 1's for items in set and 0's for items not in set
	inset_mat <- matrix(0, nrow = length(input_df$item), ncol = length(enr_test))
	dimnames(inset_mat) <- list(input_df$item, enr_test)
	# if list of items provided for each set, check to make sure each item is in dataset and set inset_mat to 1 if it is and 0 if not
	if(thresh_type=="list"&is.list(thresh)){
		thresh_action <- "exclude"
		for(a in 1:length(thresh)){
			skip = ""
			if(length(setdiff(thresh[[a]], input_df$item)) > 0){
				skip <- setdiff(thresh[[a]], input_df$item)
				warning(paste0(paste(skip, collapse = ", "), " are not items in the input dataframe and will be skipped.\n"))
			}
			for(b in 1:length(thresh[[a]])){
				if(!thresh[[a]][b] %in% skip){
					inset_mat[thresh[[a]][b], a] <- 1
				}
			}
		}
		# if numeric threshold provided, set inset_mat to 1 for items that meet threshold or 0 for those that don't
	} else if(is.numeric(thresh)){
		if(thresh_type=="values" & length(thresh)==length(enr_test)){
			for(a in 1:length(thresh)){
				items_in_set <- input_df$item[input_df[ , enr_test[a]]>=thresh[a]]
				inset_mat[items_in_set, enr_test[a]] <- 1
			}
		} else if(thresh_type=="percentile" & thresh>0 & thresh<1){
			thresh1 <- vector(mode = "numeric", length = length(enr_test))
			for(a in 1:length(enr_test)){
				thresh1[a] = quantile(input_df[ , enr_test[a]], probs = thresh)
				items_in_set <- input_df$item[input_df[ , enr_test[a]]>=thresh1[a]]
				inset_mat[items_in_set, enr_test[a]] <- 1
			}
			thresh <- thresh1
		} else if(thresh_type=="val" & length(thresh)==1){
			for(a in 1:ncol(inset_mat)){
				items_in_set <- input_df$item[input_df[ , enr_test[a]]>=thresh]
				inset_mat[items_in_set, enr_test[a]] <- 1
			}
			thresh <- rep(thresh, times=ncol(inset_mat))
		}
	} else { stop("improper threshold specified") }

	# check to make sure minimum number of items present in each set; if not, skip or adjust according to thresh_action
	skipped_sets <- character(0)
	for(c in 1:ncol(inset_mat)){
		if(sum(inset_mat[ , c]) < min_set_size){
			too_small <- colnames(inset_mat)[c]
			warning(paste0(expt, " does not contain minimum number of items in set for ", too_small, "\n"))
			if(thresh_action=="exclude"|thresh_action=="include"){
				skipped_sets <- c(skipped_sets, too_small)
			} else {
				# lower threshold to first value that would include minimum number of items in set
				lower_scores <- sort(input_df[input_df[too_small]<thresh, too_small], decreasing = T)
				new_thresh <- lower_scores[min_set_size-sum(inset_mat[ , c])]
				# set all inset_mat values that meet the lower threshold to 1
				inset_mat[inset_mat[ , c]>=new_thresh, c] <- 1
				# check to make sure the set meets size requirements after applying new threshold, if not exclude set
				if((sum(inset_mat[ , c])<min_set_size)|(sum(inset_mat[ , c])==length(input_df$item))){
					skipped_sets <- c(skipped_sets, too_small)
					warning(paste0("cannot adjust threshold for ", too_small, " to meet size requirements for analysis in ", expt), "; skipping set\n")
				}
			}
		}
	}

	# now check to make sure each set contains fewer items than the max_set_size threshold (default=500); adjust score threshold to obtain valid set size or skip according to thresh_action
	for(c in 1:ncol(inset_mat)){
		check_col <- colnames(inset_mat)[c]
		if(sum(inset_mat[ , c]) > max_set_size){
			warning(paste0(expt, " has more than ", max_set_size, " items in set ", check_col, "\n"))
			if(thresh_action=="exclude"|thresh_action=="include"){
				skipped_sets <- c(skipped_sets, check_col)
			} else {
				# set items with minimum value to 0 to reduce set size (note: while loop not necessary here because max set size is all items in the dataset, but may be useful if we decide to impose a maximum set size later)
				while(sum(inset_mat[ , c]) > max_set_size){
					inset_mat[input_df$item[input_df[, check_col] == min(input_df[inset_mat[ , c]==1, check_col])], check_col] <- 0
				}
				# make sure this adjustment didn't remove too many items
				if(sum(inset_mat[ , c])<min_set_size){
					skipped_sets <- c(skipped_sets, check_col)
					warning(paste0("cannot adjust threshold for ", check_col, " to meet size requirements for analysis in ", expt), "; skipping set\n")
				}
			}
		}
	}

	# remove skipped columns
	# print(c("sets that don't contain proper number of items:", skipped_sets))
	inset_mat <- inset_mat[ , !(colnames(inset_mat)%in%skipped_sets)]

	if (is.null(ncol(inset_mat))) {
		stop("All gene sets are skipped! Please try to descrease the minimum set size.\n")
	}

	# generate list containing names of items in each set and ranks of those items
	items_in_set <- list()
	for(it in 1:ncol(inset_mat)){
		items_in_set[[colnames(inset_mat)[it]]] <- data.frame(which(inset_mat[ , it]==1), stringsAsFactors = F)
		colnames(items_in_set[[colnames(inset_mat)[it]]]) = "rank"
	}

	# adjust set scores to range from a minimum of 0+pc to a maximum of 1 (s=(Score-minScore+pc)/(maxScore-minScore+pc)); calculate s^q * |r|^p score for each item in each set; get sum of these scores for each set
	if(max_score=="max"){
		max_scores <- vector(mode = "numeric", length = ncol(inset_mat))
	} else { max_scores <- max_score }
	if(min_score=="min"){
		min_scores <- vector(mode = "numeric", length = ncol(inset_mat))
	} else { min_scores <- min_score }
	scaled_scores <- inset_mat
	set_scores <- inset_mat * input_df[ , colnames(inset_mat)]
	set_tot <- vector(mode = "numeric", length = ncol(inset_mat))
	adj_expr_val <- (abs(input_df$expression_val))^p
	for(d in 1:ncol(set_scores)){
		if(max_score=="max"){
			max_scores[d] <- max(set_scores[ , d])
		}
		if(min_score=="min"){
			min_scores[d] <- min(set_scores[set_scores[ , d]>0, d])
		}
		set_scores[set_scores[ , d] > max_scores[d], d] <- max_scores[d]
		set_scores[set_scores[ , d] < min_scores[d], d] <- min_scores[d]

		# if the max set score equals the minimum set score, all scores for items in set will be 1; therefore, let user know that this analysis now reverts to standard GSEA
		# if(max_scores[d]==min_scores[d]){
		# 	print(paste0("max score for set ", colnames(inset_mat)[d], " equals min score; results will be equivalent to standard GSEA"))
		# }

		# scales scores for sets
		scaled_scores[ , d] <- ((set_scores[ , d] - min_scores[d] + pc)/(max_scores[d] - min_scores[d] + pc))^q
		# multiple by inset_mat to set scores for items not in set to 0
		scaled_scores[ , d] <- scaled_scores[ ,d ] * inset_mat[ , d]
		set_scores[ , d] <- scaled_scores[ , d] * adj_expr_val
		if (sum(set_scores[ , d]) == 0) {
			set_scores[ , d] <- scaled_scores[ , d]
		}
		set_tot[d] <- sum(set_scores[ , d])
	}

	# get set of items not in in_set
	outset_mat <- 1 - inset_mat

	# convert to vectors to matrices with same dimensions as inset_mat to use to calculate Running_Sum and for permutations below
	# expr_mat <- matrix(rep(input_df$expression_val, times = ncol(inset_mat)), nrow = nrow(input_df), dimnames = list(rownames(inset_mat), colnames(inset_mat)))
	set_tot <- t(matrix(rep(set_tot, times=nrow(inset_mat)), nrow=length(set_tot), dimnames=list(colnames(inset_mat), rownames(inset_mat))))
	outset_mat_sums <- t(matrix(rep(colSums(outset_mat), times=nrow(outset_mat)), nrow = ncol(outset_mat), dimnames = list(colnames(outset_mat), rownames(outset_mat))))
	outset_scores <- outset_mat/outset_mat_sums
	scores_mat <- (set_scores / set_tot) - outset_scores

	# walk through ranked list and tally running total; also keep track of maximum and minimum values
	Running_Sum <- matrix(0, ncol=ncol(scores_mat), nrow = nrow(scores_mat), dimnames = dimnames(scores_mat))
	running_max <- vector(mode = "numeric", length = ncol(scores_mat))
	running_min <- vector(mode = "numeric", length = ncol(scores_mat))
	for(e in 1:ncol(Running_Sum)){
		Running_Sum[ , e] <- cumsum(scores_mat[ , e])
		running_max[e] <- max(Running_Sum[ , e])
		running_min[e] <- min(Running_Sum[ , e])
	}

	# permute df 1000x and use to determine p-value of max/min score from previous section
	set.seed(rng_seed)
	# FORK is an option for Unix machines to reduce memory footprint
	if(fork==T) {
		cl <- makeCluster(nThreads, type = "FORK")
	} else { cl <- makeCluster(nThreads) }
	registerDoParallel(cl)
	# use dorng instead of dopar to properly pass rng seed to foreach loop
	rand_stats <- foreach(i=1:perms, .combine = 'rbind') %dorng% {
		gseaPermutation(scaled_scores, outset_scores, adj_expr_val)

		## R implementation of permutations
		#
		# rand_df <- input_df[sample(nrow(input_df)), , drop = F]
		# rand_df$expression_val <- input_df$expression_val
		# rand_df <- rand_df[ , c("item","expression_val", colnames(inset_mat)), drop = F]
		# rand_scaled_scores <- scaled_scores[rand_df$item, , drop=F]
		# rand_outset_scores <- outset_scores[rand_df$item, , drop=F]
		#
		# rand_adj_scores <- rand_scaled_scores * (abs(expr_mat)^p)
		#
		# rand_set_tot <- colSums(rand_adj_scores)
		# rand_set_tot <- t(matrix(rep(rand_set_tot, times = nrow(rand_adj_scores)), nrow = length(rand_set_tot), dimnames = list(colnames(rand_adj_scores), rownames(rand_adj_scores))))
		# rand_scores <- (rand_adj_scores / (rand_set_tot + 0.000001)) - rand_outset_scores
		#
		# rand_tot <- matrix(0, nrow = nrow(rand_scores), ncol = ncol(rand_scores))
		# rand_max <- numeric(ncol(inset_mat))
		# rand_min <- numeric(ncol(inset_mat))
		# rand_best <- numeric(ncol(inset_mat))
		# for(j in 1:ncol(rand_tot)){
		# 	rand_tot[ , j] <- cumsum(rand_scores[ , j])
		# 	rand_max[j] <- max(rand_tot[ , j])
		# 	rand_min[j] <- min(rand_tot[ , j])
		# 	if(rand_max[j] >= abs(rand_min[j])){
		# 		rand_best[j] <- rand_max[j]
		# 	} else { rand_best[j] <- rand_min[j] }
		# }
		# c(rand_min, rand_max, rand_best)
	}
	stopCluster(cl)
	cat(paste0(perms, " permutations of ", expt, " complete...\n"))

	# split output from permutations into iteration by set dataframes for random running totals, random maxes, and random mins
	rand_mins <- rand_stats[ , 1:ncol(inset_mat)]
	colnames(rand_mins) <- colnames(inset_mat)
	rand_maxes <- rand_stats[ , (ncol(inset_mat)+1):(2*ncol(inset_mat))]
	colnames(rand_maxes) <- colnames(inset_mat)
	rand_best <- rand_stats[ , (2*ncol(inset_mat)+1):ncol(rand_stats)]
	colnames(rand_best) <- colnames(inset_mat)

	# calculate max and min NES for each set
	NES_max <- running_max/(colSums((rand_best>=0)*rand_best)/(colSums(rand_best>=0) + 0.000001) + 0.000001)
	NES_min <- running_min/(colSums(abs(rand_best)*(rand_best<=0))/(colSums(rand_best<=0) + 0.000001) + 0.000001)
	NES_max <- NES_max * (colSums(rand_best>=0) > 0)
	NES_min <- NES_min * (colSums(rand_best<=0) > 0)

	pval_max <- colSums(t(t(rand_best) >= running_max))/(colSums(rand_best >= 0) + 0.000001)
	pval_min <- colSums(t(t(rand_best) <= running_min))/(colSums(rand_best <= 0) + 0.000001)

	rand_mins_NES <- t((t(rand_mins)/(colSums(abs(rand_best)*(rand_best<=0))/(colSums(rand_best<=0) +0.000001) + 0.000001)) * (colSums(rand_best<=0) > 0))
	rand_maxes_NES <- t((t(rand_maxes)/(colSums(rand_best*(rand_best>=0))/(colSums(rand_best>=0) + 0.000001) + 0.000001)) * (colSums(rand_best>=0) > 0))
	#rand_mins_NES <- rand_mins_NES * (colSums(rand_best>=0) > 0)
	#rand_maxes_NES <- rand_maxes_NES * (colSums(rand_best<=0) > 0)

	# calculate FDR for each NES
	n_all_rand_min <- sum(rand_best<=0)
	if(n_all_rand_min==0){
		n_all_rand_min <- 0.000001
	}
	n_all_rand_max <- sum(rand_best>=0)
	if(n_all_rand_max==0){
		n_all_rand_min <- 0.000001
	}

	#output_mat <- matrix(0, nrow = ncol(Running_Sum), ncol = 8, dimnames = list(colnames(Running_Sum), c("ES", "NES", "p_val", "fdr", "pos_ES", "neg_ES", "pos_NES", "neg_NES")))
	output_mat <- matrix(0, nrow = ncol(Running_Sum), ncol = 4, dimnames = list(colnames(Running_Sum), c("ES", "NES", "p_val", "fdr")))
	output_df <- data.frame(output_mat, stringsAsFactors = F)

	for(k in 1:length(NES_min)){
		#output_df$pos_ES[k] <- running_max[k]
		#output_df$neg_ES[k] <- running_min[k]
		#output_df$pos_NES[k] <- NES_max[k]
		#output_df$neg_NES[k] <- NES_min[k]
		if(abs(running_min[k])<running_max[k]){
			output_df$ES[k] <- running_max[k]
			output_df$NES[k] <- NES_max[k]
			output_df$p_val[k] <- pval_max[k]
			max_top <- sum(rowSums(rand_maxes_NES*(rand_best>=0) >= NES_max[k])) / n_all_rand_max
			max_bottom <- sum(NES_max[running_max>=abs(running_min)] >= NES_max[k]) / (sum(running_max>=abs(running_min)))
			output_df$fdr[k] <- max_top / max_bottom
		} else {
			output_df$ES[k] <- running_min[k]
			output_df$NES[k] <- NES_min[k]
			output_df$p_val[k] <- pval_min[k]
			min_top <- sum(rowSums(rand_mins_NES*(rand_best<=0) <= NES_min[k])) / n_all_rand_min
			min_bottom <- sum(NES_min[abs(running_min)>running_max] <= NES_min[k]) / sum(abs(running_min)>running_max)
			output_df$fdr[k] <- min_top / min_bottom
		}
	}
	if((thresh_action=="include")&(length(skipped_sets)>0)){
		new_row <- data.frame(matrix(0, nrow=1, ncol=4), stringsAsFactors = F)
		colnames(new_row) <- colnames(output_df)
		new_row$p_val <- 1
		new_row$fdr <- 1
		for(i in 1:length(skipped_sets)){
			rownames(new_row) <- skipped_sets[i]
			output_df <- rbind(output_df, new_row)
		}
	}
	output_df$fdr[output_df$fdr>1] <- 1
	return(list(Enrichment_Results=output_df, Running_Sums=Running_Sum, Items_in_Set=items_in_set))
}


#' Prepare input for standard GSEA
#'
#' A helper to read files for performing standard GSEA.
#'
#' @param rankFile Path of the rnk file
#' @param gmtFile Path of the GMT file
#'
#' @return a data frame to be used in \code{swGsea}
#'
#' @importFrom readr read_tsv
#'
#' @export
#'
prepareGseaInput <- function(rankFile, gmtFile) {
	rank <- read_tsv(rankFile, col_names=c("gene", "score"), col_types="cd")
	gmt <- readGmt(gmtFile)
	inputDf <- prepareInputMatrixGsea(rank, gmt)
	return(inputDf)
}

#' Comprehensive R function for the enrichment analysis
#'
#' Main function for enrichment analysis
#'
#' WebGestaltR function can perform three enrichment analyses:
#' ORA (Over-Representation Analysis) and GSEA (Gene Set Enrichment Analysis).and
#' NTA (Network Topology Analysis). Based on the user-uploaded gene list or gene list
#' with scores, WebGestaltR function will first map the gene list to the entrez gene
#' ids and then summarize the gene list based on the GO (Gene Ontology) Slim. After
#' performing the enrichment analysis, WebGestaltR function also returns a user-friendly
#' HTML report containing GO Slim summary and the enrichment analysis result. If functional
#' categories have DAG (directed acyclic graph) structure or genes in the functional
#' categories have network structure, those relationship can also be visualized in the
#' report.
#'
#' @param enrichMethod Enrichment methods: \code{ORA}, \code{GSEA} or \code{NTA}.
#' @param organism Currently, WebGestaltR supports 12 organisms. Users can use the function
#'   \code{listOrganism} to check available organisms. Users can also input \code{others} to
#'   perform the enrichment analysis for other organisms not supported by WebGestaltR. For
#'   other organisms, users need to provide the functional categories, interesting list and
#'   reference list (for ORA method). Because WebGestaltR does not perform the ID mapping for
#'   the other organisms, the above data should have the same ID type.
#' @param enrichDatabase The functional categories for the enrichment analysis. Users can use
#'   the function \code{listGeneSet} to check the available functional databases for the
#'   selected organism. Multiple databases in a vector are supported for ORA and GSEA.
#' @param enrichDatabaseFile Users can provide one or more GMT files as the functional
#'   category for enrichment analysis. The extension of the file should be \code{gmt} and the
#'   first column of the file is the category ID, the second one is the external link for the
#'   category. Genes annotated to the category are from the third column. All columns are
#'   separated by tabs. The GMT files will be combined with \code{enrichDatabase}.
#' @param enrichDatabaseType The ID type of the genes in the \code{enrichDatabaseFile}.
#'   If users set \code{organism} as \code{others}, users do not need to set this ID type because
#'   WebGestaltR will not perform ID mapping for other organisms. The supported ID types of
#'   WebGestaltR for the selected organism can be found by the function \code{listIdType}.
#' @param enrichDatabaseDescriptionFile Users can also provide description files for the custom
#'   \code{enrichDatabaseFile}. The extension of the description file should be \code{des}. The
#'   description file contains two columns: the first column is the category ID that should be
#'   exactly the same as the category ID in the custom \code{enrichDatabaseFile} and the second
#'   column is the description of the category. All columns are separated by tabs.
#' @param interestGeneFile If \code{enrichMethod} is \code{ORA} or \code{NTA}, the extension of
#'   the \code{interestGeneFile} should be \code{txt} and the file can only contain one column:
#'   the interesting gene list. If \code{enrichMethod} is \code{GSEA}, the extension of the
#'   \code{interestGeneFile} should be \code{rnk} and the file should contain two columns
#'   separated by tab: the gene list and the corresponding scores.
#' @param interestGene Users can also use an R object as the input. If \code{enrichMethod} is
#'   \code{ORA} or \code{NTA}, \code{interestGene} should be an R \code{vector} object
#'   containing the interesting gene list. If \code{enrichMethod} is \code{GSEA},
#'   \code{interestGene} should be an R \code{data.frame} object containing two columns: the
#'   gene list and the corresponding scores.
#' @param interestGeneType The ID type of the interesting gene list. The supported ID types of
#'   WebGestaltR for the selected organism can be found by the function \code{listIdType}. If
#'   the \code{organism} is \code{others}, users do not need to set this parameter.
#' @param collapseMethod The method to collapse duplicate IDs with scores. \code{mean},
#'   \code{median}, \code{min} and \code{max} represent the mean, median, minimum and maximum
#'   of scores for the duplicate IDs.
#' @param referenceGeneFile For the ORA method, the users need to upload the reference gene
#'   list. The extension of the \code{referenceGeneFile} should be \code{txt} and the file can
#'   only contain one column: the reference gene list.
#' @param referenceGene For the ORA method, users can also use an R object as the reference
#'   gene list. \code{referenceGene} should be an R \code{vector} object containing the
#'   reference gene list.
#' @param referenceGeneType The ID type of the reference gene list. The supported ID types
#'   of WebGestaltR for the selected organism can be found by the function \code{listIdType}.
#'   If the \code{organism} is \code{others}, users do not need to set this parameter.
#' @param referenceSet Users can directly select the reference set from existing platforms in
#'   WebGestaltR and do not need to provide the reference set through \code{referenceGeneFile}.
#'   All existing platforms supported in WebGestaltR can be found by the function
#'   \code{listReferenceSet}. If \code{referenceGeneFile} and \code{refereneceGene} are
#'   \code{NULL}, WebGestaltR will use the \code{referenceSet} as the reference gene set.
#'   Otherwise, WebGestaltR will use the user supplied reference set for enrichment analysis.
#' @param minNum WebGestaltR will exclude the categories with the number of annotated genes
#'   less than \code{minNum} for enrichment analysis. The default is \code{10}.
#' @param maxNum WebGestaltR will exclude the categories with the number of annotated genes
#'   larger than \code{maxNum} for enrichment analysis. The default is \code{500}.
#' @param sigMethod Two methods of significance are available in WebGestaltR: \code{fdr} and
#'   \code{top}. \code{fdr} means the enriched categories are identified based on the FDR and
#'   \code{top} means all categories are ranked based on FDR and then select top categories
#'   as the enriched categories. The default is \code{fdr}.
#' @param fdrMethod For the ORA method, WebGestaltR supports five FDR methods: \code{holm},
#'   \code{hochberg}, \code{hommel}, \code{bonferroni}, \code{BH} and \code{BY}. The default
#'   is \code{BH}.
#' @param fdrThr The significant threshold for the \code{fdr} method. The default is \code{0.05}.
#' @param topThr The threshold for the \code{top} method. The default is \code{10}.
#' @param reportNum The number of enriched categories visualized in the final report. The default
#'   is \code{20}. A larger \code{reportNum} may be slow to render in the report.
#' @param perNum The number of permutations for the GSEA method. The default is \code{1000}.
#' @param gseaP The exponential scaling factor of the phenotype score. The default is \code{1}.
#'   When p=0, ES reduces to standard K-S statistics (See original paper for more details).
#' @param isOutput If \code{isOutput} is TRUE, WebGestaltR will create a folder named by
#'   the \code{projectName} and save the results in the folder. Otherwise, WebGestaltR will
#'   only return an R \code{data.frame} object containing the enrichment results. If
#'   hundreds of gene list need to be analyzed simultaneously, it is better to set
#'   \code{isOutput} to \code{FALSE}. The default is \code{TRUE}.
#' @param outputDirectory The output directory for the results.
#' @param projectName The name of the project. If \code{projectName} is \code{NULL},
#'   WebGestaltR will use time stamp as the project name.
#' @param dagColor If \code{dagColor} is \code{binary}, the significant terms in the DAG
#'   structure will be colored by steel blue for ORA method or steel blue (positive related)
#'   and dark orange (negative related) for GSEA method. If \code{dagColor} is \code{continous},
#'   the significant terms in the DAG structure will be colored by the color gradient based on
#'   corresponding FDRs.
#' @param saveRawGseaResult Whether the raw result from GSEA is saved as a RDS file, which can be
#'   used for plotting. Defaults to \code{FALSE}. The list includes
#'   \describe{
#'     \item{Enrichment_Results}{A data frame of GSEA results with statistics}
#'     \item{Running_Sums}{A matrix of running sum of scores for each gene set}
#'     \item{Items_in_Set}{A list with ranks of genes for each gene set}
#'  }
#' @param gseaPlotFormat The graphic format of GSEA enrichment plots. Either \code{svg},
#'   \code{png}, or \code{c("png", "svg")} (default).
#' @param setCoverNum The number of expected gene sets after set cover to reduce redundancy.
#'   It could get fewer sets if the coverage reaches 100\%. The default is \code{10}.
#' @param networkConstructionMethod Netowrk construction method for NTA. Either
#'   \code{Network_Retrieval_Prioritization} or \code{Network_Expansion}. Network Retrieval &
#'   Prioritization first uses random walk analysis to calculate random walk probabilities
#'   for the input seeds, then identifies the relationships among the seeds in the selected
#'   network and returns a retrieval sub-network. The seeds with the top random walk
#'   probabilities are highlighted in the sub-network. Network Expansion first uses random
#'   walk analysis to rank all genes in the selected network based on their network
#'   proximity to the input seeds and then return an expanded sub-network in which nodes
#'   are the input seeds and their top ranking neighbors and edges represent their
#'   relationships.
#' @param neighborNum The number of neighbors to include in NTA Network Expansion method.
#' @param highlightType The type of nodes to highlight in the NTA Network Expansion method,
#'   either \code{Seeds} or \code{Neighbors}.
#' @param highlightSeedNum The number of top input seeds to highlight in NTA Network Retrieval
#'   & Prioritizaiton method.
#' @param nThreads The number of cores to use for GSEA and set cover, and in batch function.
#' @param cache A directory to save data cache for reuse. Defaults to \code{NULL} and disabled.
#' @param hostName The server URL for accessing data. Mostly for development purposes.
#' @param ... In batch function, passes parameters to WebGestaltR function.
#'   Also handles backward compatibility for some parameters in old versions.
#'
#' @return The WebGestaltR function returns a data frame containing the enrichment analysis
#'   result and also outputs an user-friendly HTML report if \code{isOutput} is \code{TRUE}.
#'   The columns in the data frame depend on the \code{enrichMethod} and they are the following:
#'   \describe{
#'     \item{geneSet}{ID of the gene set.}
#'     \item{description}{Description of the gene set if available.}
#'     \item{link}{Link to the data source.}
#'     \item{size}{The number of genes in the set after filtering by \code{minNum} and \code{maxNum}.}
#'     \item{overlap}{The number of mapped input genes that are annotated in the gene set.}
#'     \item{expect}{Expected number of input genes that are annotated in the gene set.}
#'     \item{enrichmentRatio}{Enrichment ratio, overlap / expect.}
#'     \item{enrichmentScore}{Enrichment score, the maximum running sum of scores for the ranked list.}
#'     \item{normalizedEnrichmentScore}{Normalized enrichment score, normalized against the average
#'       enrichment score of all permutations.}
#'     \item{leadingEdgeNum}{Number of genes/phosphosites in the leading edge.}
#'     \item{pValue}{P-value from hypergeometric test for ORA. For GSEA, please refer to its original
#'       publication or online at \url{https://software.broadinstitute.org/gsea/doc/GSEAUserGuideTEXT.htm}.}
#'     \item{FDR}{Corrected P-value for mulilple testing with \code{fdrMethod} for ORA.}
#'     \item{overlapId}{The gene/phosphosite IDs of \code{overlap} for ORA (entrez gene IDs or
#'       phosphosite sequence).}
#'     \item{leadingEdgeId}{Genes/phosphosites in the leading edge in entrez gene ID or
#'       phosphosite sequence.}
#'     \item{userId}{The gene/phosphosite IDs of \code{overlap} for ORA or \code{leadingEdgeId}
#'       for GSEA in User input IDs.}
#'     \item{plotPath}{Path of the GSEA enrichment plot.}
#'     \item{database}{Name of the source database if multiple enrichment databases are given.}
#'     \item{goId}{In NTA, like \code{geneSet}, the enriched GO terms of genes in the
#'       returned subnetwork.}
#'     \item{interestGene}{In NTA, the gene IDs in the subnetwork with 0/1 annotations indicating
#'       if it is from user input.}
#'   }
#'
#' @export
#'
#' @examples
#' \dontrun{
#' ####### ORA example #########
#' geneFile <- system.file("extdata", "interestingGenes.txt", package="WebGestaltR")
#' refFile <- system.file("extdata", "referenceGenes.txt", package="WebGestaltR")
#' outputDirectory <- getwd()
#' enrichResult <- WebGestaltR(enrichMethod="ORA", organism="hsapiens",
#'   enrichDatabase="pathway_KEGG", interestGeneFile=geneFile,
#'   interestGeneType="genesymbol", referenceGeneFile=refFile,
#'   referenceGeneType="genesymbol", isOutput=TRUE,
#'   outputDirectory=outputDirectory, projectName=NULL)
#'
#' ####### GSEA example #########
#' rankFile <- system.file("extdata", "GeneRankList.rnk", package="WebGestaltR")
#' outputDirectory <- getwd()
#' enrichResult <- WebGestaltR(enrichMethod="GSEA", organism="hsapiens",
#'   enrichDatabase="pathway_KEGG", interestGeneFile=rankFile,
#'   interestGeneType="genesymbol", sigMethod="top", topThr=10, minNum=5,
#'   outputDirectory=outputDirectory)
#'
#' ####### NTA example #########
#' enrichResult <- WebGestaltR(enrichMethod="NTA", organism="hsapiens",
#'   enrichDatabase="network_PPI_BIOGRID", interestGeneFile=geneFile,
#'   interestGeneType="genesymbol", sigMethod="top", topThr=10,
#'   outputDirectory=getwd(), highlightSeedNum=10,
#'   networkConstructionMethod="Network_Retrieval_Prioritization")
#' }
#'
WebGestaltR <- function(enrichMethod="ORA", organism="hsapiens", enrichDatabase=NULL, enrichDatabaseFile=NULL, enrichDatabaseType=NULL, enrichDatabaseDescriptionFile=NULL, interestGeneFile=NULL, interestGene=NULL, interestGeneType=NULL, collapseMethod="mean", referenceGeneFile=NULL, referenceGene=NULL, referenceGeneType=NULL, referenceSet=NULL, minNum=10, maxNum=500, sigMethod="fdr", fdrMethod="BH", fdrThr=0.05, topThr=10, reportNum=20, perNum=1000, gseaP=1, isOutput=TRUE, outputDirectory=getwd(), projectName=NULL, dagColor="continuous", saveRawGseaResult=FALSE, gseaPlotFormat=c("png", "svg"), setCoverNum=10, networkConstructionMethod=NULL, neighborNum=10, highlightType="Seeds", highlightSeedNum=10, nThreads=1, cache=NULL, hostName="https://www.webgestalt.org/", ...) {
	extraArgs <- list(...)
	if ('keepGSEAFolder' %in% names(extraArgs) | 'keepGseaFolder' %in% names(extraArgs)) {
		warning("Parameter keepGSEAFolder is obsolete.\n")
	}
	if ('is.output' %in% names(extraArgs)) {
		isOutput <- extraArgs$is.output
		warning("Parameter is.output is deprecated and changed to isOutput!\n")
		warning("Column names of the result data frame are modified.")
	}
	if ('methodType' %in% names(extraArgs)) {
		warning("Parameter methodType is obsolete.\n")
	}
	if ('lNum' %in% names(extraArgs)) {
		warning("Parameter lNum is obsolete.\n")
	}
	if ('dNum' %in% names(extraArgs)) {
		warning("Parameter dNum is deprecated and changed to reportNum.\n")
		reportNum <- extraArgs$dNum
	}

	if (!is.null(cache)) {
		cat("Use cache data if available.\n")
	}

	## TODO: add para test for NTA
	errorTest <- parameterErrorMessage(enrichMethod=enrichMethod, organism=organism, collapseMethod=collapseMethod, minNum=minNum, maxNum=maxNum, fdrMethod=fdrMethod, sigMethod=sigMethod, fdrThr=fdrThr, topThr=topThr, reportNum=reportNum, perNum=perNum, isOutput=isOutput, outputDirectory=outputDirectory, dagColor=dagColor, hostName=hostName, cache=cache)
	if(!is.null(errorTest)){
		return(errorTest)
	}

	if(is.null(projectName)){
		projectName <- as.character(as.integer(Sys.time()))
	}
	projectName <- sanitizeFileName(projectName) # use for GOSlim summary file name, convert punct to _
	if (enrichMethod == "ORA") {
		enrichR <- WebGestaltROra(organism=organism, enrichDatabase=enrichDatabase, enrichDatabaseFile=enrichDatabaseFile, enrichDatabaseType=enrichDatabaseType, enrichDatabaseDescriptionFile=enrichDatabaseDescriptionFile,  interestGeneFile=interestGeneFile, interestGene=interestGene, interestGeneType=interestGeneType, collapseMethod=collapseMethod, referenceGeneFile=referenceGeneFile, referenceGene=referenceGene, referenceGeneType=referenceGeneType, referenceSet=referenceSet, minNum=minNum, maxNum=maxNum, fdrMethod=fdrMethod, sigMethod=sigMethod, fdrThr=fdrThr, topThr=topThr, reportNum=reportNum, setCoverNum=setCoverNum, isOutput=isOutput, outputDirectory=outputDirectory, projectName=projectName, dagColor=dagColor, nThreads=nThreads, cache=cache, hostName=hostName)
	} else if (enrichMethod == "GSEA") {
		enrichR <- WebGestaltRGsea(organism=organism, enrichDatabase=enrichDatabase, enrichDatabaseFile=enrichDatabaseFile, enrichDatabaseType=enrichDatabaseType, enrichDatabaseDescriptionFile=enrichDatabaseDescriptionFile,  interestGeneFile=interestGeneFile, interestGene=interestGene, interestGeneType=interestGeneType, collapseMethod=collapseMethod, minNum=minNum, maxNum=maxNum, fdrMethod=fdrMethod, sigMethod=sigMethod, fdrThr=fdrThr, topThr=topThr, reportNum=reportNum, setCoverNum=setCoverNum, perNum=perNum, p=gseaP, isOutput=isOutput, outputDirectory=outputDirectory, projectName=projectName, dagColor=dagColor, saveRawGseaResult=saveRawGseaResult, plotFormat=gseaPlotFormat, nThreads=nThreads, cache=cache, hostName=hostName)
	} else if (enrichMethod == "NTA") {
		enrichR <- WebGestaltRNta(organism=organism, network=enrichDatabase, method=networkConstructionMethod, neighborNum=neighborNum, highlightSeedNum=highlightSeedNum, inputSeed=interestGene, inputSeedFile=interestGeneFile, interestGeneType=interestGeneType, sigMethod=sigMethod, fdrThr=fdrThr, topThr=topThr, outputDirectory=outputDirectory, projectName=projectName, highlightType=highlightType, cache=cache, hostName=hostName)
	}

	return(enrichR)
}

urlToFile <- function(dataUrl) {
	result <- sub("^http://", "", dataUrl)
	result <- sub("^https://", "", result)
	result <- gsub("\\?[^?]+?=", "_", result)
	result <- gsub("&[^&]+?=", "_", result)
	result <- gsub("[:/.]", "_", result)
	result <- gsub("_+", "_", result, fixed=TRUE)
	return(result)
}

#' cacheUrl
#'
#' Get data from a URL or cache and optionally save in cache for reuse
#'
#' @param dataUrl The URL of data
#' @param cache The cache directory. Defaults to \code{NULL} and fetches data from server.
#' @param query The list of queries passed on to httr methods
#'
#' @return response object from httr request
#'
#' @importFrom httr GET
#' @keywords internal
#'
cacheUrl <- function(dataUrl, cache=NULL, query=NULL) {
	if (!is.null(cache)) {
		dir.create(cache, showWarnings=FALSE)
		if (!is.null(query)) {
			localFilePrefix <- urlToFile(paste0(dataUrl, "_", paste0(query, collapse="_")))
		} else {
			localFilePrefix <- urlToFile(dataUrl)
		}
		localFile <- file.path(cache, paste0(localFilePrefix, ".rds"))
	}
	if (!is.null(cache) && file.exists(localFile)) {
		#cat("Reading from cache: ", localFile, "\n")
		response <- readRDS(localFile)
	} else {
		#cat("Reading from server: ", dataUrl, "\n")
		if (!is.null(query)) {
			response <- GET(dataUrl, query=query)
		} else {
			response <- GET(dataUrl)
		}
		if (response$status_code != 200) {
			return(response)
		}
		if (!is.null(cache)) {
			saveRDS(response, localFile)
		}
	}
	return(response)
}

#' Wrapper for batch WebGestaltR runs
#'
#' @param interestGeneFolder Run WebGestaltR for gene list files in the folder.
#' @param isParallel If jobs are run parallelly in the batch.
#'
#' @return The WebGestaltRBatch function returns a list of enrichment results.
#'
#' @importFrom parallel makeCluster stopCluster
#' @importFrom doParallel registerDoParallel
#' @importFrom foreach foreach %dopar%
#'
#' @export
#' @aliases WebGestaltR_batch
#' @rdname WebGestaltR
#'
WebGestaltRBatch <- function(interestGeneFolder=NULL, enrichMethod="ORA", isParallel=FALSE, nThreads=3, ...) {
	args <- list(...)
	if(enrichMethod=="ORA" || enrichMethod=="NTA"){
		interestGeneFiles <- list.files(interestGeneFolder,pattern="\\.txt",full.names=TRUE)
	}

	if(enrichMethod=="GSEA"){
		interestGeneFiles <- list.files(interestGeneFolder,pattern="\\.rnk",full.names=TRUE)
	}

	projectNames <- unlist(lapply(strsplit(basename(interestGeneFiles),split=".",fixed=TRUE),function(e){return(paste(e[-length(e)],collapse="."))}))

	resultList <- list()

	if(isParallel==TRUE){
		cl <- makeCluster(nThreads)
		registerDoParallel(cl)
		resultList <- foreach(i=1:length(interestGeneFiles), .packages="WebGestaltR") %dopar% {
			args$interestGeneFile <- interestGeneFiles[i]
			args$projectName <- projectNames[i]
			args$enrichMethod <- enrichMethod
			sig <- do.call(WebGestaltR, args)
			re <- list(filename=interestGeneFiles[i], enrichResult=sig)
			return(re)
		}
		stopCluster(cl)
	}else{
		for(i in c(1:length(interestGeneFiles))){
			cat("Process file: ",interestGeneFiles[i],"\n",sep="")
			args$interestGeneFile <- interestGeneFiles[i]
			args$projectName <- projectNames[i]
			args$enrichMethod <- enrichMethod
			sig <- do.call(WebGestaltR, args)
			re <- list(filename=interestGeneFiles[i], enrichResult=sig)
			resultList[[i]] <- re
		}
	}
	return(resultList)
}

#' @export
WebGestaltR_batch <- function(is.output=TRUE, ...) {
	warning("Function WebGestaltR_batch is deprecated and changed to WebGestaltRBatch!\n")
	return(WebGestaltRBatch(isOutput=is.output, ...))
}

#' Weighted Set Cover
#'
#' Size constrained weighted set cover problem to find top N sets while
#' maximizing the coverage of all elements.
#'
#' @param idsInSet A list of set names and their member IDs.
#' @param costs A vector of the same length to add weights for penalty, i.e. 1/-logP.
#' @param topN The number of sets (or less when it completes early) to return.
#' @param nThreads The number of processes to use. In Windows, it fallbacks to 1.
#'
#' @return A list of \code{topSets} and \code{coverage}.
#' \describe{
#'  \item{topSets}{A list of set IDs.}
#'  \item{coverage}{The percentage of IDs covered in the top sets.}
#' }
#'
#' @importFrom dplyr bind_rows %>%
#' @importFrom parallel mclapply
#' @export
#' @author Zhiao Shi, Yuxing Liao
#'
weightedSetCover <- function(idsInSet, costs, topN, nThreads=4) {
  cat("Begin weighted set cover...\n")
  names(costs) <- names(idsInSet)
  if (.Platform$OS.type == "windows") {
    nThreads = 1
  }
  multiplier <- 10
  # we only start with the top (multiplier * topN) most
  # significant sets to reduce computational cost
  max_num_set <- multiplier * topN
  if (length(idsInSet) > max_num_set) {
    # sort by absolute of cost (1/signedLogP)
    index <- order(abs(costs), decreasing=FALSE)
    costs <- costs[index][1:max_num_set]
    idsInSet <- idsInSet[index][1:max_num_set]
  }


  s.hat <- 1.0
  # get all unique genes in all enriched sets
  all.genes <- unique(unlist(idsInSet))
  remain <- s.hat * length(all.genes)

  # final results, contains a list of gene set names
  cur.res <- c()
  # current candidates with marginal gain and size
  all.set.names <- names(idsInSet)
  mc_results <- mclapply(all.set.names, function(cur_name, cur_res, idsInSet, costs) {
      cur_gain <- marginalGain(cur_name, cur_res, idsInSet, costs)
      cur_size <- length(idsInSet[[cur_name]])
      return(data.frame(geneset.name=cur_name, gain=cur_gain, size=cur_size, stringsAsFactors=FALSE))
    }, cur_res=cur.res, idsInSet=idsInSet, costs=costs, mc.cores=nThreads)
  candidates <- mc_results %>% bind_rows()
  topN <- min(topN, nrow(candidates))
  for (i in seq(topN)) {
    # if there is no candidates, return
    if (nrow(candidates) == 0) {
      covered.genes <- unique(unlist(idsInSet[cur.res]))
      s.hat <- length(covered.genes) / length(all.genes)
      cat("No more candidates, ending weighted set cover\n")
      return(list(topSets=cur.res, coverage=s.hat))
    }
    # find the set with maximum marginal gain
    # tie breaker: for two sets with sname marginal gain, pick the one with
    # larger size
    candidates <- candidates[order(-candidates$gain, -candidates$size), ]
    # update remain
    remain <- remain - length(marginalBenefit(candidates[1, "geneset.name"], cur.res, idsInSet))
    cur.res <- c(cur.res, candidates[1,"geneset.name"])
    if (remain == 0) {
      covered.genes <- unique(unlist(idsInSet[cur.res]))
      s.hat <- length(covered.genes) / length(all.genes)
      cat("Remain is 0, ending weighted set cover\n")
      # full coverage solution
      return(list(topSets=cur.res, coverage=s.hat))
    }
    # update candidates
    # first remove the one just been selected
    candidates <- candidates[-1, ]
    # recalculate gain, remove rows with gain == 0
    if (nrow(candidates) > 0) {
      mc_results <- mclapply(seq(nrow(candidates)), function(row, candidates, cur_res, idsInSet, costs){
      cur_name <- candidates[row, "geneset.name"]
      cur_gain <- marginalGain(cur_name, cur_res, idsInSet, costs)
      if(cur_gain != 0) {
        candidates[candidates$geneset.name == cur_name, "gain"] <- cur_gain
        tmp_candidate <- candidates[candidates$geneset.name == cur_name,]
         return(tmp_candidate)
      }
    }, candidates=candidates, cur_res=cur.res, idsInSet=idsInSet, costs=costs, mc.cores=nThreads)

      new_candidates <- mc_results %>% bind_rows()
      candidates <- new_candidates
    }
  }
  # not fully covered, compute the current coverage and return
  covered.genes <- unique(unlist(idsInSet[cur.res]))
  s.hat <- length(covered.genes) / length(all.genes)
  cat("End weighted set cover...\n")
  return(list(topSets=cur.res, coverage=s.hat))
}


# return a list of genes from all.genes that has not been
# covered so far
# cur.set.name: name of the candidate gene set
# cur.res: vector of names of gene sets in current result
marginalBenefit <- function(cur.set.name, cur.res, idsInSet) {
  all.genes <- unique(unlist(idsInSet))
  cur.genes <- idsInSet[[cur.set.name]]
  if(length(cur.res) == 0){
    not.covered.genes <- cur.genes
  } else{
    covered.genes <- unique(unlist(idsInSet[cur.res]))
    not.covered.genes <- setdiff(cur.genes, covered.genes)
  }
  return(not.covered.genes)
}


marginalGain <- function(cur.set.name, cur.res, idsInSet, costs) {
  abs_cur_cost <- abs(costs[cur.set.name])
  cur.mben <- marginalBenefit(cur.set.name, cur.res, idsInSet)
  return(length(cur.mben) / abs_cur_cost)
}

#' @importFrom dplyr select distinct left_join arrange %>% mutate
#' @importFrom readr write_tsv
WebGestaltRGsea <- function(organism="hsapiens", enrichDatabase=NULL, enrichDatabaseFile=NULL, enrichDatabaseType=NULL, enrichDatabaseDescriptionFile=NULL,  interestGeneFile=NULL, interestGene=NULL, interestGeneType=NULL, collapseMethod="mean", minNum=10, maxNum=500, fdrMethod="BH", sigMethod="fdr", fdrThr=0.05, topThr=10, reportNum=20, setCoverNum=10, perNum=1000, p=1, isOutput=TRUE, outputDirectory=getwd(), projectName=NULL, dagColor="binary", saveRawGseaResult=FALSE, plotFormat="png", nThreads=1, cache=NULL, hostName="https://www.webgestalt.org/") {
	enrichMethod <- "GSEA"
	projectDir <- file.path(outputDirectory, paste0("Project_", projectName))

	######### Web server will input "NULL" to the R package, thus, we need to change "NULL" to NULL ########
	enrichDatabase <- testNull(enrichDatabase)
	enrichDatabaseFile <- testNull(enrichDatabaseFile)
	enrichDatabaseType <- testNull(enrichDatabaseType)
	enrichDatabaseDescriptionFile <- testNull(enrichDatabaseDescriptionFile)
	interestGeneFile <- testNull(interestGeneFile)
	interestGene <- testNull(interestGene)
	interestGeneType <- testNull(interestGeneType)

	################ Check parameter ################
	errorTest <- parameterErrorMessage(enrichMethod=enrichMethod, organism=organism, collapseMethod=collapseMethod, minNum=minNum, maxNum=maxNum, fdrMethod=fdrMethod, sigMethod=sigMethod, fdrThr=fdrThr, topThr=topThr, reportNum=reportNum, perNum=perNum, isOutput=isOutput, outputDirectory=outputDirectory, dagColor=dagColor, hostName=hostName, cache=cache)

	if(!is.null(errorTest)){
		stop(errorTest)
	}

	############# Check enriched database #############
	cat("Loading the functional categories...\n")
	enrichD <- loadGeneSet(organism=organism, enrichDatabase=enrichDatabase, enrichDatabaseFile=enrichDatabaseFile, enrichDatabaseType=enrichDatabaseType, enrichDatabaseDescriptionFile=enrichDatabaseDescriptionFile, cache=cache, hostName=hostName)

	geneSet <- enrichD$geneSet
	geneSetDes <- enrichD$geneSetDes
	geneSetDag <- enrichD$geneSetDag
	geneSetNet <- enrichD$geneSetNet
	databaseStandardId <- enrichD$standardId
	rm(enrichD)

	########### Check input interesting gene list ###############
	cat("Loading the ID list...\n")
	interestingGeneMap <- loadInterestGene(organism=organism, dataType="rnk", inputGeneFile=interestGeneFile, inputGene=interestGene, geneType=interestGeneType, collapseMethod=collapseMethod, cache=cache, hostName=hostName, geneSet=geneSet)

	if (organism == "others") {
		interestGeneList <- unique(interestingGeneMap)
	} else {
		interestStandardId <- interestingGeneMap$standardId
		interestGeneList <- interestingGeneMap$mapped %>% select(interestStandardId, .data$score) %>% distinct()
	}

	########## Create project folder ##############
	if (isOutput) {
		dir.create(projectDir)

	###### Summarize gene annotation based on the GOSlim ###########
		if (organism != "others") {
			if (databaseStandardId == "entrezgene") {
				cat("Summarizing the uploaded ID list by GO Slim data...\n")
				goSlimOutput <- file.path(projectDir, paste0("goslim_summary_", projectName))
				re <- goSlimSummary(organism=organism, geneList=interestGeneList[[interestStandardId]], outputFile=goSlimOutput, outputType="png", isOutput=isOutput, cache=cache, hostName=hostName)
			}
			write_tsv(interestingGeneMap$mapped, file.path(projectDir, paste0("interestingID_mappingTable_", projectName, ".txt")))
			write(interestingGeneMap$unmapped, file.path(projectDir, paste0("interestingID_unmappedList_", projectName, ".txt")))
		} else {
			write_tsv(interestGeneList, file.path(projectDir, paste0("interestList_", projectName, ".txt")), col_names=FALSE)
		}
	}

	############# Run enrichment analysis ###################
	cat("Performing the enrichment analysis...\n")

	gseaRes <- gseaEnrichment(hostName, outputDirectory, projectName, interestGeneList,
		geneSet, geneSetDes=geneSetDes, minNum=minNum, maxNum=maxNum, sigMethod=sigMethod, fdrThr=fdrThr,
		topThr=topThr, perNum=perNum, p=p, nThreads=nThreads, saveRawGseaResult=saveRawGseaResult, plotFormat=plotFormat, isOutput=isOutput
	)
	if (is.null(gseaRes)) {
		return(NULL)
	}

	enrichedSig <- gseaRes$enriched
	insig <- gseaRes$background


	clusters <- list()
	geneTables <- list()
	if (!is.null(enrichedSig)) {
		if (!is.null(geneSetDes)) { ####### Add extra description information ###########
			enrichedSig <- enrichedSig %>%
				left_join(geneSetDes, by="geneSet") %>%
				select(.data$geneSet, .data$description, .data$link, .data$enrichmentScore, .data$normalizedEnrichmentScore, .data$pValue, .data$FDR, .data$size, .data$plotPath, .data$leadingEdgeNum, .data$leadingEdgeId) %>%
				arrange(.data$FDR, .data$pValue, desc(.data$normalizedEnrichmentScore)) %>%
				mutate(description=ifelse(is.na(.data$description), "", .data$description))
		} else {
			enrichedSig <- enrichedSig %>%
				select(.data$geneSet, .data$link, .data$enrichmentScore, .data$normalizedEnrichmentScore, .data$pValue, .data$FDR, .data$size, .data$plotPath, .data$leadingEdgeNum, .data$leadingEdgeId) %>%
				arrange(.data$FDR, .data$pValue, desc(.data$normalizedEnrichmentScore))
		}

		geneTables <- getGeneTables(organism, enrichedSig, "leadingEdgeId", interestingGeneMap)
		if (organism != "others") {
			enrichedSig$link <- mapply(function(link, geneList) linkModification("GSEA", link, geneList, interestingGeneMap),
				enrichedSig$link,
				enrichedSig$leadingEdgeId
			)
		}

		if ("database" %in% colnames(geneSet)) {
			# add source database for multiple databases
			enrichedSig <- enrichedSig %>% left_join(unique(geneSet[, c("geneSet", "database")]), by="geneSet")
		}

		if (organism != "others" && interestGeneType != interestStandardId) {
			outputEnrichedSig <- mapUserId(enrichedSig, "leadingEdgeId", interestingGeneMap)
		} else {
			outputEnrichedSig <- enrichedSig
		}

		if (isOutput) {
			write_tsv(outputEnrichedSig, file.path(projectDir, paste0("enrichment_results_", projectName, ".txt")))
			idsInSet <- sapply(enrichedSig$leadingEdgeId, strsplit, split=";")
			names(idsInSet) <- enrichedSig$geneSet
			pValue <- enrichedSig$pValue
			pValue[pValue == 0] <- .Machine$double.eps
			signedLogP <- -log(pValue) * sign(enrichedSig$enrichmentScore)
			apRes <- affinityPropagation(idsInSet, signedLogP)
			wscRes <- weightedSetCover(idsInSet, 1 / signedLogP, setCoverNum, nThreads)
			if (!is.null(apRes)) {
				writeLines(sapply(apRes$clusters, paste, collapse="\t"), file.path(projectDir, paste0("enriched_geneset_ap_clusters_", projectName, ".txt")))
			} else {
				apRes <- NULL
			}
			clusters$ap <- apRes
			if (!is.null(wscRes$topSets)) {
				writeLines(c(paste0("# Coverage: ", wscRes$coverage), wscRes$topSets), file.path(projectDir, paste0("enriched_geneset_wsc_topsets_", projectName, ".txt")))
				clusters$wsc <- list(representatives=wscRes$topSets,  coverage=wscRes$coverage)
			} else {
				clusters$wsc <- NULL
			}
		}
	}

	if (isOutput) {
		############## Create report ##################
		cat("Generate the final report...\n")
		createReport(hostName=hostName, outputDirectory=outputDirectory, organism=organism, projectName=projectName, enrichMethod=enrichMethod, geneSet=geneSet, geneSetDes=geneSetDes, geneSetDag=geneSetDag, geneSetNet=geneSetNet, interestingGeneMap=interestingGeneMap, enrichedSig=enrichedSig, background=insig, geneTables=geneTables, clusters=clusters, enrichDatabase=enrichDatabase, enrichDatabaseFile=enrichDatabaseFile, enrichDatabaseType=enrichDatabaseType, enrichDatabaseDescriptionFile=enrichDatabaseDescriptionFile, interestGeneFile=interestGeneFile, interestGene=interestGene, interestGeneType=interestGeneType, collapseMethod=collapseMethod, minNum=minNum, maxNum=maxNum, fdrMethod=fdrMethod, sigMethod=sigMethod, fdrThr=fdrThr, topThr=topThr, reportNum=reportNum, perNum=perNum, p=p, dagColor=dagColor)

		cwd <- getwd()
		setwd(projectDir)
		zip(paste0("Project_", projectName, ".zip"), ".", flags="-rq")
		setwd(cwd)

		cat("Results can be found in the ", projectDir, "!\n", sep="")
	}
	return(outputEnrichedSig)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' Fill relation data frame for GSEA input
#'
#' Fill 1 for gene in gene set
#'
#' @param gmt A Data Frame with geneSet and gene columns from the GMT file
#' @param genes A vector of genes
#' @param geneSets A vector of gene sets
#'
#' @return A Data Frame with the first column of gene and 1 or 0 for other columns of gene sets.
#' @author Yuxing Liao
#' @keywords internal
fillInputDataFrame <- function(gmt, genes, geneSets) {
    .Call('_WebGestaltR_fillInputDataFrame', PACKAGE = 'WebGestaltR', gmt, genes, geneSets)
}

#' Permutaion in GSEA algorithm
#'
#' @param inset_scores Scaled score matrix for genes in sets
#' @param outset_scores Normalized score matrix for genes not in sets
#' @param expression_value Vector of gene rank scores
#'
#' @return A vector of concatenated random minimal,maimum and best running sum scores for each set.
#' @author Yuxing Liao
#' @keywords internal
gseaPermutation <- function(inset_scores, outset_scores, expression_value) {
    .Call('_WebGestaltR_gseaPermutation', PACKAGE = 'WebGestaltR', inset_scores, outset_scores, expression_value)
}

#' List Gene Sets
#'
#' List available gene sets for the given organism on WebGestalt server.
#'
#' @inheritParams WebGestaltR
#'
#' @return A data frame of available gene sets.
#'
#' @importFrom httr content
#' @importFrom jsonlite fromJSON
#' @export
#'
listGeneSet <- function(organism="hsapiens", hostName="https://www.webgestalt.org/", cache=NULL) {
	if (startsWith(hostName, "file://")) {
		jsonData <- fromJSON(removeFileProtocol(file.path(hostName, "genesetsummary.json")))
		ids <- jsonData[[organism]]
		name1 <- names(ids)
		idList <- data.frame(name="",description="",idType="",stringsAsFactors=F)
		di <- 1
		for(i in c(1:length(name1))){
			x <- ids[[i]]$name
			if(!(is.null(x) || (is.list(x) && length(x)==0))){
				y <- ids[[i]]$description
				z <- ids[[i]]$idtype
				x <- paste(name1[i],"_",x,sep="")
				idList[di:(di+length(x)-1),1] <- x
				idList[di:(di+length(x)-1),2] <- y
				idList[di:(di+length(x)-1),3] <- z
				di <- di+length(x)
			}
	}
	} else {
		response <- cacheUrl(file.path(hostName, "api", "summary", "geneset"), cache)
		if (response$status_code != 200) {
			return(webRequestError(response))
		}
		jsonData <- content(response)
	ids <- jsonData[[organism]]
	name1 <- names(ids)
	idList <- data.frame(name="",description="",idType="",stringsAsFactors=F)
	di <- 1
	for(i in c(1:length(name1))){
		x <- .getName(ids[[i]])
		if(!(is.null(x) || (is.list(x) && length(x)==0))){
			y <- .getDescription(ids[[i]])
			z <- .getIdType(ids[[i]])
			x <- paste(name1[i],"_",x,sep="")
			idList[di:(di+length(x)-1),1] <- x
			idList[di:(di+length(x)-1),2] <- y
			idList[di:(di+length(x)-1),3] <- z
			di <- di+length(x)
		}
	}
	}
	return(idList)
}

.getName <- function(ids){
	return(sapply(ids,function(e){return(e$name)}))
}

.getDescription <- function(ids){
	return(sapply(ids,function(e){return(e$description)}))
}

.getIdType <- function(ids){
	return(sapply(ids,function(e){return(e$idtype)}))
}

#' Check Format and Read Data
#'
#' @inheritParams idMapping
#'
#' @return A list of data frame
#'
#' @importFrom tools file_ext
#' @importFrom readr read_tsv stop_for_problems
#'
#' @export
#'
formatCheck <- function(dataType="list", inputGeneFile=NULL, inputGene=NULL) {
	dataTypeA <- c("list", "rnk")
	if (length(dataType)>1 || !is.character(dataType) || length(which(dataTypeA==dataType))==0) {
		stop("ERROR: DataType parameter can only be one of 'list' or 'rnk'.")
	}

	if (dataType=="list") {
		if (!is.null(inputGeneFile)) {
			if (file_ext(inputGeneFile)!="txt") {
				stop("ERROR: For the user ID list, please upload a 'txt' file with only one column.")
			} else {
				inputGene <- read_tsv(inputGeneFile, col_names=FALSE, col_types="c")
				error <- tryCatch(stop_for_problems(inputGene),
					error=function(e) stop("ERROR: The format of the uploaded gene list is incorrect. Please make sure the file only contains one column of gene IDs.")
				)
				return(inputGene[[1]])
			}
		} else {
			if (!is.null(inputGene)) {
				if (!is.vector(inputGene)) {
					stop("ERROR: For the user ID list, please upload an R vector object.")
				} else {
					inputGene <- unname(sapply(inputGene, as.character)) # as.character will convert length-1 vector to primitive
					return(inputGene)
				}
			} else {
				stop("ERROR: Please upload a file or an R object for the ID list.")
			}
		}
	}

	if (dataType=="rnk") {
		if (!is.null(inputGeneFile)) {
			if (file_ext(inputGeneFile)!="rnk") {
				stop("ERROR: For the ranked list, please upload a 'rnk' file with two columns (ids and scores).")
			} else {
				inputGene <- read_tsv(inputGeneFile, col_names=c("gene", "score"), col_types="cd")
				# readr just gives warning
				if (all(is.na(inputGene$gene)) || all(is.na(inputGene$score))) {
					stop("ERROR: For the ranked list, please upload a 'rnk' file with two columns (ids and scores).")
				}

				error <- tryCatch(stop_for_problems(inputGene),
					error=function(e) stop("ERROR: The format of the uploaded ranked list is incorrect. Please make sure the file contains two column of gene IDs and scores.")
				 )

				#########GSEA do not allow the second column contains NA. Thus, we should remove NA first############
				return(filter(inputGene, !is.na(.data$score)))
			}
		} else {
			if (!is.null(inputGene)) {
				if (!is.data.frame(inputGene)) {
					stop("ERROR: For the ranked list, please upload an R data.frame object.")
				} else {
					if (!is.numeric(inputGene[[2]]) && !is.integer(inputGene[[2]])) {
						stop("ERROR: The second column of the ranked list should be the numeric scores.")
					} else {
						#########GSEA do not allow the second column contains NA. Thus, we should remove NA first############
						inputGene <- inputGene[!is.na(inputGene[[2]]), ]
						inputGene[[1]] <- as.character(inputGene[[1]])
						colnames(inputGene) <- c("gene", "score")
						return(inputGene)
					}
				}
			} else {
				stop("ERROR: Please upload a file or an R object for the ranked list.")
			}
		}
	}
}

#' @importFrom httr POST content
#' @importFrom dplyr right_join select left_join %>%
idMappingPhosphosite <- function(organism="hsapiens", dataType="list", inputGeneFile=NULL, inputGene=NULL, sourceIdType, targetIdType, collapseMethod="mean", mappingOutput=FALSE,  outputFileName="", hostName="https://www.webgestalt.org/") {

	###########Check input data type###############
	inputGene <- idMappingInput(dataType=dataType,inputGeneFile=inputGeneFile,inputGene=inputGene)

	##########ID Mapping Specify to phosphosite level###############
	if(dataType=="list"){
		inputGeneL <- unique(inputGene)
	}

	if(dataType=="rnk"){
		######Collapse the gene ids with multiple scores##########
		x <- tapply(inputGene$score, inputGene$gene, collapseMethod)
		inputGene <- data.frame(gene=names(x),score=as.numeric(x),stringsAsFactors=FALSE)
		inputGeneL <- inputGene$gene
		colnames(inputGene) <- c(sourceIdType,"score")
	}

	if(dataType=="gmt"){
		colnames(inputGene) <- c("geneSet", "link", sourceIdType)
		inputGeneL <- unique(inputGene$gene)
	}

	if (startsWith(hostName, "file://")) {
		sourceMap <- read_tsv(
			removeFileProtocol(file.path(hostName, "xref", paste(organism, sourceIdType, "phosphositeSeq.table", sep="_"))),
			col_names=c("phosphositeSeq", "userId"), col_types="cc", quote=""
		) %>% filter(.data$userId %in% inputGeneL)
		if (targetIdType == "phosphositeSeq" || targetIdType == sourceIdType) {
			mappedInputGene <- sourceMap
		} else {
			targetMap <- read_tsv(
				removeFileProtocol(file.path(hostName, "xref", paste(organism, targetIdType, "phosphositeSeq.table", sep="_"))),
				col_names=c("phosphositeSeq", targetIdType), col_types="cc", quote=""
			)
			mappedInputGene <- inner_join(sourceMap, targetMap, by=c("phosphositeSeq"))
		}
		if (nrow(mappedInputGene) == 0) { return(idMappingError("empty")) }
		mappedInputGene <- select(mappedInputGene, .data$userId, targetIdType)
		unmappedIds <- setdiff(inputGeneL, mappedInputGene$userId)
	} else {
		response <- POST(file.path(hostName, "api", "idmapping"), encode="json",
			body=list(organism=organism, sourceType=sourceIdType,
			targetType=targetIdType, ids=inputGeneL, standardId="phosphositeSeq")
		)
		if (response$status_code != 200) {
			stop(webRequestError(response))
		}
		mapRes <- content(response)
		if (mapRes$status == 1) {
			stop(webApiError(mapRes))
		}

		mappedIds <- mapRes$mapped
		unmappedIds <- unlist(mapRes$unmapped)

		if (length(mappedIds) == 0) { stop(idMappingError("empty")) }
		names <- c("sourceId", "targetId")
		mappedInputGene <- data.frame(matrix(unlist(lapply(replace_null(mappedIds), FUN=function(x) { x[names] })), nrow=length(mappedIds), byrow=TRUE), stringsAsFactors=FALSE)
		colnames(mappedInputGene) <- c("userId", targetIdType)
	}

	### Get gene name and symbol in 2nd step, either direct by geneid or mapping to uniprot ambiguously
	# TODO mapping to target other than 15mer may introduce ambiguity, like DTQIKRNtFVGTPFW maps to three STKs in uniprot.
	# Not essential for WG, but could use protein ID to determine
	if (grepl("Uniprot", sourceIdType, fixed=TRUE) || grepl("Ensembl", sourceIdType, fixed=TRUE) || grepl("Refseq", sourceIdType, fixed=TRUE)) { ##if the sourceIdType is Uniprot, Ensembl or Refseq, directly extract the gene level id####
		mappedInputGene$gene <- unlist(lapply(strsplit(mappedInputGene$userId, "_"), .combineG))
	}else{
		###If the input id type is sequence, we will first map the sequence to uniprot. And then map the uniprot to gene name####
		if (targetIdType == "phosphositeUniprot") {
			mappedInputGene$gene <- unlist(lapply(strsplit(mappedInputGene[, targetIdType], "_"), .combineG))
		} else {
			if (startsWith(hostName, "file://")) {
				uniMapRes <- read_tsv(
					removeFileProtocol(file.path(hostName, "xref", paste(organism, "phosphositeUniprot", "phosphositeSeq.table", sep="_"))),
					col_names=c("phosphositeSeq", "phosphositeUniprot"), col_types="cc", quote=""
				) %>% filter(.data$phosphositeSeq %in% mappedInputGene$phosphositeSeq)
			} else {
				response <- POST(file.path(hostName, "api", "idmapping"), encode="json",
					body=list(organism=organism, sourceType="phosphositeSeq", standardId="phosphositeSeq",
					targetType="phosphositeUniprot", ids=inputGeneL)
				)

				if (response$status_code != 200) {
					stop(webRequestError(response))
				}
				uniMapRes <- content(response)
				if (uniMapRes$status == 1) {
					stop(webApiError(uniMapRes))
				}
				if (length(uniMapRes$mapped) == 0) { return(idMappingError("empty")) }

				names <- c("sourceId", "targetId")
				uniMapRes <- data.frame(matrix(unlist(lapply(replace_null(uniMapRes$mapped), FUN=function(x) { x[names] })), nrow=length(uniMapRes$mapped), byrow=TRUE), stringsAsFactors=FALSE)
				colnames(uniMapRes) <- c("phosphositeSeq", "phosphositeUniprot")
			}

			uniMapRes$gene <- unlist(lapply(strsplit(uniMapRes[, "phosphositeUniprot"], "_"), .combineG))
			# Map ID may change nrow due to unmapped ones
			mappedInputGene <- uniMapRes %>% select(.data$phosphositeSeq, .data$gene) %>% right_join(mappedInputGene, by="phosphositeSeq")
		}
	}
	#####Hard code#######
	if (grepl("Uniprot", sourceIdType, fixed=TRUE) || sourceIdType == "phosphositeSeq") {
		geneType <- "uniprotswissprot"
		outLink <- "http://www.uniprot.org/uniprot/"
	}

	if (grepl("Ensembl", sourceIdType, fixed=TRUE)) {
		geneType <- "ensembl_peptide_id"
		outLink <- paste("http://www.ensembl.org/",organism,"/Gene/Summary?db=core;t=",sep="")
	}

	if (grepl("Refseq", sourceIdType, fixed=TRUE)) {
		geneType <- "refseq_peptide"
		outLink <- "https://www.ncbi.nlm.nih.gov/protein/"
	}
	mappedInputGene$gLink <- paste0(outLink, mappedInputGene$gene)

	########Get gene level information#########
	entrezgeneMapRes <- idMappingGene(organism=organism, dataType="list", inputGene=mappedInputGene$gene, sourceIdType=geneType, targetIdType="entrezgene", mappingOutput=FALSE, hostName=hostName)

	mergedRes <- entrezgeneMapRes$mapped %>% select(gene=.data$userId, .data$geneSymbol, .data$geneName) %>%
		right_join(mappedInputGene, by="gene")

	if(dataType=="list"){
		inputGene <- select(mergedRes, .data$userId, .data$geneSymbol, .data$geneName, targetIdType, .data$gLink)
	}

	if(dataType=="rnk"){
		inputGene <- mergedRes %>% left_join(inputGene, by=c("userId"=sourceIdType)) %>%
			select(.data$userId, .data$geneSymbol, .data$geneName, targetIdType, .data$score, .data$gLink)
	}

	if(dataType=="gmt"){
		inputGene <- mergedRes %>% left_join(inputGene, by=c("userId"=sourceIdType)) %>%
			select(.data$geneSet, .data$link, .data$userId, .data$geneSymbol, .data$geneName, targetIdType, .data$gLink)
	}

	#############Output#######################
	if (mappingOutput) {
		idMappingOutput(outputFileName, inputGene, unmappedIds, dataType, sourceIdType, targetIdType=targetIdType)
	}
	r <- list(mapped=inputGene,unmapped=unmappedIds)
	return(r)
}


.combineG <- function(e){
	e <- e[-length(e)]
	e <- paste(e,collapse="_")
	return(e)
}

#' summaryDescription
#'
#' Render job summary section
#'
#' @importFrom whisker whisker.render
#' @keywords internal
#'
summaryDescription <- function(projectName, organism, interestGeneFile, interestGene, interestGeneType, enrichMethod, enrichDatabase, enrichDatabaseFile, enrichDatabaseType, enrichDatabaseDescriptionFile, interestingGeneMap, referenceGeneList, referenceGeneFile, referenceGene, referenceGeneType, referenceSet, minNum, maxNum, sigMethod, fdrThr, topThr, fdrMethod, enrichedSig, reportNum, perNum, p, geneSet, repAdded, numAnnoRefUserId, hostName) {
	if (enrichMethod == "ORA") {
		methodSpecificContent <- specificParameterSummaryOra(organism, referenceGeneList, geneSet, referenceGeneFile, referenceGene, referenceGeneType, referenceSet, minNum, maxNum, sigMethod, fdrThr, topThr, fdrMethod, enrichedSig, reportNum, repAdded, numAnnoRefUserId, interestingGeneMap, hostName)
	}

	if (enrichMethod == "GSEA") {
		methodSpecificContent <- specificParameterSummaryGsea(organism, interestingGeneMap, geneSet, minNum, maxNum, sigMethod, fdrThr, topThr, perNum, p, enrichedSig, reportNum, repAdded)
	}

	template <- readLines(system.file("templates/summary.mustache", package="WebGestaltR"))
	enrichDatabaseInfo <- list()
	for (enrichDb in enrichDatabase) {
		if (!is.null(enrichDb)) {
			enrichDatabaseInfo <- c(enrichDatabaseInfo, list(list(isBuiltIn=TRUE, enrichDatabase=enrichDb)))
		}
	}
	for (i in 1:length(enrichDatabaseFile)) {
		if (!is.null(enrichDatabaseFile[[i]])) {
			enrichDatabaseInfo <- c(enrichDatabaseInfo, list(list(isBuiltIn=FALSE, enrichDatabaseFile=enrichDatabaseFile[[i]],
				hasEnrichDatabaseDescriptionFile=is.null(enrichDatabaseDescriptionFile[[i]]),
				enrichDatabaseDescriptionFile=enrichDatabaseDescriptionFile[[i]])))
		}
	}
	if (organism != "others") {
		standardId <- unname(interestingGeneMap$standardId)
		data <- list(projectName=projectName, enrichMethod=enrichMethod, organism=organism, organismIsOthers=FALSE,
			enrichDatabaseInfo=enrichDatabaseInfo, enrichDatabaseType=enrichDatabaseType,
			hasInterestGeneFile=!is.null(interestGeneFile),
			interestGeneFileBase=ifelse(is.null(interestGeneFile), "", basename(interestGeneFile)), interestGeneType=interestGeneType,
			numUserId=nrow(interestingGeneMap$mapped)+length(interestingGeneMap$unmapped),
			numMappedUserId=nrow(interestingGeneMap$mapped), numUniqueMappedId=length(unique(interestingGeneMap$mapped[[standardId]])),
			numUnmappedUserId=length(interestingGeneMap$unmapped), idIsEntrezGene=standardId=="entrezgene", standardId=standardId,
			methodSpecificContent=methodSpecificContent, hostName=hostName
		)
	}else {
		data <- list(projectName=projectName, enrichMethod=enrichMethod, organism=organism, organismIsOthers=TRUE,
			enrichDatabaseInfo=enrichDatabaseInfo, enrichDatabaseType=enrichDatabaseType,
			hasInterestGeneFile=!is.null(interestGeneFile),
			interestGeneFileBase=ifelse(is.null(interestGeneFile), "", basename(interestGeneFile)), interestGeneType=interestGeneType,
			idIsEntrezGene=FALSE, methodSpecificContent=methodSpecificContent
		)
	}

	return(whisker.render(template, data))
}


#' specificParameterSummaryOra
#'
#' Render job summary section of ORA specific parameters
#'
#' @importFrom whisker whisker.render
#'
#' @keywords internal
#'
specificParameterSummaryOra <- function(organism, referenceGeneList, geneSet, referenceGeneFile, referenceGene, referenceGeneType, referenceSet, minNum, maxNum, sigMethod, fdrThr, topThr, fdrMethod, enrichedSig, reportNum, repAdded, numAnnoRefUserId, interestingGeneMap, hostName) {
	organismIsOthers <- organism == "others"
	if (!organismIsOthers) {
		standardId <- interestingGeneMap$standardId
		interestGeneList <- unique(interestingGeneMap$mapped[[standardId]])
	} else {  ### for others
		standardId <- NULL
		interestGeneList <- unique(interestingGeneMap)
	}
	numAnnoRefId <- length(intersect(referenceGeneList, geneSet$gene))
	hasEnrichedSig <- !is.null(enrichedSig)
	if (hasEnrichedSig) {
		numEnrichedSig <- nrow(enrichedSig)
		showAll <- reportNum >= numEnrichedSig
	} else {
		numEnrichedSig <- NULL
		showAll <- NULL
	}
	data <- list(organismIsOthers=organismIsOthers, numUniqueUserId=length(interestGeneList), standardId=standardId,
		numAnnoRefUserId=numAnnoRefUserId, hasRefGeneFile=!is.null(referenceGeneFile), referenceGeneFile=referenceGeneFile,
		referenceGeneType=referenceGeneType, hasRefGene=!is.null(referenceGene), referenceSet=referenceSet,
		numRefGene=length(referenceGeneList), numAnnoRefId=length(intersect(referenceGeneList, geneSet$gene)), minNum=minNum,
		maxNum=maxNum, fdrMethod=fdrMethod, methodIsFdr=sigMethod=="fdr", methodIsTop=sigMethod=="top", fdrThr=fdrThr,
		topThr=topThr, hasEnrichedSig=hasEnrichedSig, showAll=showAll, reportNum=reportNum, hasRepAdded=repAdded, numEnrichedSig=numEnrichedSig, organism=organism, hostName=hostName
		)
	template <- readLines(system.file("templates/summaryOra.mustache", package="WebGestaltR"))
	return(whisker.render(template, data))
}

#' specificParameterSummaryGsea
#'
#' Render job summary section of GSEA specific parameters
#'
#' @importFrom whisker whisker.render
#' @importFrom dplyr filter
#'
#' @keywords internal
#'
specificParameterSummaryGsea <- function(organism, interestingGeneMap, geneSet, minNum, maxNum, sigMethod, fdrThr, topThr, perNum, p, enrichedSig, reportNum, repAdded) {
	organismIsOthers <- organism == "others"
	if (!organismIsOthers) {
		standardId <- interestingGeneMap$standardId
		interestGeneList <- unique(interestingGeneMap$mapped[[standardId]])
		numUniqueUserId <- length(interestGeneList)
		numAnnoUserId <- length(intersect(interestGeneList, geneSet$gene))
	} else {
		standardId <- NULL
		interestGeneList <- unique(interestingGeneMap)
		numUniqueUserId <- nrow(interestGeneList)
		numAnnoUserId <- NULL
	}

	hasEnrichedSig <- !is.null(enrichedSig)
	data <- list(organismIsOthers=organismIsOthers, numUniqueUserId=numUniqueUserId, standardId=standardId, numAnnoUserId=numAnnoUserId,
		minNum=minNum, maxNum=maxNum, methodIsFdr=sigMethod=="fdr", methodIsTop=sigMethod=="top", fdrThr=fdrThr, topThr=topThr,
		perNum=perNum, reportNum=reportNum, hasRepAdded=repAdded, hasEnrichedSig=hasEnrichedSig, p=p, isPShown=p!=1
		)

	if(hasEnrichedSig){
		data$numPosRel <- nrow(filter(enrichedSig, .data$normalizedEnrichmentScore > 0))
		data$numNegRel <- nrow(filter(enrichedSig, .data$normalizedEnrichmentScore < 0))
		data$isPosRel <- data$numPosRel > 0
		data$isNegRel <- data$numNegRel > 0
		data$showAll <- reportNum >= nrow(enrichedSig)
	}

	template <- readLines(system.file("templates/summaryGsea.mustache", package="WebGestaltR"))
	return(whisker.render(template, data))
}

#' List ID Types
#'
#' List supported ID types for the given organism on WebGestalt server.
#'
#' @inheritParams WebGestaltR
#'
#' @return A list of supported gene sets.
#'
#' @importFrom httr content
#' @export
#' @aliases listIDType
#'
listIdType <- function(organism="hsapiens", hostName="https://www.webgestalt.org/", cache=NULL) {
	if (startsWith(hostName, "file://")) {
        jsonData <- fromJSON(removeFileProtocol(file.path(hostName, "idtypesummary.json")))
		idType <- jsonData[[organism]]
		idType <- idType$name
	} else {
		response <- cacheUrl(file.path(hostName, "api", "summary", "idtype"), cache)
		if (response$status_code != 200) {
			return(webRequestError(response))
		}
		jsonData <- content(response)
        idType <- jsonData[[organism]]
        idType <- sapply(idType,function(e){return(e$name)})
    }
	return(idType)
}

#' @export
listIDType <- function(...) {
	warning("Function listIDType is deprecated and changed to listIdType!\n")
	return(listIdType(...))
}

#' @importFrom dplyr filter select left_join mutate arrange %>%
#' @importFrom stats p.adjust phyper
oraEnrichment <- function(interestGene, referenceGene, geneSet, minNum=10, maxNum=500, fdrMethod="BH", sigMethod="fdr", fdrThr=0.05, topThr=10) {
	#before running this code, the main code has checked the overlap among interestGene, referenceGene and geneSet.
	#And this three sets should have overlapping genes.

	# The calculation is based on the genes (input and reference)
	# with annotations in GMT (i.e. effective genes shown in final HTML report)
	# One common question is why input GMT affects results
	# While GSEA does not have this
	referenceGene <- intersect(referenceGene, geneSet$gene)

	geneSet <- filter(geneSet, .data$gene %in% referenceGene)

	geneSetNum <- tapply(geneSet$gene, geneSet$geneSet,length)
	geneSetNum <- geneSetNum[geneSetNum>=minNum & geneSetNum<=maxNum]
	if (length(geneSetNum) == 0) {
		stop("ERROR: The number of annotated genes for all functional categories are not from ", minNum, " to ", maxNum, " for the ORA enrichment method.")
	}

	interestGene <- intersect(interestGene, geneSet$gene)
	interestGene <- intersect(interestGene, referenceGene)
	if (length(interestGene) == 0) {
		stop("ERROR: No genes in the interesting list can annotate to any functional category.")
	}

	###############Enrichment analysis###################
	ra <- length(interestGene) / length(referenceGene) # ratio
	refG <- data.frame(geneSet=names(geneSetNum), size=as.numeric(geneSetNum), stringsAsFactors=FALSE)

	intG <- filter(geneSet, .data$gene %in% interestGene)
	intGNum <- tapply(intG$gene, intG$geneSet, length) # a vector of overlap with geneset as name
	intGNum <- data.frame(geneSet=names(intGNum), overlap=as.numeric(intGNum), stringsAsFactors=FALSE)

	intGId <- tapply(intG$gene, intG$geneSet, paste, collapse=";")
	intGId <- data.frame(geneSet=names(intGId), overlapId=as.character(intGId), stringsAsFactors=FALSE)

	# .data from rlang needed for package to pass R CMD check
	# https://cran.r-project.org/web/packages/dplyr/vignettes/programming.html
	enrichedResult <- geneSet %>% filter(!is.na(.data$geneSet)) %>%
		filter(.data$geneSet %in% names(geneSetNum)) %>%
		select(.data$geneSet, link=.data$description) %>% distinct() %>%
		left_join(refG, by="geneSet") %>%
		left_join(intGNum, by="geneSet") %>% # this may just be inner_join. O is NA should not be meaningful anyway
		mutate(overlap=ifelse(is.na(.data$overlap), 0, .data$overlap), expect=.data$size * ra, enrichmentRatio=.data$overlap /.data$expect,
			   pValue=1-phyper(.data$overlap - 1, length(interestGene), length(referenceGene) - length(interestGene), .data$size, lower.tail=TRUE, log.p=FALSE),
			   FDR=p.adjust(.data$pValue, method=fdrMethod)
		) %>%
		left_join(intGId, by="geneSet") %>% # get overlapping gene IDs
		arrange(.data$FDR, .data$pValue, .data$enrichmentRatio)

	if (sigMethod == "fdr") {
		enrichedResultSig <- filter(enrichedResult, .data$FDR<fdrThr)
		if (nrow(enrichedResultSig) == 0) {
			warning("No significant gene set is identified based on FDR ", fdrThr, "!")
			return(NULL)
		} else {
			enrichedResultInsig <- enrichedResult %>% filter(.data$FDR >= fdrThr, .data$overlap != 0) %>% select(.data$geneSet, .data$enrichmentRatio, .data$FDR, .data$overlap)
			return(list(enriched=enrichedResultSig, background=enrichedResultInsig))
		}
	} else {
		#for the top method, we only select the terms with at least one annotated interesting gene
		enrichedResult <- enrichedResult %>% filter(.data$overlap != 0)
		if (nrow(enrichedResult)>topThr) {
			enrichedResultSig <- enrichedResult[1:topThr, ]
			enrichedResultInsig <- enrichedResult[(topThr+1):nrow(enrichedResult), c("geneSet", "enrichmentRatio", "FDR", "overlap")]
		}else{
			enrichedResultSig <- enrichedResult
			enrichedResultInsig <- data.frame()
		}
		return(list(enriched=enrichedResultSig, background=enrichedResultInsig))
	}
}

#' @importFrom httr POST content
#' @importFrom readr read_tsv cols
#' @importFrom jsonlite toJSON
WebGestaltRNta <- function(organism="hsapiens", network="network_PPI_BIOGRID", method="Network_Retrieval_Prioritization", inputSeed, inputSeedFile, interestGeneType="genesymbol", neighborNum=10, highlightSeedNum=10, sigMethod="fdr", fdrThr=0.05, topThr=10, highlightType="Seeds", outputDirectory=getwd(), projectName=NULL, cache=NULL, hostName="https://www.webgestalt.org/") {
	projectDir <- file.path(outputDirectory, paste0("Project_", projectName))
	dir.create(projectDir)

	if (length(network) > 1) {
		stop("NTA does not support multiple databases.")
	}
	inputGene <- formatCheck("list", inputGeneFile=inputSeedFile, inputGene=inputSeed)
	# only networks are in gene symbols
	# mapping always returns gene symbol, could map to genesymbol but takes two requests
	inputGene <- idMappingGene(organism=organism, dataType="list", inputGene=inputGene, sourceIdType=interestGeneType, targetIdType="entrezgene", mappingOutput=FALSE, hostName=hostName)
	inputGene <- inputGene$mapped$geneSymbol

	if (startsWith(hostName, "file://")) {
		dagInfo <- read_tsv(
			removeFileProtocol(file.path(hostName, "geneset", paste(organism, "geneontology_Biological_Process", "entrezgene.dag", sep="_"))),
			col_names=c("source", "target"), col_types="cc"
		)
	} else {
		geneSetUrl <- file.path(hostName, "api", "geneset")
		response <- cacheUrl(geneSetUrl, cache=cache, query=list(organism=organism, database="geneontology_Biological_Process", standardId="entrezgene", fileType="dag"))
		dagInfo <- read_tsv(content(response), col_names=c("source", "target"), col_types="cc")
	}

	## networks <- unlist(strsplit(network, ",", fixed=TRUE))
	## May need to bring back analysis of multiple networks
	fileName <- paste(projectName, network, method, sep=".")
	goEnrichRes <- randomWalkEnrichment(organism=organism, network=network, method=method, highlightSeedNum=highlightSeedNum, inputSeed=inputGene,
						sigMethod=sigMethod, fdrThr=fdrThr, topThr=topThr, projectDir=projectDir,
						topRank=neighborNum, projectName=projectName, cache=cache, hostName=hostName)
	if (is.null(goEnrichRes)) {
		return(NULL)
	}
	enrichResFile <- file.path(projectDir, paste0(fileName, "_enrichedResult.txt"))

	goTermList <- read_tsv(enrichResFile, col_types=cols())$goId
	inputEndIndex <- length(goTermList)

	dagTree <- expandDag(goTermList, dagInfo)
	goTermList <- dagTree$allNodes
	edges <- dagTree$edges
	rm(dagTree)

	if (startsWith(hostName, "file://")) {
		goId2Term <- read_tsv(
			removeFileProtocol(file.path(hostName, "geneset", paste(organism, "geneontology_Biological_Process", "entrezgene.des", sep="_"))),
			col_names=c("id", "name"), col_types="cc"
		)
	} else {
		response <- POST(geneSetUrl, body=list(organism=organism, database="geneontology_Biological_Process",
			fileType="des", ids=goTermList), encode="json")
		goId2Term <- read_tsv(content(response), col_names=c("id", "name"), col_types="cc")
	}

	jsonFile <- file.path(projectDir, paste0(fileName, ".json"));
	jsonData <- vector(mode="list", length=length(goTermList))

	for (i in 1:length(goTermList)) {
		goId <- goTermList[[i]]
		goName <- filter(goId2Term, .data$id == goId)[[1, "name"]]
		dataSets <- i <= inputEndIndex
		jsonData[[i]] <- list(data=list(id=goId, name=goName, datasets=dataSets))
	}
	jsonData <- unname(c(jsonData, edges))

	cat(toJSON(jsonData, auto_unbox=TRUE), "\n", sep="", file=jsonFile)

	createNtaReport(networkName=network, method=method, sigMethod=sigMethod, fdrThr=fdrThr, topThr=topThr,
					highlightType=highlightType, outputDirectory=outputDirectory, projectDir=projectDir,
					projectName=projectName, hostName=hostName)

	cwd <- getwd()
	setwd(projectDir)
	zip(paste0(projectName, ".zip"), ".", flags="-rq")
	setwd(cwd)

	cat("Results can be found in the ", projectDir, "!\n", sep="")
	return(goEnrichRes)
}

#' WebGestaltR: The R interface for enrichment analysis with WebGestalt.
#'
#' @docType package
#' @name WebGestaltR
#' @import methods
#' @import grDevices
#' @import graphics
#' @import utils
#' @importFrom Rcpp sourceCpp
#' @importFrom rlang .data
#' @useDynLib WebGestaltR
#'
NULL


.onAttach <- function(lib, pkg) {
	packageStartupMessage("******************************************\n")
	packageStartupMessage("*                                        *\n")
	packageStartupMessage("*          Welcome to WebGestaltR !      *\n")
	packageStartupMessage("*                                        *\n")
	packageStartupMessage("******************************************\n")
}

#' @importFrom readr write_tsv
#' @importFrom dplyr left_join select arrange %>% desc mutate
WebGestaltROra <- function(organism="hsapiens", enrichDatabase=NULL, enrichDatabaseFile=NULL, enrichDatabaseType=NULL, enrichDatabaseDescriptionFile=NULL,  interestGeneFile=NULL, interestGene=NULL, interestGeneType=NULL, collapseMethod="mean", referenceGeneFile=NULL, referenceGene=NULL, referenceGeneType=NULL, referenceSet=NULL, minNum=10, maxNum=500, fdrMethod="BH", sigMethod="fdr", fdrThr=0.05, topThr=10, reportNum=20, setCoverNum=10, isOutput=TRUE, outputDirectory=getwd(), projectName=NULL, dagColor="binary", nThreads=1, cache=NULL, hostName="https://www.webgestalt.org/") {
	enrichMethod <- "ORA"
	projectDir <- file.path(outputDirectory, paste0("Project_", projectName))

	######### Web server will input "NULL" to the R package, thus, we need to change "NULL" to NULL ########
	enrichDatabase <- testNull(enrichDatabase)
	enrichDatabaseFile <- testNull(enrichDatabaseFile)
	enrichDatabaseType <- testNull(enrichDatabaseType)
	enrichDatabaseDescriptionFile <- testNull(enrichDatabaseDescriptionFile)
	interestGeneFile <- testNull(interestGeneFile)
	interestGene <- testNull(interestGene)
	interestGeneType <- testNull(interestGeneType)
	referenceGeneFile <- testNull(referenceGeneFile)
	referenceGene <- testNull(referenceGene)
	referenceGeneType <- testNull(referenceGeneType)
	referenceSet <- testNull(referenceSet)

	################ Check parameter ################
	errorTest <- parameterErrorMessage(enrichMethod=enrichMethod, organism=organism, collapseMethod=collapseMethod, minNum=minNum, maxNum=maxNum, fdrMethod=fdrMethod, sigMethod=sigMethod, fdrThr=fdrThr, topThr=topThr, reportNum=reportNum, isOutput=isOutput, outputDirectory=outputDirectory, dagColor=dagColor, hostName=hostName, cache=cache)

	if (!is.null(errorTest)) {
		stop(errorTest)
	}

	############# Check enriched database #############
	cat("Loading the functional categories...\n")
	enrichD <- loadGeneSet(organism=organism, enrichDatabase=enrichDatabase, enrichDatabaseFile=enrichDatabaseFile, enrichDatabaseType=enrichDatabaseType, enrichDatabaseDescriptionFile=enrichDatabaseDescriptionFile, cache=cache, hostName=hostName)

	geneSet <- enrichD$geneSet
	geneSetDes <- enrichD$geneSetDes
	geneSetDag <- enrichD$geneSetDag
	geneSetNet <- enrichD$geneSetNet
	databaseStandardId <- enrichD$standardId
	rm(enrichD)

	########### Check input interesting gene list ###############
	cat("Loading the ID list...\n")
	interestingGeneMap <- loadInterestGene(organism=organism, dataType="list", inputGeneFile=interestGeneFile, inputGene=interestGene, geneType=interestGeneType, collapseMethod=collapseMethod, cache=cache, hostName=hostName, geneSet=geneSet)

	if (organism == "others") {
		interestGeneList <- unique(interestingGeneMap)
	} else {
		interestStandardId <- interestingGeneMap$standardId
		interestGeneList <- unique(interestingGeneMap$mapped[[interestStandardId]])
	}

	################### Load reference gene set ##############
	cat("Loading the reference list...\n")
	referenceGeneList <- loadReferenceGene(organism=organism, referenceGeneFile=referenceGeneFile, referenceGene=referenceGene, referenceGeneType=referenceGeneType, referenceSet=referenceSet, collapseMethod=collapseMethod, hostName=hostName, geneSet=geneSet, interestGeneList=interestGeneList, cache=cache)

	########## Create project folder ##############
	if (isOutput) {
		dir.create(projectDir)

	###### Summarize gene annotation based on the GOSlim ###########
		if (organism != "others") {
			if (databaseStandardId == "entrezgene") {
				cat("Summarizing the input ID list by GO Slim data...\n")
				goSlimOutput <- file.path(projectDir, paste0("goslim_summary_", projectName))
				re <- goSlimSummary(organism=organism, geneList=interestGeneList, outputFile=goSlimOutput, outputType="png", isOutput=isOutput, cache=cache, hostName=hostName)
			}
			write_tsv(interestingGeneMap$mapped, file.path(projectDir, paste0("interestingID_mappingTable_", projectName , ".txt")))
			write(interestingGeneMap$unmapped, file.path(projectDir, paste0("interestingID_unmappedList_", projectName, ".txt")))
		} else {
			write(interestGeneList, file.path(projectDir, paste0("interestList_", projectName, ".txt")))
		}
	}

	############# Run enrichment analysis ###################
	cat("Performing the enrichment analysis...\n")
	oraRes <- oraEnrichment(interestGeneList, referenceGeneList, geneSet, minNum=minNum, maxNum=maxNum, fdrMethod=fdrMethod, sigMethod=sigMethod, fdrThr=fdrThr, topThr=topThr)
	if (is.null(oraRes)) {
		return(NULL)
	}
	enrichedSig <- oraRes$enriched
	insig <- oraRes$background

	clusters <- list()
	geneTables <- list()

	if (!is.null(enrichedSig)) {
		if (!is.null(geneSetDes)) { ####### Add extra description information ###########
			enrichedSig <- enrichedSig %>%
				left_join(geneSetDes, by="geneSet") %>%
				select(.data$geneSet, .data$description, .data$link, .data$size, .data$overlap, .data$expect, .data$enrichmentRatio, .data$pValue, .data$FDR, .data$overlapId) %>%
				arrange(.data$FDR, .data$pValue, desc(.data$size)) %>%
				mutate(description=ifelse(is.na(.data$description), "", .data$description)) # now des could be mixture
		} else {
			enrichedSig <- enrichedSig %>%
				select(.data$geneSet, .data$link, .data$size, .data$overlap, .data$expect, .data$enrichmentRatio, .data$pValue, .data$FDR, .data$overlapId) %>%
				arrange(.data$FDR, .data$pValue, desc(.data$size))
		}

		geneTables <- getGeneTables(organism, enrichedSig, "overlapId", interestingGeneMap)
		if (organism != "others") {
			enrichedSig$link <- mapply(function(link, geneList) linkModification("ORA", link, geneList, interestingGeneMap),
				enrichedSig$link,
				enrichedSig$overlapId
			)
		}

		if ("database" %in% colnames(geneSet)) {
			# Add source database for multiple databases
			enrichedSig <- enrichedSig %>% left_join(unique(geneSet[, c("geneSet", "database")]), by="geneSet")
		}

		if (organism != "others" && interestGeneType != interestStandardId) {
			outputEnrichedSig <- mapUserId(enrichedSig, "overlapId", interestingGeneMap)
		} else {
			outputEnrichedSig <- enrichedSig
		}

		if (isOutput) {
			write_tsv(outputEnrichedSig, file.path(projectDir, paste0("enrichment_results_", projectName, ".txt")))
			idsInSet <- sapply(enrichedSig$overlapId, strsplit, split=";")
			names(idsInSet) <- enrichedSig$geneSet
			minusLogP <- -log(enrichedSig$pValue)
			minusLogP[minusLogP == Inf] <- -log(.Machine$double.eps)
			apRes <- affinityPropagation(idsInSet, minusLogP)
			wscRes <- weightedSetCover(idsInSet, 1 / minusLogP, setCoverNum, nThreads)
			if (!is.null(apRes)) {
				writeLines(sapply(apRes$clusters, paste, collapse="\t"), file.path(projectDir, paste0("enriched_geneset_ap_clusters_", projectName, ".txt")))
			} else {
				apRes <- NULL
			}
			clusters$ap <- apRes
			if (!is.null(wscRes$topSets)) {
				writeLines(c(paste0("# Coverage: ", wscRes$coverage), wscRes$topSets), file.path(projectDir, paste0("enriched_geneset_wsc_topsets_", projectName, ".txt")))
				clusters$wsc <- list(representatives=wscRes$topSets,  coverage=wscRes$coverage)
			} else {
				clusters$wsc <- NULL
			}
		}
	}

	if (isOutput) {
	############## Create report ##################
		cat("Generate the final report...\n")
		createReport(hostName=hostName, outputDirectory=outputDirectory, organism=organism, projectName=projectName, enrichMethod=enrichMethod, geneSet=geneSet, geneSetDes=geneSetDes, geneSetDag=geneSetDag, geneSetNet=geneSetNet, interestingGeneMap=interestingGeneMap, referenceGeneList=referenceGeneList, enrichedSig=enrichedSig, background=insig, geneTables=geneTables, clusters=clusters, enrichDatabase=enrichDatabase, enrichDatabaseFile=enrichDatabaseFile, enrichDatabaseType=enrichDatabaseType, enrichDatabaseDescriptionFile=enrichDatabaseDescriptionFile, interestGeneFile=interestGeneFile, interestGene=interestGene, interestGeneType=interestGeneType, collapseMethod=collapseMethod, referenceGeneFile=referenceGeneFile, referenceGene=referenceGene, referenceGeneType=referenceGeneType, referenceSet=referenceSet, minNum=minNum, maxNum=maxNum, fdrMethod=fdrMethod, sigMethod=sigMethod, fdrThr=fdrThr, topThr=topThr, reportNum=reportNum, dagColor=dagColor)

		cwd <- getwd()
		setwd(projectDir)
		zip(paste0("Project_", projectName, ".zip"), ".", flags="-rq")
		setwd(cwd)

		cat("Results can be found in the ", projectDir, "!\n", sep="")
	}

	return(outputEnrichedSig)
}

#' @importFrom httr content modify_url
#' @importFrom readr read_tsv
#' @importFrom dplyr filter arrange %>% desc
#' @importFrom igraph graph.edgelist V
randomWalkEnrichment <- function(organism, network, method, inputSeed, topRank, highlightSeedNum, sigMethod, fdrThr, topThr, projectDir, projectName, cache, hostName) {
	fileName <- paste(projectName, network, method, sep=".")
	if (startsWith(hostName, "file://")) {
		net <- as.matrix(read_tsv(
			removeFileProtocol(file.path(hostName, "geneset", paste(organism, network, "entrezgene.net", sep="_"))),
			col_names=FALSE, col_types="cc"))
		goAnn <- readGmt(removeFileProtocol(file.path(hostName, "geneset", paste(organism, "geneontology_Biological_Process", "genesymbol.gmt", sep="_"))))
	} else {
		geneSetUrl <- file.path(hostName, "api", "geneset")
		# actually standard id is gene symbol for network
		response <- cacheUrl(geneSetUrl, cache=cache, query=list(organism=organism, database=network, standardId="entrezgene", fileType="net"))
		net <- as.matrix(read_tsv(content(response), col_names=FALSE, col_types="cc"))
		gmtUrl <- modify_url(geneSetUrl, query=list(organism=organism, database="geneontology_Biological_Process", standardId="genesymbol", fileType="gmt"))
		goAnn <- readGmt(gmtUrl, cache=cache)
	}
	netGraph <- graph.edgelist(net, directed=FALSE)
	netNode <- V(netGraph)$name


	cat("Start Random Walk...\n")

	#seeds <- unlist(strsplit(inputSeed, ","))
	seeds <- inputSeed

	allNum <- length(seeds)
	write(seeds, file.path(projectDir, paste0(fileName, "_seeds.txt")))
	seeds <- intersect(netNode, seeds)
	write(seeds, file.path(projectDir, paste0(fileName, "_seedsInNetwork.txt")))

	if (method == "Network_Retrieval_Prioritization") {
		if (length(seeds) < highlightSeedNum) {
			highlightSeedNum <- length(seeds)
		}
	}

	pt1 <- .netwalker(seeds, netGraph, r=0.5)
	gS <- data.frame(name=netNode, score=pt1, per=1, stringsAsFactors=F)
	if (method == "Network_Expansion") {
		gS <- gS %>% filter(.data$name %in% setdiff(.data$name, seeds)) %>%
			arrange(desc(.data$score))
		candidate <- gS[1:topRank, ]
		allN <- c(seeds, candidate$name)
	} else {
		gS <- gS %>% filter(.data$name %in% seeds) %>%
			arrange(desc(.data$score))
		highSeeds <- gS[1:highlightSeedNum, "name"]
		allN <- seeds
		candidate <- gS
	}

	subNet <- net[net[, 1] %in% allN & net[, 2] %in% allN, ]
	allN <- union(subNet[, 1], subNet[, 2])

	if (length(allN) != 0){
		termInfo <- .enrichmentFunction(organism, netNode, allN, goAnn, seeds, sigMethod, fdrThr, topThr, hostName)
	} else {
		warning("Error: No sub-network is generated.")
		return(NULL)
	}

	cat("Output\n")
	if (method == "Network_Retrieval_Prioritization") {
		write(highSeeds, file.path(projectDir, paste0(fileName, "_highlightedSeeds.txt")))
	}
	overlapSeeds <- intersect(allN, seeds)
	x <- c(paste("Total number of genes in the selected network:", length(netNode), "(used for the enrichment analysis)"),
		paste("Total number of seeds:", allNum),
		paste("Total number of seeds in the selected network:", length(seeds))
		)
	if (method=="Network_Expansion") {
		x <- c(x,
			paste("Total number of genes in the expanded sub-network:", length(allN), "(used for the enrichment analysis)"),
			paste("Total number of seeds in the expanded sub-network:", length(overlapSeeds)),
			paste("We select top", topRank, "neighbors based on the probability of random walk method. All seeds and top ranking neighbors in the expanded sub-network can enrich to", nrow(termInfo), "GO BP categories.")
			)
	} else {
		x <- c(x,
			paste("Total number of seeds in the retrieved sub-network:", length(allN), "(used for the enrichment analysis)"),
			paste("All seeds in the retrieved sub-network can enrich to", nrow(termInfo), "GO BP categories.")
			)
	}

	write(x, file.path(projectDir, paste0(fileName, "_resultSummary.txt")))
	write(overlapSeeds, file.path(projectDir, paste0(fileName, "_seedsInSubnetwork.txt")))
	write_tsv(as.data.frame(subNet), file.path(projectDir, paste0(fileName, "_randomWalkNetwork.txt")), col_names=FALSE)
	write_tsv(candidate, file.path(projectDir, paste0(fileName, "_candidate.txt")), col_names=FALSE)
	if (!is.null(termInfo)) {
		write_tsv(termInfo, file.path(projectDir, paste0(fileName, "_enrichedResult.txt")))
	}
	return(termInfo)
}

#' @importFrom igraph get.adjacency
.netwalker <- function(seed, network, r=0.5) {
	adjMatrix <- get.adjacency(network, sparse=FALSE)
	de <- apply(adjMatrix, 2, sum)
	w <- t(t(adjMatrix)/de)

	p0 <- array(0, dim=c(nrow(w), 1))
	rownames(p0) <- rownames(w)
	p0[seed, 1] <- 1/length(seed)

	pt <- p0
	pt1 <- (1-r)*(w%*%pt)+r*p0

	while (sum(abs(pt1-pt)) > 1e-6) {
		pt <- pt1
		pt1 <- (1-r)*(w%*%pt)+r*p0
	}

	return(pt1)
}

#' @importFrom dplyr select filter arrange left_join mutate %>%
#' @importFrom httr POST content
#' @importFrom readr read_tsv
#' @importFrom stats p.adjust phyper
.enrichmentFunction <- function(organism, reference, interest, goAnn, seeds, sigMethod, fdrThr, topThr, hostName) {
	goAnn <- select(goAnn, .data$gene, .data$geneSet)
	annRef <- filter(goAnn, .data$gene %in% reference)
	annInterest <- filter(goAnn, .data$gene %in% interest)

	allRefNum <- length(unique(annRef$gene))
	allInterestNum <- length(unique(annInterest$gene))

	refTermCount <- tapply(annRef$gene, annRef$geneSet, length)

	if (startsWith(hostName, "file://")) {
		refTermName <- read_tsv(
			removeFileProtocol(file.path(hostName, "geneset", paste(organism, "geneontology_Biological_Process", "entrezgene.des", sep="_"))),
			col_names=c("id", "description"), col_types="cc"
		) %>% filter(.data$id %in% names(refTermCount))
	} else {
		geneSetUrl <- file.path(hostName, "api", "geneset")
		response <- POST(geneSetUrl, body=list(organism=organism, database="geneontology_Biological_Process",
			fileType="des", ids=unique(annRef$geneSet)), encode="json")
		refTermName <- read_tsv(content(response), col_names=c("id", "description"), col_types="cc") %>%
			filter(.data$id %in% names(refTermCount))
	}

	refTermCount <- data.frame(goId=names(refTermCount), size=as.numeric(refTermCount), stringsAsFactors=FALSE) %>%
		left_join(refTermName, by=c("goId"="id")) %>%
		select(.data$goId, .data$description, .data$size) %>%
		arrange(.data$goId)
	interestTermCount <- tapply(annInterest$gene, annInterest$geneSet, length)

	interestTermGene <- tapply(annInterest$gene, annInterest$geneSet, .getGenes, seeds)


	interestTermCount <- data.frame(goId=names(interestTermCount), overlap=as.numeric(interestTermCount), interestGene=interestTermGene, stringsAsFactors=FALSE) %>%
		arrange(.data$goId)

	refInterestTermCount <- refTermCount

	refInterestTermCount$overlap <- vector("numeric", nrow(refInterestTermCount))
	refInterestTermCount[refInterestTermCount$goId %in% interestTermCount$goId, "overlap"] <- interestTermCount$overlap

	refInterestTermCount$interestGene <- vector("numeric", nrow(refInterestTermCount))
	refInterestTermCount[refInterestTermCount$goId %in% interestTermCount$goId, "interestGene"] <- interestTermCount$interestGene

	refInterestTermCount <- refInterestTermCount %>%
		mutate(expect = (allInterestNum / allRefNum) * .data$size,
			enrichmentRatio = .data$overlap / .data$expect,
			pValue = 1 - phyper(.data$overlap - 1, allInterestNum, allRefNum - allInterestNum, .data$size, lower.tail=TRUE, log.p=FALSE),
			FDR = p.adjust(.data$pValue, method="BH")
		) %>%
		arrange(.data$FDR, .data$pValue)


	if(sigMethod=="fdr"){
		refInterestTermCountSig <- filter(refInterestTermCount, .data$FDR < fdrThr)
	}else{
		refInterestTermCountSig <- refInterestTermCount[1:topThr, ]
	}

	return(refInterestTermCountSig)
}


.getGenes <- function(genelist, seeds) {
	genelist <- data.frame(gene=genelist, id=0, stringsAsFactors=F)
	genelist[genelist[, 1] %in% seeds, 2] <- 1
	genelist <- paste(genelist[, 1], genelist[, 2], sep="|")
	genelist <- paste(genelist, collapse=";")
	return(genelist)
}

#' @importFrom dplyr select distinct filter arrange mutate left_join %>%
#' @importFrom readr write_tsv
gseaEnrichment <- function (hostName, outputDirectory, projectName, geneRankList, geneSet, geneSetDes=NULL, collapseMethod="mean", minNum=10, maxNum=500, sigMethod="fdr", fdrThr=0.05, topThr=10, perNum=1000, p=1, isOutput=TRUE, saveRawGseaResult=FALSE, plotFormat="png", nThreads=1) {
	projectFolder <- file.path(outputDirectory, paste("Project_", projectName, sep=""))
	if (!dir.exists(projectFolder)) {
		dir.create(projectFolder)
	}

	colnames(geneRankList) <- c("gene", "score")
	sortedScores <- sort(geneRankList$score, decreasing=TRUE)

	geneSetName <- geneSet %>% select(.data$geneSet, link=.data$description) %>% distinct()
	effectiveGeneSet <- geneSet %>% filter(.data$gene %in% geneRankList$gene)

	geneSetNum <- tapply(effectiveGeneSet$gene, effectiveGeneSet$geneSet, length)
	geneSetNum <- geneSetNum[geneSetNum>=minNum & geneSetNum<=maxNum]
	if (length(geneSetNum)==0) {
		stop("ERROR: The number of annotated IDs for all functional categories are not from ", minNum," to ", maxNum, " for the GSEA enrichment method.")
	}

	# collapse rank list
	a <- tapply(geneRankList$score, geneRankList$gene, collapseMethod, na.rm=TRUE)
	geneRankList <- data.frame(gene=names(a), score=as.numeric(a), stringsAsFactors=FALSE)

	gseaRnk <- file.path(projectFolder, paste("Project_", projectName, "_GSEA.rnk", sep=""))
	write_tsv(geneRankList, gseaRnk, col_names=FALSE)

	outputF <- file.path(projectFolder, paste0("Project_", projectName, "_GSEA/"))
	relativeF <- file.path(".", paste0("Project_", projectName, "_GSEA"))
	if (!dir.exists(outputF) && isOutput) {
		dir.create(outputF)
	}

	inputDf <- prepareInputMatrixGsea(geneRankList, effectiveGeneSet)

	gseaRes <- swGsea(inputDf, thresh_type="val", perms=perNum,
		min_set_size=minNum, max_set_size=maxNum, p=p,
		nThreads=nThreads, rng_seed=as.integer(format(Sys.time(), "%H%M%S"))
	)
	if (saveRawGseaResult) {
		saveRDS(gseaRes, file=file.path(outputF, "rawGseaResult.rds"))
	}

	enrichRes <- gseaRes$Enrichment_Results %>%
		mutate(geneSet = rownames(gseaRes$Enrichment_Results)) %>%
		select(.data$geneSet, enrichmentScore=.data$ES, normalizedEnrichmentScore=.data$NES, pValue=.data$p_val, FDR=.data$fdr)
	# TODO: handle errors

	if (sigMethod == "fdr") {
		sig <- filter(enrichRes, .data$FDR < fdrThr)
		insig <- filter(enrichRes, .data$FDR >= fdrThr)
	} else if (sigMethod == "top") {
		enrichRes <- arrange(enrichRes, .data$FDR, .data$pValue)
		tmpRes <- getTopGseaResults(enrichRes, topThr)
		sig <- tmpRes[[1]]
		insig <-tmpRes[[2]]
	}
	numSig <- nrow(sig)
	if (numSig == 0) {
		warning("ERROR: No significant set is identified based on FDR ", fdrThr, "!\n")
		return(NULL)
	}

	if (!is.null(insig)) {
		insig$leadingEdgeNum <- unname(sapply(insig$geneSet, function(geneSet) {
			rsum <- gseaRes$Running_Sums[, geneSet] # Running sum is a matrix of gene by gene set
			maxPeak <- max(rsum)
			minPeak <- min(rsum)
			if (abs(maxPeak) >= abs(minPeak)) {
				peakIndex <- match(max(rsum), rsum)
				leadingEdgeNum <- sum(gseaRes$Items_in_Set[[geneSet]]$rank <= peakIndex)
			} else {
				peakIndex <- match(min(rsum), rsum)
				leadingEdgeNum <- sum(gseaRes$Items_in_Set[[geneSet]]$rank >= peakIndex)
			}
			return(leadingEdgeNum)
		}))
	}
	plotSuffix <- ifelse("png" %in% plotFormat, "png", "svg")
	sig <- sig %>% left_join(geneSetName, by="geneSet") %>%
		mutate(size = unname(sapply(geneSet, function(x) nrow(gseaRes$Items_in_Set[[x]])))) %>%
		mutate(plotPath = unname(sapply(geneSet, function(x) file.path(relativeF, paste0(sanitizeFileName(x), ".", plotSuffix)))))

	leadingGeneNum <- vector("integer", numSig)
	leadingGenes <- vector("character", numSig)
	for (i in 1:numSig) {
		geneSet <- sig[[i, "geneSet"]]
		es <- sig[[i, "enrichmentScore"]]
		genes <- gseaRes$Items_in_Set[[geneSet]] # rowname is gene and one column called rank
		rsum <- gseaRes$Running_Sums[, geneSet]
		peakIndex <- match(ifelse(es > 0, max(rsum), min(rsum)), rsum)
		if (es > 0) {
			indexes <- genes$rank <= peakIndex
		} else {
			indexes <- genes$rank >= peakIndex
		}
		leadingGeneNum[[i]] <- sum(indexes)
		leadingGenes[[i]] <- paste(rownames(genes)[indexes], collapse=";")

		if (isOutput) {
			# Plot GSEA-like enrichment plot
			if (!is.null(geneSetDes)) {
				# same name of variable and column name, use quasiquotation !!
				title <- as.character((geneSetDes %>% filter(.data$geneSet == !!geneSet))[1, "description"])
			} else {
				title <- geneSet
			}

			if (!is.vector(plotFormat)) {
				plotEnrichmentPlot(title, outputF, geneSet, format=plotFormat, gseaRes$Running_Sums[, geneSet], genes$rank, sortedScores, peakIndex)
			} else {
				for (format in plotFormat) {
					plotEnrichmentPlot(title, outputF, geneSet, format=format, gseaRes$Running_Sums[, geneSet], genes$rank, sortedScores, peakIndex)
				}
			}
		}
	}
	sig$leadingEdgeNum <- leadingGeneNum
	sig$leadingEdgeId <- leadingGenes

	return(list(enriched=sig, background=insig))
}

#' @importFrom svglite svglite
plotEnrichmentPlot <- function(title, outputDir, fileName, format="png", runningSums, ranks, scores, peakIndex) {
	if (format == "png") {
		png(file.path(outputDir, paste0(sanitizeFileName(fileName), ".png")), bg="transparent", width=2000, height=2000)
		cex <- list(main=5, axis=2.5, lab=3.2)
	} else if (format == "svg") {
		svglite(file.path(outputDir, paste0(sanitizeFileName(fileName), ".svg")), bg="transparent", width=7, height=7)
		cex <- list(main=1.5, axis=0.6, lab=0.8)
		# svg seems to have a problem with long title (figure margins too large)
		if (nchar(title) > 80) {
			title = paste0(substr(title, 1, 80), "...")
		}
	}
	wrappedTitle <- strwrap(paste0("Enrichment plot: ", title), 60)
	plot.new()
	par(fig=c(0, 1, 0.5, 1), mar=c(0, 6, 6 * length(wrappedTitle), 2), cex.axis=cex$axis, cex.main=cex$main, cex.lab=cex$lab, lwd=2, new=TRUE)
	plot(1:length(runningSums), runningSums,
		type="l", main=paste(wrappedTitle, collapse="\n"),
		xlab="", ylab="Enrichment Score", xaxt='n', lwd=3)
	abline(v=peakIndex, lty=3)
	par(fig=c(0, 1, 0.35, 0.5), mar=c(0, 6, 0, 2), new=TRUE)
	plot(ranks, rep(1, length(ranks)), type="h",
		xlim=c(1, length(scores)), ylim=c(0, 1), axes=FALSE, ann=FALSE)
	par(fig=c(0, 1, 0, 0.35), mar=c(6, 6, 0, 2), cex.axis=cex$axis, cex.lab=cex$lab, new=TRUE)
	# use polygon to greatly reduce file size of SVG
	plot(1:length(scores), scores, type="n",
		ylab="Ranked list metric", xlab="Rank in Ordered Dataset")
	polygon(c(1, 1:length(scores), length(scores)), c(0, scores, 0), col="black")
	abline(v=peakIndex, lty=3)
	dev.off()
}

#' GO Slim Summary
#'
#' Outputs a brief summary of input genes based on GO Slim data.
#'
#' @inheritParams WebGestaltR
#' @param geneList A list of input genes.
#' @param outputFile Output file name.
#' @param isOutput Boolean if a plot is save to \code{outputFile}.
#' @param outputType File format of the plot: \code{pdf}, \code{bmp} or \code{png}.
#'
#' @return A list of the summary result.
#' @export
#' @aliases GOSlimSummary
#'
goSlimSummary <- function(organism="hsapiens", geneList, outputFile, outputType="pdf", isOutput=TRUE, cache=NULL, hostName="http://www.webgestalt.org") {
	organisms <- listOrganism(hostName=hostName, cache=cache)
	if (!organism %in% organisms) {
		stop("ERROR: ", organism, " can not be supported.")
	}

	if (!is.vector(geneList)) {
		stop("ERROR: Please upload a list of entrez gene ids.")
	} else {
		geneList <- as.character(geneList)
	}

	outputTypeList <- c("pdf","png","bmp")
	if (!outputType %in% outputTypeList) {
	stop("ERROR: The output Type ", outputType ," is invalid. Please select one from pdf, png, and bmp.")
	}

	bpRes <- .processData(organism, hostName, geneList, "Biological_Process")
	bpGoCnts <- bpRes$goTermCounts
	bpUnclassified <- bpRes$dataUnclassified
	if (isOutput) {
		write.table(bpGoCnts, paste0(outputFile, "_bp.txt"), row.names=FALSE, col.names=FALSE, sep="\t", quote=FALSE)
	}

	ccRes <- .processData(organism, hostName, geneList, "Cellular_Component")
	ccGoCnts <- ccRes$goTermCounts
	ccUnclassified <- ccRes$dataUnclassified
	if (isOutput) {
		write.table(ccGoCnts, paste0(outputFile, "_cc.txt"), row.names=FALSE, col.names=FALSE, sep="\t", quote=FALSE)
	}

	mfRes <- .processData(organism, hostName, geneList, "Molecular_Function")
	mfGoCnts <- mfRes$goTermCounts
	mfUnclassified <- mfRes$dataUnclassified
	if (isOutput) {
		write.table(mfGoCnts, paste0(outputFile, "_mf.txt"), row.names=FALSE, col.names=FALSE, sep="\t", quote=FALSE)
	}

	if (outputType=="pdf") {
		pdf(paste0(outputFile, ".pdf"), height=8, width=16)
	} else if (outputType=="png") {
		png(paste0(outputFile, ".png"), width=2800, height=1300, res=200)
	} else if (outputType=="bmp") {
		bmp(paste0(outputFile, ".bmp"), width=2800, height=1300, res=200)
	}

	layout(matrix(1:3,1,3))

	.plotData(geneList, bpGoCnts, bpUnclassified, "Biological Process", "red")
	.plotData(geneList, ccGoCnts, ccUnclassified, "Cellular Component", "blue")
	.plotData(geneList, mfGoCnts, mfUnclassified, "Molecular Function", "green")

	dev.off()
	return(NULL)
}

#' @importFrom httr POST content
#' @importFrom dplyr select distinct inner_join arrange filter %>%
#' @importFrom readr read_tsv
.processData <- function(organism, hostName, geneList, ontology) {
	names <- c("entrezgene", "accession", "name")
	if (startsWith(hostName, "file://")) {
		goPath <- removeFileProtocol(file.path(hostName, "goslim", paste0(organism, "_GOSlim_", ontology, ".table")))
		goSlimData <- read_tsv(goPath, col_names=names, col_types="ccc", quote="")
		goSlimData <- filter(goSlimData, .data$entrezgene %in% geneList)
	} else {
		goUrl <- file.path(hostName, "api", "goslim")
		response <- POST(goUrl, body=list(organism=organism, ontology=ontology, entrezgenes=geneList), encode="json")
		if (response$status_code != 200) {
			stop(webRequestError(response))
		}
		goSlimData <- content(response)[["goslim"]]
		goSlimData <- data.frame(matrix(unlist(lapply(goSlimData, FUN=function(x) { x[names] })), nrow=length(goSlimData), byrow=TRUE), stringsAsFactors=FALSE)
		colnames(goSlimData) <- names
	}


	if (nrow(goSlimData) == 0) {
		stop("ERROR: The ID type of the uploaded list is not entrez gene id. Please use IDMapping function to map other id types to entrez gene id.")
	}

	dataUnclassified <- setdiff(geneList, unique(goSlimData[["entrezgene"]]))

	goTermCount <- tapply(goSlimData[["entrezgene"]], goSlimData[["accession"]], length)
	goTermCount <- data.frame(accession=names(goTermCount), geneNum=as.numeric(goTermCount), stringsAsFactors=FALSE)
	uniqueGoCounts <- goSlimData %>% select(.data$accession, .data$name) %>% distinct() %>%
		inner_join(goTermCount, by="accession") %>%
		arrange(desc(.data$geneNum))

	re <- list(goTermCounts=uniqueGoCounts,dataUnclassified=dataUnclassified)
	return(re)
}

.plotData <- function(geneList,goCounts,dataUnclassified,ontology,color){
	par(mar=c(20,5,2,2))
	c <- c(length(geneList), goCounts$geneNum, length(dataUnclassified))
	names(c) <- c("all", goCounts$name, "unclassified")
	maxC <- max(c)
	xx <- barplot(c,main=paste("Bar chart of ",ontology," categories",sep=""),col=color,xaxt="n",xlab="",ylim=c(0,1.2*maxC),cex.axis=1.5,las=2,cex.main=1.7)
	text(x=xx+0.3,y=c+maxC*0.05,labels=c,pos=3,cex=1.5,col="black",srt=90)
	axis(side=1,at=xx,labels=names(c),las=2,cex.axis=1.5)
}


#' @export
GOSlimSummary <- function(...) {
	warning("Function GOSlimSummary is deprecated and changed to goSlimSummary!")
	return(goSlimSummary(...))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readGmt.R
\name{prepareInputMatrixGsea}
\alias{prepareInputMatrixGsea}
\title{Prepare Input Matrix for GSEA}
\usage{
prepareInputMatrixGsea(rank, gmt)
}
\arguments{
\item{rank}{A 2 column Data Frame of gene and score}

\item{gmt}{3 column Data Frame of geneSet, description, and gene}
}
\value{
A matrix used for input to \code{swGsea}.
}
\description{
Prepare Input Matrix for GSEA
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cacheFile.R
\name{cacheUrl}
\alias{cacheUrl}
\title{cacheUrl}
\usage{
cacheUrl(dataUrl, cache = NULL, query = NULL)
}
\arguments{
\item{dataUrl}{The URL of data}

\item{cache}{The cache directory. Defaults to \code{NULL} and fetches data from server.}

\item{query}{The list of queries passed on to httr methods}
}
\value{
response object from httr request
}
\description{
Get data from a URL or cache and optionally save in cache for reuse
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reportUtils.R
\name{keepRep}
\alias{keepRep}
\title{keepRep}
\usage{
keepRep(topResult, allResult, reps)
}
\description{
Add representatives of redundancy-reduced clusters to
topResult if they are missing.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrichResultSection.R
\name{enrichResultSection}
\alias{enrichResultSection}
\title{enrichResultSection}
\usage{
enrichResultSection(
  enrichMethod,
  enrichedSig,
  geneSet,
  geneSetDes,
  geneSetDag,
  geneSetNet,
  clusters
)
}
\description{
Conditionally render template of main result section. Actual work is carried out in front end
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/swGsea.R
\name{prepareGseaInput}
\alias{prepareGseaInput}
\title{Prepare input for standard GSEA}
\usage{
prepareGseaInput(rankFile, gmtFile)
}
\arguments{
\item{rankFile}{Path of the rnk file}

\item{gmtFile}{Path of the GMT file}
}
\value{
a data frame to be used in \code{swGsea}
}
\description{
A helper to read files for performing standard GSEA.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linkModification.R
\name{linkModification}
\alias{linkModification}
\title{Modify the link to highlight the genes in the pathways}
\usage{
linkModification(enrichMethod, enrichPathwayLink, geneList, interestingGeneMap)
}
\description{
Currently, we only have wikipathway and kegg pathways that need to modify the link
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatCheck.R
\name{formatCheck}
\alias{formatCheck}
\title{Check Format and Read Data}
\usage{
formatCheck(dataType = "list", inputGeneFile = NULL, inputGene = NULL)
}
\arguments{
\item{dataType}{Type of data, either \code{list}, \code{rnk} or \code{gmt}.
Could be \code{list}, \code{rnk} or \code{matrix} for \code{idToSymbol}.}

\item{inputGeneFile}{The data file to be mapped.}

\item{inputGene}{Or the input could be given as an R object.
GMT file should be read with \code{readGmt}.}
}
\value{
A list of data frame
}
\description{
Check Format and Read Data
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/listArchiveURL.R
\name{listArchiveUrl}
\alias{listArchiveUrl}
\alias{listArchiveURL}
\title{List WebGestalt Servers}
\usage{
listArchiveUrl()
}
\value{
A data frame of available servers.
}
\description{
List available WebGestalt servers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/swGsea.R
\name{swGsea}
\alias{swGsea}
\title{Site Weighted Gene Set Enrichment Analysis}
\usage{
swGsea(
  input_df,
  thresh_type = "percentile",
  thresh = 0.9,
  thresh_action = "exclude",
  min_set_size = 10,
  max_set_size = 500,
  max_score = "max",
  min_score = "min",
  psuedocount = 0.001,
  perms = 1000,
  p = 1,
  q = 1,
  nThreads = 1,
  rng_seed = 1,
  fork = FALSE
)
}
\arguments{
\item{input_df}{A data frame in which first column is name of item of interest
(gene, protein, phosphosite, etc.), the second is the correlation of that item
of interest with the phenotype (typically log ratio of expression for phenotype
vs. normal), and the remaining columns are the scores for the likelihood that
the item belongs in each set (one column per set).}

\item{thresh_type}{The type of \code{thresh}. Use 'percentile' to include all
scores over that percentile given in \code{thresh} (i.e., 0.9 would be all items
in 90th percentile, or top 10 percent); 'list' to include a list of set lists
where the set lists are in the same order as the corresponding set columns in
the \code{input_df}; 'val' to apply a single threshold value to all sets; or
'values' to use a vector of unique cutoffs for each set (needs to be in the
same order as the sets are specified in the columns of \code{input_df}")}

\item{thresh}{Depends on \code{thresh_type}. A list of lists of the items in
each set (with same names as colnames of the scores); a numeric vector of
threshold scores for each set (in the same order as the colnames of the scores
in the input_df), or a single percentile value between 0 and 1 (i.e., if
\code{thresh}=0.9, the 90th percentile of the score or the highest scoring 10%
of of the items are included in the set for each scoring regimen) (\code{thresh}
="all" is not supported at this time, as it doesn't result in a Kolgorov-Smirnoff
statistic; this may be worked in as an alternate scoring method later on).}

\item{thresh_action}{Either "include", "exclude (default)", or "adjust";
this specifies how to treat each set if it doesn't contain a minimum number of
items or contains all of the items; this option cannot be used with predefined
lists of items in sets (if the number of items in a given set doesn't meet
requirements, that set will be skipped).}

\item{min_set_size, max_set_size}{The minimum/maximum number of items each
set needs for the analysis to proceed.}

\item{max_score, min_score}{A optional numeric vector of minimum/maximum boundaries
to clip scores for each set.}

\item{psuedocount}{Psuedocount (pc) is used for rescaling set scores:
\code{(score - min_score + pc)/(max_score - min_score +pc)}; this is needed to
prevent division by 0 if \code{max_score==min_score} (in this case, all scores
for items in set will be 1, which is equivalent to standard GSEA); it also allows
users to adjust weights for scores that are close to the minimum for the scores in
the set (unless min_score==max_score): as psuedocount value approaches 0, scaled
minimum scores also approach 0; as psuedocount approaches infinity, scaled minimum
scores approach the scaled maximum scores (which equal 1); this value must be
larger than 0.}

\item{perms}{The number of permutations.}

\item{p}{The exponential scaling factor of the phenotype score (second column in
\code{input_df}).}

\item{q}{The exponential scaling factor of the likelihood score (weights).}

\item{nThreads}{The number of threads to use in calculating permutaions.}

\item{rng_seed}{Random seed.}

\item{fork}{A boolean. Whether pass "fork" to \code{type} parameter of
\code{makeCluster} on Unix-like machines.}
}
\value{
A list of \code{Enrichment_Results}, \code{Items_in_Set} and \code{Running_Sums}.
\describe{
 \item{Enrichment_Results}{A data frame with row names of gene set and columns of
 "ES", "NES", "p_val", "fdr".}
 \item{Items_in_Set}{A list of one-column data frames. Describes genes and their
 ranks in each set.}
 \item{Running_Sums}{Running sum scores along genes sorted by ranked scores,
 with gene sets as columns.}
}
}
\description{
Performs site weighted gene set enrichment analysis or standard GSEA when
likelihood/weight columns in \code{input_df} are 1 or 0, \code{p=1},
\code{q=1} and \code{thresh_type="val"}.
}
\details{
The formula for weighting is as follows
\deqn{\frac{s_{j}^{q}|r_{j}|^{p}}{\sum s^{q}|r|^{p}}}{(s_j^q * abs(r_j)^p) / (\sum s^q * abs(r)^p)}
Where r is log ratio score, s is likelihood score, j is the index of the gene.
}
\author{
Eric Jaehnig
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createReport.R
\name{createReport}
\alias{createReport}
\title{createReport}
\usage{
createReport(
  hostName,
  outputDirectory,
  organism = "hsapiens",
  projectName,
  enrichMethod,
  geneSet,
  geneSetDes,
  geneSetDag,
  geneSetNet,
  interestingGeneMap,
  referenceGeneList,
  enrichedSig,
  geneTables,
  clusters,
  background,
  enrichDatabase = NULL,
  enrichDatabaseFile = NULL,
  enrichDatabaseType = NULL,
  enrichDatabaseDescriptionFile = NULL,
  interestGeneFile = NULL,
  interestGene = NULL,
  interestGeneType = NULL,
  collapseMethod = "mean",
  referenceGeneFile = NULL,
  referenceGene = NULL,
  referenceGeneType = NULL,
  referenceSet = NULL,
  minNum = 10,
  maxNum = 500,
  fdrMethod = "BH",
  sigMethod = "fdr",
  fdrThr = 0.05,
  topThr = 10,
  reportNum = 20,
  perNum = 1000,
  p = 1,
  dagColor = "binary"
)
}
\description{
Generate HTML report for ORA and GSEA
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/listGeneSet.R
\name{listGeneSet}
\alias{listGeneSet}
\title{List Gene Sets}
\usage{
listGeneSet(
  organism = "hsapiens",
  hostName = "http://www.webgestalt.org/",
  cache = NULL
)
}
\arguments{
\item{organism}{Currently, WebGestaltR supports 12 organisms. Users can use the function
\code{listOrganism} to check available organisms. Users can also input \code{others} to
perform the enrichment analysis for other organisms not supported by WebGestaltR. For
other organisms, users need to provide the functional categories, interesting list and
reference list (for ORA method). Because WebGestaltR does not perform the ID mapping for
the other organisms, the above data should have the same ID type.}

\item{hostName}{The server URL for accessing data. Mostly for development purposes.}

\item{cache}{A directory to save data cache for reuse. Defaults to \code{NULL} and disabled.}
}
\value{
A data frame of available gene sets.
}
\description{
List available gene sets for the given organism on WebGestalt server.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weightedSetCover.R
\name{weightedSetCover}
\alias{weightedSetCover}
\title{Weighted Set Cover}
\usage{
weightedSetCover(idsInSet, costs, topN, nThreads = 4)
}
\arguments{
\item{idsInSet}{A list of set names and their member IDs.}

\item{costs}{A vector of the same length to add weights for penalty, i.e. 1/-logP.}

\item{topN}{The number of sets (or less when it completes early) to return.}

\item{nThreads}{The number of processes to use. In Windows, it fallbacks to 1.}
}
\value{
A list of \code{topSets} and \code{coverage}.
\describe{
 \item{topSets}{A list of set IDs.}
 \item{coverage}{The percentage of IDs covered in the top sets.}
}
}
\description{
Size constrained weighted set cover problem to find top N sets while
maximizing the coverage of all elements.
}
\author{
Zhiao Shi, Yuxing Liao
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summaryDescription.R
\name{summaryDescription}
\alias{summaryDescription}
\title{summaryDescription}
\usage{
summaryDescription(
  projectName,
  organism,
  interestGeneFile,
  interestGene,
  interestGeneType,
  enrichMethod,
  enrichDatabase,
  enrichDatabaseFile,
  enrichDatabaseType,
  enrichDatabaseDescriptionFile,
  interestingGeneMap,
  referenceGeneList,
  referenceGeneFile,
  referenceGene,
  referenceGeneType,
  referenceSet,
  minNum,
  maxNum,
  sigMethod,
  fdrThr,
  topThr,
  fdrMethod,
  enrichedSig,
  reportNum,
  perNum,
  p,
  geneSet,
  repAdded,
  numAnnoRefUserId,
  hostName
)
}
\description{
Render job summary section
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affinityPropagation.R
\name{jaccardSim}
\alias{jaccardSim}
\title{Jaccard Similarity}
\usage{
jaccardSim(idsInSet, score)
}
\arguments{
\item{idsInSet}{A list of set names and their member IDs.}

\item{score}{A vector of addible scores with the same length used to assign input preference;
higher score has larger weight, i.e. -logP.}
}
\value{
A list of similarity matrix \code{sim.mat} and input preference vector \code{ip.vec}.
}
\description{
Calculate Jaccard Similarity.
}
\author{
Zhiao Shi, Yuxing Liao
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affinityPropagation.R
\name{affinityPropagation}
\alias{affinityPropagation}
\title{Affinity Propagation}
\usage{
affinityPropagation(idsInSet, score)
}
\arguments{
\item{idsInSet}{A list of set names and their member IDs.}

\item{score}{A vector of addible scores with the same length used to assign input preference;
higher score has larger weight, i.e. -logP.}
}
\value{
A list of \code{clusters} and \code{representatives} for each cluster.
\describe{
 \item{clusters}{A list of character vectors of set IDs in each cluster.}
 \item{representatives}{A character vector of representatives for each cluster.}
}
}
\description{
Use affinity propagation to cluster similar gene sets to reduce redundancy in report.
}
\author{
Zhiao Shi, Yuxing Liao
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readGmt.R
\name{readGmt}
\alias{readGmt}
\title{Read GMT File}
\usage{
readGmt(gmtFile, cache = NULL)
}
\arguments{
\item{gmtFile}{The file path or URL of the GMT file.}

\item{cache}{A directory to save data cache for reuse. Defaults to \code{NULL} and disabled.}
}
\value{
A data frame with columns of "geneSet", "description", "gene".
}
\description{
Read GMT File
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/listOrganism.R
\name{listOrganism}
\alias{listOrganism}
\title{List Organisms}
\usage{
listOrganism(hostName = "https://www.webgestalt.org/", cache = NULL)
}
\arguments{
\item{hostName}{The server URL for accessing data. Mostly for development purposes.}

\item{cache}{A directory to save data cache for reuse. Defaults to \code{NULL} and disabled.}
}
\value{
A list of supported organisms.
}
\description{
List supported organisms on WebGestalt server.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{fillInputDataFrame}
\alias{fillInputDataFrame}
\title{Fill relation data frame for GSEA input}
\usage{
fillInputDataFrame(gmt, genes, geneSets)
}
\arguments{
\item{gmt}{A Data Frame with geneSet and gene columns from the GMT file}

\item{genes}{A vector of genes}

\item{geneSets}{A vector of gene sets}
}
\value{
A Data Frame with the first column of gene and 1 or 0 for other columns of gene sets.
}
\description{
Fill 1 for gene in gene set
}
\author{
Yuxing Liao
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{gseaPermutation}
\alias{gseaPermutation}
\title{Permutaion in GSEA algorithm}
\usage{
gseaPermutation(inset_scores, outset_scores, expression_value)
}
\arguments{
\item{inset_scores}{Scaled score matrix for genes in sets}

\item{outset_scores}{Normalized score matrix for genes not in sets}

\item{expression_value}{Vector of gene rank scores}
}
\value{
A vector of concatenated random minimal,maimum and best running sum scores for each set.
}
\description{
Permutaion in GSEA algorithm
}
\author{
Yuxing Liao
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createNtaReport.R
\name{createNtaReport}
\alias{createNtaReport}
\title{Create HTML Report for NTA}
\usage{
createNtaReport(
  networkName,
  method,
  sigMethod,
  fdrThr,
  topThr,
  highlightType,
  outputDirectory,
  projectDir,
  projectName,
  hostName
)
}
\description{
Create HTML Report for NTA
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dagUtils.R
\name{expandDag}
\alias{expandDag}
\title{Expand enriched GO IDs to include ancestors up to the root}
\usage{
expandDag(goTermList, dagEdgeList)
}
\description{
Returns expanded nodes and DAG tree edges
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summaryDescription.R
\name{specificParameterSummaryGsea}
\alias{specificParameterSummaryGsea}
\title{specificParameterSummaryGsea}
\usage{
specificParameterSummaryGsea(
  organism,
  interestingGeneMap,
  geneSet,
  minNum,
  maxNum,
  sigMethod,
  fdrThr,
  topThr,
  perNum,
  p,
  enrichedSig,
  reportNum,
  repAdded
)
}
\description{
Render job summary section of GSEA specific parameters
}
\keyword{internal}
